#!/bin/bash
EOSNF=1
KVER=`uname -r`
prefix=
MODULES="myri_snf myri_eosnf"
MODULES_ALL="myri_snf,myri_eosnf,myri_snf_kbridge,myri_snf_kapi"
DEFINES="-DMAL_KERNEL -DMX_THREAD_SAFE=1"

KMAJOR=$(echo $KVER | cut -d- -f1 | sed 's/\.//g')
KSTEP=$(echo $KVER | cut -d- -f2 | cut -d. -f2)

usage() {
    echo "`basename $0` [--kernel|-k <kernel version>] [--modules|-m <modules>] [<snf base dir>]"
    echo ""
    echo "  --modules : Specify comma-separated list of modules to build from:"
    echo "              $MODULES_ALL"
    echo "              If unspecified, defaults to 'myri_snf,myri_eosnf'"
    echo "              NOTE: Overriden by the value of SNF_MODULES in the environment"
    echo " -D[DEFINE] : Add a preprocessor definition as part of the kernel build"
    echo ""
    exit $1
}

build_sub_module()
{
  local module=$1; shift
  local obj_deps=$1; shift
  local mod_desc=$1; shift

  echo -n "  *** ${module}.ko ($mod_desc) ... ";
  (
  set -x
  set -e
  cat >$KBUILD_FILE <<EOF
EXTRA_CFLAGS += $DEFINES
EXTRA_CFLAGS += -DMYRI_DRIVER=myri_snf
EXTRA_CFLAGS += -DMYRI_LINUX_MAJOR=$KMAJOR
EXTRA_CFLAGS += -DMYRI_LINUX_STEPPING=$KSTEP
EXTRA_CFLAGS += -I$prefix/src/driver/linux
EXTRA_CFLAGS += -I$prefix/src/driver/common
EXTRA_CFLAGS += -I$prefix/src/common
EXTRA_CFLAGS += -I$prefix/src/common
EXTRA_CFLAGS += -I$prefix/src/common/fapi
EXTRA_CFLAGS += -I$prefix/include
EXTRA_CFLAGS += -g -Wall -Werror
KBUILD_EXTRA_SYMBOLS = $prefix/src/driver/linux/kbuild/myri_snf.symvers
obj-m = ${module}.o
${module}-objs = $obj_deps
EOF
  make -C /lib/modules/$KVER/build M=$prefix/src/driver/linux/kbuild clean
  make -C /lib/modules/$KVER/build M=$prefix/src/driver/linux/kbuild
  cp $prefix/src/driver/linux/kbuild/${module}.ko $prefix/sbin
  ) >> /tmp/myri_snf.log 2>&1
  err=$?
  if test $err != 0; then
    echo "error"
    echo "ERROR building ${module}.ko, output available in /tmp/myri_snf.log"
    exit 1
  else
    echo "ok"
  fi
  return $err
}

while test $# -gt 0; do
  case "$1" in
    --kernel|-k)
      shift
      KVER=$1
      ;;
    --help|-h)
      usage 0
      ;;
    --modules|-m)
      shift
      test -z "$1" && usage 1 
      MODULES=$1
      ;;
    -D*)
      DEFINES="$DEFINES $1"
      ;;
    -*)
      usage 1
      ;;
    *) test -z "$prefix" || usage 1
       prefix=$1
       ;;
  esac
  shift
done

if test ! -d "/lib/modules/$KVER"; then
    echo "'/lib/modules/$KVER/build' doesn't exits. Check your kernel version"
    echo "and/or that you have installed the correct packages to enable compiling"
    echo "of kernel modules."
    usage
fi

if test -z "$prefix"; then
  if test -n "$SNF_PREFIX"; then
    prefix=$SNF_PREFIX
  else
    prefix=`cd $(dirname $0)/..; pwd`
  fi
fi
KBUILD_FILE=$prefix/src/driver/linux/kbuild/Kbuild

build_myri_snf() {
  if test -n "$MYRI_MCP" -a \
          -f "$prefix/src/driver/common/$MYRI_MCP/mcp_array_10g_prebuilt.c"; then
    myri_mcp=$MYRI_MCP
    cp $prefix/src/driver/common/$MYRI_MCP/mcp_array_10g_prebuilt.c $prefix/src/driver/linux/kbuild/mcp_array_10g.c
  else
    myri_mcp=default
  fi
  echo -n "  *** myri_snf.ko with MYRI_MCP=$myri_mcp ... "

  #regenerate mal checks
  (
    set -x
    set -e
    $prefix/src/driver/linux/mal_check_headers.sh /lib/modules/$KVER/source \
  	/lib/modules/$KVER/build $prefix/src/driver/linux/mal_checks.h
  
    #replace Kbuild file
    cat >$KBUILD_FILE <<EOF
EXTRA_CFLAGS += $DEFINES
EXTRA_CFLAGS += -DMYRI_DRIVER=myri_snf
EXTRA_CFLAGS += -DMYRI_LINUX_MAJOR=$KMAJOR
EXTRA_CFLAGS += -DMYRI_LINUX_STEPPING=$KSTEP
EXTRA_CFLAGS += -I$prefix/src/driver/linux
EXTRA_CFLAGS += -I$prefix/src/driver/common
EXTRA_CFLAGS += -I$prefix/src/common
EXTRA_CFLAGS += -I$prefix/src/common
EXTRA_CFLAGS += -I$prefix/src/common/fapi
EXTRA_CFLAGS += -I$prefix/include
EXTRA_CFLAGS += -g -Wall -Werror
obj-m = myri_snf.o
myri_snf-objs = mx.o mx_ether.o mx_common.o mx_instance.o mx_lanai_command.o mcp_wrapper_common.o mx_lx.o mx_lz.o mx_ether_common.o kraw.o myri_ptp_common.o myri_raw_ops.o myri_snf_ops.o
ifndef CONFIG_ZLIB_INFLATE
EXTRA_CFLAGS += -I$prefix/src/driver/zlib
myri_snf-objs += adler32.o crc32.o infblock.o infcodes.o inffast.o inflate.o inftrees.o infutil.o trees.o uncompr.o zutil.o
endif
EOF
  make -C /lib/modules/$KVER/build M=$prefix/src/driver/linux/kbuild clean
  make -C /lib/modules/$KVER/build M=$prefix/src/driver/linux/kbuild
  cp $prefix/src/driver/linux/kbuild/myri_snf.ko $prefix/sbin/
  cp $prefix/src/driver/linux/kbuild/Module.symvers $prefix/src/driver/linux/kbuild/myri_snf.symvers
  for file in $prefix/sbin/*_start_stop.orig; do
      newfile=${file%.orig}
      sed -e "s,MYRI_MODULE_DIR=/opt/snf,MYRI_MODULE_DIR=$prefix," < ${file} > ${newfile}
      chmod +x ${newfile}
  done

) >> /tmp/myri_snf.log 2>&1
  if test $? != 0; then
    echo "error"
    echo "ERROR building myri_snf.ko, output available in /tmp/myri_snf.log"
    exit 1
  else
    echo "ok"
  fi
  return $?
}

#echo "obj-m += myri_eosnf.o" >> $KBUILD_FILE
#echo "myri_eosnf-objs += myri_snf_eosnf.o myri_snf_kapi_obj.o" >> $KBUILD_FILE

if test ! -e $prefix/sbin/rebuild.sh; then
  echo "'$prefix' doesn't seem to be the snf base directory"
  echo "If you specified the prefix on the command line or through"
  echo "SNF_PREFIX, please check your path."
  usage 1
fi

if test -n "$SNF_MODULES"; then
  echo "WARNING: Overriding --modules with environment variable SNF_MODULES=$SNF_MODULES"
  MODULES=$SNF_MODULES
fi
if test "$MODULES" = "all"; then
  MODULES=$MODULES_ALL
fi

rm -f /tmp/myri_snf.log

hdr=
for module in `echo $MODULES | tr ',:;' ' '`; do
  if test -z "$hdr"; then
    hdr=1
    echo "kernel      = $KVER"
    echo "destination = $prefix/sbin"
  fi

  case $module in
    myri_snf)
      build_myri_snf
      ;;
    myri_eosnf)
      build_sub_module $module "myri_snf_eosnf.o myri_snf_kapi_obj.o" \
                       "Virtual Ethernet adapter over SNF/klib"
      ;;
    myri_snf_kbridge)
      build_sub_module $module "myri_snf_kbridge_module.o snf_bridge.o myri_snf_kapi_obj.o" \
                        "Example of transparent bridge over SNF/klib"
      ;;
    myri_snf_kapi)
      build_sub_module $module "myri_snf_kapi_module.o myri_snf_kapi_obj.o" \
                        "Standalone kernel API/SNF module"
      ;;

    *)
      echo "ERROR: Unknown module $module"
      exit 1
      ;;
  esac
done


