/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2008 - 2010 by Myricom, Inc.  All rights reserved.
 * Copyright 2015 by CSP, Inc.  All rights reserved.
 ***********************************************************************/

#ifndef _WIN32
#include <inttypes.h>
#endif
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#ifndef _WIN32
#define __USE_GNU
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <netinet/in.h>
#else
#include <winsock2.h>
#include <windows.h>
#include <signal.h>
#include "getopt.h"
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "pcap.h"
#ifdef SNF_TREE
#include "mal_timing.h"
#else
#include "mal_simplified.h"
#endif

#ifdef _WIN32
#define C_API __cdecl
#else
#define C_API
#endif

#ifdef __FreeBSD__
#include <sys/param.h>
#if __FreeBSD_version < 702100
#define PTHREAD_SETAFFINITY_NP_MISSING
#else
#include <pthread_np.h>
#define cpu_set_t cpuset_t
#endif /* __FreeBSD_version */
#elif defined(__GLIBC__)
#if !__GLIBC_PREREQ(2,4)
#define PTHREAD_SETAFFINITY_NP_MISSING
#endif
#endif
#define MAX_WORKERS               32
#define MAX_CONC_RECVS            1024

#ifndef _WIN32
#define DBG1(vargs...) if (G.verbose) printf(vargs)
#else
#define DBG1(...) if (G.verbose) printf(__VA_ARGS__)
#endif


#define TV_TO_US(tv) ((tv)->tv_sec * 1000000 + (tv)->tv_usec)

struct worker {
  int      worker_id;
  int      cpu_id;
  pcap_t   *hring;
#ifndef _WIN32
  pthread_t thread_id;
#else
  HANDLE thread_id;
#endif
  uint64_t num_pkts[2];
  uint64_t num_bytes[2];
  mal_cycles_t cyc_itvl;
};
typedef struct worker worker_t;

struct G {
  char     *device;
  unsigned snap_len;
  unsigned toggle;
  uint64_t num_pkts[MAX_WORKERS];
  uint64_t num_bytes[MAX_WORKERS];
  uint64_t num_pkts_expected;
  uint64_t total_pkts;
  uint64_t total_bytes;
  uint64_t exit_usecs;
  unsigned duration;
  unsigned custom_hash_rr;
  int64_t  delay_nsecs;
  unsigned validate;
  unsigned verbose;
  int      recv_wait;
  unsigned num_workers;
  unsigned num_workers_ready;
  unsigned expired;
  uint64_t cpumask;
  worker_t worker[MAX_WORKERS];
#ifndef _WIN32
  pthread_mutex_t wrk_mtx;
  pthread_cond_t  wrk_cond;
#else
  HANDLE wrk_mtx;
  HANDLE wrk_sem;
#endif
  struct timeval tv_last;
  uint64_t usecs_last;
  unsigned periodic_stats;
} G;

static void
usage(const char *progname)
{
  printf("usage: %s -w <num_workers> [options]\n\n", progname);
  printf(" options:\n");
  printf("     -n <num pkts>: number of packet to receive (default: 0 - infinite)\n");
  printf("                -t: show periodic stats, every second\n");
  printf("                -v: verbose, or -vv very verbose\n");
  printf("       -d <device>: packet capture device\n");
  printf("     -D <nanosecs>: add <nanosecs> of synthetic processing delay in packet handling\n");
  printf("     -S <snap len>: return first <snap len> bytes of each packet\n");
  printf("      -T <timeout>: quit application after timeout seconds\n");
  printf("\n");
  exit(1);
}

void
stats(void)
{
  worker_t *wrk;
  unsigned int w;
  uint64_t num_pkts = 0;
  uint64_t num_bytes = 0;
  int have_nic_stats = 0;
  struct pcap_stat stats;

  for (w = 0; w < G.num_workers; w++) {
    wrk = &G.worker[w];
    num_pkts += G.num_pkts[w];
    num_bytes += G.num_bytes[w];
    if (wrk->hring == NULL || pcap_stats(wrk->hring, &stats)) {
      printf("[%u] %" PRIu64 " pkts, %" PRIu64" bytes\n", w, G.num_pkts[w],
              G.num_bytes[w]);
    }
    else {
      have_nic_stats = 1;
      printf("Ring %d Packets received,    ring:%12" PRIu32 ", app:%12" PRIu64 
             "\n", w, stats.ps_recv, G.num_pkts[w]);
    }
  }

  if (have_nic_stats) {
    /* NIC-level stats */
    printf("NIC dropped overflow SW: %12" PRIu32 "\n", stats.ps_drop);
    printf("NIC dropped overflow HW: %12" PRIu32 "\n", stats.ps_ifdrop);
  }

  printf("\n");
  printf("Total packets:                app:%12" PRIu64 "\n", num_pkts);
  printf("Average Packet Length:        app:%12" PRIu64 " bytes\n\n",
            (num_pkts ? (num_bytes / num_pkts) : 0));
  printf("Total bytes:                  app:%12" PRIu64 " (%" PRIu64 " MB)\n",
          num_bytes, num_bytes / 1024 / 1024);
}

static int
cpu_getmask_ith(unsigned cpumask, int id)
{
  int i, i_found = 0;
  unsigned mask;
  for (i = 0; i < sizeof(cpumask)*8; i++) {
    mask = (1U<<i);
    if (mask & cpumask) {
      if (i_found == id) /* cpu i to be used for id */
        return i;
      i_found++;
    }
  }
  /* not found, return any cpu */
  return -1;
}

void
print_periodic_stats(void)
{
  float delta_secs;
  uint32_t pps;
  float gbps;
  float bps;
  uint64_t num_pkts;
  uint64_t num_bytes;
  struct timeval tv_now;
  uint64_t usecs_now;
  worker_t *wrk;
  unsigned int w;
  int flip;

  gettimeofday(&tv_now, NULL);
  usecs_now = TV_TO_US(&tv_now);
  delta_secs = (float)((usecs_now - G.usecs_last) / 1000000.0);

  if (G.duration && G.expired == 0) {
    if (!G.exit_usecs)
      G.exit_usecs = usecs_now + G.duration * 1000000;
    else if (usecs_now >= G.exit_usecs) {
      G.expired = 1;
    }
  }

  /* flip the flip-flop so threads are using other counter set */
  flip = G.toggle;
  G.toggle = (flip ? 0 : 1);
    
  num_pkts = 0;
  num_bytes = 0;
  for (w = 0; w < G.num_workers; w++) {

    wrk = &G.worker[w];

    /* it's possible that the periodic time fires before all rings are started */
    if (wrk->hring == NULL)
      continue;

    G.num_pkts[w] += wrk->num_pkts[flip];
    num_pkts += wrk->num_pkts[flip];
    G.num_bytes[w] += wrk->num_bytes[flip];
    num_bytes += wrk->num_bytes[flip];
  }

  if (num_pkts) {
    /* Update the last count only if we are going to print stats below.
     * Otherwise, num_pkts and num_bytes may get out of sync.
     * For example, delta_pkts > 0 and delta_bytes = 0.
     * The same issue applies to the saved timestamps.
     */
    G.total_pkts += num_pkts;
    G.total_bytes += num_bytes;
    G.usecs_last = usecs_now;
    G.tv_last = tv_now;

    if (G.periodic_stats) {
      pps = (uint32_t)(num_pkts / delta_secs);
      bps = ((float) num_bytes * 8) / delta_secs;
      gbps = (float)(bps / 1000000000.0);

      for (w = 0; w < G.num_workers; w++) {
        wrk = &G.worker[w];
        DBG1("[%u] %" PRIu64 " ", w, wrk->num_pkts[flip]);
      }

      printf  ("%" PRIu64 " pkts (%" PRIu64 "B) in %.3f secs (%u pps), Avg Pkt: %"
               PRIu64 ", BW (Gbps): %6.3f\n",
               num_pkts, num_bytes, delta_secs,
               pps, num_bytes / num_pkts, gbps);
      fflush(stdout);
    }
  }

  for (w = 0; w < G.num_workers; w++) {
    wrk = &G.worker[w];

    /* clear the counters for next flip */
    wrk->num_pkts[flip] = 0;
    wrk->num_bytes[flip] = 0;

    if (wrk->hring == NULL)
      continue;

    /* break the worker out if task completed */
    if (G.total_pkts >= G.num_pkts_expected) {
      DBG1("[%d] done, %llu packets received\n", w, (unsigned long long) G.total_pkts);
      pcap_breakloop(wrk->hring);
    } 
    else if (G.expired) {
      DBG1("[%d] done, times up\n", w);
      pcap_breakloop(wrk->hring);
    }
  }
}

void C_API
sigexit(int sig)
{
#ifndef _WIN32
  static pthread_mutex_t exit_mtx = PTHREAD_MUTEX_INITIALIZER;
  static int once = 0;

  pthread_mutex_lock(&exit_mtx);
  if (!once) {
    once = 1;
    stats();
    fflush(stdout);
  }
  pthread_mutex_unlock(&exit_mtx);
  if (sig >= 0)
    exit(0);
#else
  static volatile LONG once = 0;
  LONG l;

  l = InterlockedIncrement(&once);
  if (l == 1) {
    stats();
    fflush(stdout);
  }
  if (sig >= 0)
    exit(0);
#endif
}


#ifndef _WIN32
void
sigalrm(int sig)
{
  print_periodic_stats();
  alarm(1);
}
#else
DWORD
stats_thread(LPVOID lpParameter)
{
  for (;;) {
    print_periodic_stats();
    Sleep(1000);
  }
}
#endif

/* default snap length (maximum bytes per packet to capture) */
#define SNAP_LEN 1518

/* ethernet headers are always exactly 14 bytes [1] */
#define SIZE_ETHERNET 14

/* Ethernet addresses are 6 bytes */
#define ETHER_ADDR_LEN  6

/* Ethernet header */
struct sniff_ethernet {
        u_char  ether_dhost[ETHER_ADDR_LEN];    /* destination host address */
        u_char  ether_shost[ETHER_ADDR_LEN];    /* source host address */
        u_short ether_type;                     /* IP? ARP? RARP? etc */
};


/* IP header */
struct sniff_ip {
        u_char  ip_vhl;                 /* version << 4 | header length >> 2 */
        u_char  ip_tos;                 /* type of service */
        u_short ip_len;                 /* total length */
        u_short ip_id;                  /* identification */
        u_short ip_off;                 /* fragment offset field */
        #define IP_RF 0x8000            /* reserved fragment flag */
        #define IP_DF 0x4000            /* dont fragment flag */
        #define IP_MF 0x2000            /* more fragments flag */
        #define IP_OFFMASK 0x1fff       /* mask for fragmenting bits */
        u_char  ip_ttl;                 /* time to live */
        u_char  ip_p;                   /* protocol */
        u_short ip_sum;                 /* checksum */
        struct  in_addr ip_src,ip_dst;  /* source and dest address */
};
#define IP_HL(ip)               (((ip)->ip_vhl) & 0x0f)
#define IP_V(ip)                (((ip)->ip_vhl) >> 4)

/* TCP header */
typedef u_int tcp_seq;

struct sniff_tcp {
        u_short th_sport;               /* source port */
        u_short th_dport;               /* destination port */
        tcp_seq th_seq;                 /* sequence number */
        tcp_seq th_ack;                 /* acknowledgement number */
        u_char  th_offx2;               /* data offset, rsvd */
#define TH_OFF(th)      (((th)->th_offx2 & 0xf0) >> 4)
        u_char  th_flags;
        #define TH_FIN  0x01
        #define TH_SYN  0x02
        #define TH_RST  0x04
        #define TH_PUSH 0x08
        #define TH_ACK  0x10
        #define TH_URG  0x20
        #define TH_ECE  0x40
        #define TH_CWR  0x80
        #define TH_FLAGS        (TH_FIN|TH_SYN|TH_RST|TH_ACK|TH_URG|TH_ECE|TH_CWR)
        u_short th_win;                 /* window */
        u_short th_sum;                 /* checksum */
        u_short th_urp;                 /* urgent pointer */
};

static void
got_packet(u_char *stuff, const struct pcap_pkthdr *header, const u_char *packet)
{
  worker_t *wrk;
  mal_cycles_t cyc_start;
  
  if (stuff == NULL)
    return;

  wrk = (worker_t *)stuff;

  wrk->num_pkts[G.toggle]++;
  wrk->num_bytes[G.toggle] += header->len;

  if (G.delay_nsecs > 0) {
    cyc_start = mal_get_cycles();
    while (mal_get_cycles() < cyc_start + wrk->cyc_itvl);
  }

  // send slowly if enabled
  // DBG1(" got packet worker %d %lu %u bytes\n", wrk->worker_id, wrk->num_pkts, header->len );
}


#ifndef _WIN32
void *
#else
DWORD WINAPI
#endif
work_thread(void *stuff)
{
  int worker_id;
  worker_t *wrk;
  int rc=0;
#ifndef _WIN32
  sigset_t sigmask;
#endif
  mal_cycles_t cycles_per_second;
  pcap_t *handle;
  char *dev;
  char errbuf[PCAP_ERRBUF_SIZE];

  wrk = (worker_t *)stuff;
  wrk->hring = NULL;
  worker_id = wrk->worker_id;

#ifdef _WIN32
  if (wrk->cpu_id >= 0)
  {
    SYSTEM_INFO info;
    DWORD tmp;

    GetSystemInfo(&info);
    tmp = wrk->cpu_id % info.dwNumberOfProcessors;
    SetThreadIdealProcessor(GetCurrentThread(), tmp);
    printf("Worker %d running on CPU %d\n", worker_id, wrk->cpu_id);
  }
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  if (wrk->cpu_id >= 0)
  {
    cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(wrk->cpu_id, &cpuset);
    pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
    printf("Worker %d running on CPU %d\n", worker_id, wrk->cpu_id);
  }
#endif

#ifndef _WIN32
  /* mask signals except SIGINT to allow ^C */
  sigfillset(&sigmask);
#if !defined __FreeBSD__
  sigdelset(&sigmask, SIGINT);
#endif
  rc = pthread_sigmask(SIG_BLOCK, &sigmask, NULL);
  if (rc)
    printf("Worker %d error %d setting sigmask\n", wrk->worker_id, rc);
#endif

  mal_cycles_counter_init();
  cycles_per_second = mal_cycles_per_second();
  wrk->cyc_itvl = (mal_cycles_t) (G.delay_nsecs * cycles_per_second / 1000000000.0);

  dev = G.device;
  if (dev == NULL) {
    dev = pcap_lookupdev(errbuf);
    if (dev == NULL) {
      fprintf(stderr, "Couldn't find default device: %s\n", errbuf);
      exit(EXIT_FAILURE);
    }
  }

  handle = pcap_open_live(dev, G.snap_len, 1, 1000, errbuf);
  if (handle == NULL) {
    fprintf(stderr, "Couldn't open device %s: %s\n", dev, errbuf);
    exit(EXIT_FAILURE);
  }

  if (pcap_datalink(handle) != DLT_EN10MB) {
    fprintf(stderr, "[%d] Can't open ring %d as ethernet\n", worker_id, worker_id);
    goto done_thread;
  }

  wrk->hring = handle;
  DBG1("[%u] started\n", worker_id);

#ifndef _WIN32
  pthread_mutex_lock(&G.wrk_mtx);
  if (++G.num_workers_ready == G.num_workers) {
    pthread_mutex_unlock(&G.wrk_mtx);
    pthread_cond_broadcast(&G.wrk_cond);
  }
  else {
    pthread_cond_wait(&G.wrk_cond, &G.wrk_mtx);
    pthread_mutex_unlock(&G.wrk_mtx);
  }
#else
  WaitForSingleObject(G.wrk_mtx, INFINITE);
  if (++G.num_workers_ready == G.num_workers) {
    ReleaseMutex(G.wrk_mtx);
    ReleaseSemaphore(G.wrk_sem, G.num_workers, NULL);
  }
  else {
    ReleaseMutex(G.wrk_mtx);
    WaitForSingleObject(G.wrk_sem, INFINITE);
  }
#endif

  /* now we can set our callback function */
  pcap_loop(handle, -1, got_packet, (unsigned char *) wrk);

done_thread:
  DBG1("[%d] exiting\n", worker_id);
  return 0;

}/*end work_thread()*/

int
C_API
main(int argc, char **argv)
{
  int rc = 0;
  char c;
  int32_t app_id = -1;
  int w;
  worker_t *wrk;
#ifndef _WIN32
  void *thread_rc;
#endif
  char buf[64];

  /* set defaults */
  G.device = NULL;
  G.snap_len = 0;
  G.num_workers = 0;
  G.recv_wait = 500;
  G.num_pkts_expected = -1;
  G.duration = 0;
  G.delay_nsecs = 0;
  G.verbose = 0;
  G.cpumask = ~0ULL;
  G.toggle = 0;

  /* get args */
  while ((c = getopt(argc, argv, "vn:td:w:D:W:hA:S:T:")) != -1) {
    if (c == 'A') {
      app_id = strtoul(optarg, NULL, 0);
    } else if (c == 'n') {
      G.num_pkts_expected = strtoul(optarg, NULL, 0);
    } else if (c == 't') {
      G.periodic_stats = 1;
    } else if (c == 'v') {
      G.verbose++;
    } else if (c == 'S') {
      G.snap_len = strtoul(optarg, NULL, 0);
    } else if (c == 'T') {
      G.duration = strtoul(optarg, NULL, 0);
    } else if (c == 'd') {
      G.device = optarg;
    } else if (c == 'D') {
      G.delay_nsecs = strtoll(optarg, NULL, 0);
    } else if (c == 'w') {
      long long cpumask;
      int n;
      if ((n = sscanf(optarg, "%u:%llx", &G.num_workers, &cpumask)) < 1)
        usage(argv[0]);
      if (G.num_workers > MAX_WORKERS) {
        printf("Invalid number of worker threads (%d, max:%d)\n",
            G.num_workers, MAX_WORKERS);
        exit(1);
      }
      if (n >= 2)
        G.cpumask = cpumask;
    } else if (c == 'W') {
      G.recv_wait = (int)strtoull(optarg, NULL, 0);
    } else {
      if (c != 'h')
        printf("ERROR: Unknown option: %c\n", c);
      usage(argv[0]);
    }
  }

  if (!G.num_workers) {
    printf("\nERROR: Number of workers/rings not specified with -w option:\n\n");
    usage(argv[0]);
  }

  snprintf(buf, 64, "SNF_NUM_RINGS=%d", G.num_workers);
  putenv(buf);


  if (app_id != -1){
    snprintf(buf, 64, "SNF_APP_ID=%d", app_id);
    putenv(buf);
  }

#ifdef _WIN32
  if (G.cpumask) {
    SetThreadAffinityMask(GetCurrentThread(), (DWORD_PTR)G.cpumask);
  }
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  if (G.cpumask) {
    cpu_set_t cpuset;
    int i;
    CPU_ZERO(&cpuset);
    for (i = 0; i < sizeof(G.cpumask) * 8; i++) {
      if (G.cpumask & (1<<i))
        CPU_SET(i, &cpuset);
    }
    pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
  }
#endif

#ifndef _WIN32
  if (SIG_ERR == signal(SIGINT, sigexit))
    exit(1);
  if (SIG_ERR == signal(SIGTERM, sigexit))
    exit(1);
#else
  signal(SIGINT, sigexit);
#endif

#ifndef _WIN32
  pthread_mutex_init(&G.wrk_mtx, NULL);
  pthread_cond_init(&G.wrk_cond, NULL);
#else
  G.wrk_mtx = CreateMutex(NULL, FALSE, NULL);
  G.wrk_sem = CreateSemaphore(NULL, 0, G.num_workers, NULL);
#endif
  G.num_workers_ready = 0;

  printf("Capture using %d rings and worker threads with default RSS hashing", G.num_workers);
  if (G.delay_nsecs)
    printf(", each thread spends an %d nanosecs of additional (synthetic) processing\n", 
        (int) G.delay_nsecs);
  else
    printf("\n");

  for (w = 0; w < (int)G.num_workers; w++) {
    wrk = &G.worker[w];
    memset(wrk, 0, sizeof(worker_t));
    wrk->worker_id = w;
    wrk->cpu_id = cpu_getmask_ith((unsigned)G.cpumask, w);

#ifndef _WIN32
    rc = pthread_create(&wrk->thread_id, NULL, work_thread, (void *)wrk);
    if (rc != 0) {
      perror("pthread_create");
      exit(1);
    }
#else
    wrk->thread_id = CreateThread(NULL, 0, work_thread, wrk, 0, NULL);
    if (wrk->thread_id == NULL) {
      fprintf(stderr, "CreateThread\n");
      exit(1);
    }
#endif
  }

#ifndef _WIN32
  if (SIG_ERR == signal(SIGALRM, sigalrm)) {
    exit(1);
  }
#else
  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE *) stats_thread, NULL, 0, NULL);
#endif
  gettimeofday(&G.tv_last, NULL);
  G.usecs_last = TV_TO_US(&G.tv_last);
#ifndef _WIN32
  alarm(1);
#endif


#ifndef _WIN32
  for (w = 0; w < G.num_workers; w++)
    pthread_join(G.worker[w].thread_id, &thread_rc);
#else
  for (w = 0; w < (int)G.num_workers; w++)
    WaitForSingleObject(G.worker[w].thread_id, INFINITE);
#endif

  // show the stats before closing pcap handles
  sigexit(-1);

  // Do an orderly close for histogram display
  for (w = 0; w < (int)G.num_workers; w++) {
    wrk = &G.worker[w];
    if (wrk->hring) {
      pcap_close(wrk->hring);
      wrk->hring = NULL;
    }
  }


  exit(0);
}
