/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2010 by Myricom, Inc.  All rights reserved.
 * Copyright 2015 by CSP, Inc.  All rights reserved.
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifndef _WIN32
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#define __USE_GNU
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#else /* _WIN32 */
#include <winsock2.h>
#include <windows.h>
#include <time.h>
#include "getopt.h"
#include "snf_win.h"
#endif

#ifndef _WIN32
#ifdef __FreeBSD__
#include <sys/param.h>
#if __FreeBSD_version < 702100
#define PTHREAD_SETAFFINITY_NP_MISSING
#else
#include <pthread_np.h>
#define cpu_set_t cpuset_t
#endif /* __FreeBSD_version */
#elif defined(__GLIBC__)
#if !__GLIBC_PREREQ(2,4)
#define PTHREAD_SETAFFINITY_NP_MISSING
#endif
#endif
#endif

#include "snf.h"
#ifdef SNF_TREE
#include "mal_timing.h"
#else
#include "mal_simplified.h"
#endif

#define SNF_Safe(x) do {      \
  int ret = (x);              \
  if (ret != 0) {                                               \
    fprintf(stderr, "SNF Failure at line %d: %s (errno=%d)\n",  \
        __LINE__, strerror(ret), ret);                          \
    exit(1); \
  } } while (0)

#ifdef _WIN32
#define C_API __cdecl
#else
#define C_API
#endif

#define MAX_PKT_LEN 9000
#define MAX_THREADS 8
#define MAX_CONFIGS 2

#define PKT_BUF_LEN (16 * 1024)

#define DEF_PKT_LEN 60
#define DEF_SRC_PORT 3333
#define DEF_DST_PORT 3333

#define RAND_PKT_LEN    0x01
#define STEP_PKT_LEN    0x02
#define RAND_SRC_PORT   0x04
#define STEP_SRC_PORT   0x08
#define RAND_DST_PORT   0x10
#define STEP_DST_PORT   0x20

#define RND_TBL_SZ (1024 * 64)
uint32_t rand_tbl[MAX_THREADS][RND_TBL_SZ];

static inline
uint32_t myrand(uint32_t worker_id, uint32_t *idx, uint32_t mod)
{
  *idx = (*idx + 1) & (RND_TBL_SZ - 1);
  return rand_tbl[worker_id][*idx] % mod;
}

struct thread_info {
  int portnum;
  unsigned cpumask;
};

static int verbose = 0;
static struct thread_info tinfo[MAX_THREADS];
static uint32_t max_packets = 0;
static enum { NONE = 0, RATE_BW = 1, RATE_PPS = 2 } rate_limit = NONE;
static double rate_f = .0;
static uint32_t timeout = 0;

static int flags = 0;
static unsigned int src_mac[6];
static unsigned int dst_mac[6];
static unsigned int src_ip[4];
static unsigned int dst_ip[4];
static uint32_t pkt_len_min = DEF_PKT_LEN;
static uint32_t pkt_len_max = DEF_PKT_LEN;
static uint32_t src_port_min = DEF_SRC_PORT;
static uint32_t src_port_max = DEF_SRC_PORT;
static uint32_t dst_port_min = DEF_DST_PORT;
static uint32_t dst_port_max = DEF_DST_PORT;

/* Print usage statement and exit */
void
usage(void)
{
  printf("Usage: snf_pktgen [options]\n");
  printf("                       -v: Print verbose output\n");
  printf("              -p <config>: Configure threads (max: %d), where\n", MAX_THREADS);
  printf("                           <config> = <board>[:<nthreads>[:<cpumask>]]\n");
  printf("                           (default: 0:1)\n");
  printf("               -n <count>: Number of packets to send per thread (default: infinite)\n");
  printf("                 -r <x.y>: Replay at x.y Gbits/s (per thread, not aggregate)\n");
  printf("                 -R <x.y>: Replay at x.y Mpps (per thread, not aggregate)\n");
  printf("                       -W: Wait instead of busy-poll when injecting packets\n");
  printf("   -E <src mac>-<dst mac>: Source-destination MAC address\n");
  printf("     -I <src ip>-<dst ip>: Source-destination IP address\n");
  printf("                -s <size>: Packet size (default: %d)\n", DEF_PKT_LEN);
  printf("            -S <src port>: Source port\n");
  printf("            -D <dst port>: Destination port\n");
  printf("                           (-s/S/D only) For rand/seq ranges, specify\n");
  printf("                           <min value>-<max value>:{R,S}\n");
  exit(1);
}

/* Get the CPU in sequence for the given worker, taking into account its CPU mask */
static int
cpu_getmask_ith(int num_cpus, unsigned cpumask, int worker_id)
{
  int avail_cpus[MAX_THREADS];
  int i, avail_cnt = 0;
  unsigned mask;

  for (i = 0; i < num_cpus; i++) {
    mask = (1U<<i);
    if (mask & cpumask) {
      avail_cpus[avail_cnt++] = i;
      if (avail_cnt > MAX_THREADS)
        break;
    }
  }

  if (avail_cnt > 0) {
    int idx = worker_id % avail_cnt;
    return avail_cpus[idx];
  }

  /* Return any CPU */
  return -1;
}

/* Set thread CPU affinity to distribute work among cores */
static
void
set_cpu_affinity(int worker_id)
{
#ifdef _WIN32
  SYSTEM_INFO info;
  DWORD cpu;

  GetSystemInfo(&info);
  cpu = cpu_getmask_ith(info.dwNumberOfProcessors, tinfo[worker_id].cpumask, worker_id);
  if (cpu >= 0) {
    SetThreadAffinityMask(GetCurrentThread(), cpu);
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  cpu_set_t cpuset;
  int cpu;

  cpu = cpu_getmask_ith(8, tinfo[worker_id].cpumask, worker_id);
  if (cpu >= 0) {
    CPU_ZERO(&cpuset);
    CPU_SET(cpu, &cpuset);
    pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
#endif

    if (verbose)
      printf("[%d] Pinned to CPU %d\n", worker_id, (int) cpu);
  }
}

/* Calculate the header checksum */
static
uint16_t
checksum(uint16_t *ptr, int len)
{
  uint32_t sum = 0;

  while (len > 1) {
    sum += *ptr++;
    len -= 2;
  }

  if (len)
    sum += *(uint8_t *)ptr;

  while (sum >> 16)
    sum = (sum & 0xffff) + (sum >> 16);

  return ~sum;
}

/* Work thread for setup, timing and sending */
#ifndef _WIN32
void *
#else
DWORD WINAPI
#endif
work_thread(void *arg)
{
  int worker_id = (int)(uintptr_t)arg;
  int rc, i;
  int portnum;
  snf_inject_t hinj;
  uint64_t pkts_cnt = 0, bytes_cnt = 0;
  uint64_t t_start, t_elapsed;
  double rate_f_delta;
  mal_cycles_t t_cyc_start, cpc;
  unsigned char pkt[PKT_BUF_LEN] = {
      /* eth */
      0x00, 0x60, 0xdd, 0x45, 0x4c, 0xec, /* dst mac */
      0x00, 0x60, 0xdd, 0x45, 0x4c, 0xee, /* src mac */
      0x08, 0x00, /* type */

      /* ip */
      0x45, 0x00, /* misc */
      0x00, 0x1c, /* total len */
      0x00, 0x00, /* id */
      0x40, 0x00, /* flags and offset */
      0x40, 0x11, /* ttl and proto */
      0x00, 0x00, /* hdr cksum */
      0x0a, 0x00, 0x00, 0x02, /* src ip */
      0x0a, 0x00, 0x00, 0x01, /* dst ip */

      /* udp */
      0x0d, 0x05, /* src port */
      0x0d, 0x05, /* dst port */
      0x00, 0x08, /* len */
      0x00, 0x00, /* cksum */
  };
  uint8_t *mac_ptr = (uint8_t *) &pkt[0];
  uint16_t *ip_hdr_ptr = (uint16_t *) &pkt[14];
  uint16_t *ip_len_ptr = (uint16_t *) &pkt[16];
  uint16_t *ip_cksum_ptr = (uint16_t *) &pkt[24];
  uint8_t *ip_addr_ptr = (uint8_t *) &pkt[26];
  uint16_t *src_port_ptr = (uint16_t *) &pkt[34];
  uint16_t *dst_port_ptr = (uint16_t *) &pkt[36];
  uint16_t *udp_len_ptr = (uint16_t *) &pkt[38];
  uint32_t *pld_ptr = (uint32_t *) &pkt[48];
  uint32_t pkt_len, pkt_len_delta, pkt_len_incr;
  uint32_t src_port, src_port_delta, src_port_incr;
  uint32_t dst_port, dst_port_delta, dst_port_incr;
  uint32_t rnd_idx = 0;
  uint64_t speed;
#ifndef _WIN32
  sigset_t sigmask;
  /* Mask signals except SIGINT to allow ^C */
  sigfillset(&sigmask);
#if !defined __FreeBSD__
  sigdelset(&sigmask, SIGINT);
#endif
  rc = pthread_sigmask(SIG_BLOCK, &sigmask, NULL);
#endif

  set_cpu_affinity(worker_id);

  pkt_len = pkt_len_min;
  src_port = src_port_min;
  dst_port = dst_port_min;
  pkt_len_delta = pkt_len_max - pkt_len_min;
  src_port_delta = src_port_max - src_port_min;
  dst_port_delta = dst_port_max - dst_port_min;
  pkt_len_incr =  src_port_incr = dst_port_incr = 0;

  /* Build packet */
  for (i = 0; i < 6; i++) { mac_ptr[i] = (unsigned char) dst_mac[i]; }
  for (i = 0; i < 6; i++) { mac_ptr[i+6] = (unsigned char) src_mac[i]; }

  for (i = 0; i < 4; i++) { ip_addr_ptr[i] = (unsigned char) src_ip[i]; }
  for (i = 0; i < 4; i++) { ip_addr_ptr[i+4] = (unsigned char) dst_ip[i]; }
 
  for (i = 0; i < (PKT_BUF_LEN - 48) / 4 - 3; i++) {
    *pld_ptr = i;
    pld_ptr++;
  }

  portnum = tinfo[worker_id].portnum;

  /* Compute rate limit delta after the thread is pinned so as to get
   * consistent results from rdtsc */
  mal_cycles_counter_init();
  cpc = mal_cycles_per_second();
  if (rate_limit) {
    t_cyc_start = mal_get_cycles();
    if (rate_limit == RATE_BW)
      rate_f_delta = cpc / (rate_f * 1.0e9 / 8.0);
    else /* rate_limit == RATE_PPS */
      rate_f_delta = cpc / (rate_f * 1.0e6);
  } else {
    rate_f_delta = .0;
    t_cyc_start = 0;
  }

  rc = snf_inject_open(portnum, 0, &hinj);
  if (rc) {
    fprintf(stderr, "[%d] Cannot inject on port %d: %s\n", 
        worker_id, portnum, strerror(rc));
    goto done_thread;
  }

  snf_get_injection_speed(hinj, &speed);
  speed /= 1000000000;

  t_start = mal_get_cycles();

  /* Start sending */
  do {
    if (flags) {
      if (flags & RAND_PKT_LEN) {
        pkt_len = pkt_len_min + myrand(worker_id, &rnd_idx, pkt_len_delta + 1);
      } else if (flags & STEP_PKT_LEN) {
        pkt_len = pkt_len_min + pkt_len_incr;
        pkt_len_incr = (pkt_len_incr >= pkt_len_delta) ? 0 : pkt_len_incr + 1;
      }

      if (flags & RAND_SRC_PORT) {
        src_port = src_port_min + myrand(worker_id, &rnd_idx, src_port_delta + 1);
      } else if (flags & STEP_SRC_PORT) {
        src_port = src_port_min + src_port_incr;
        src_port_incr = (src_port_incr >= src_port_delta) ? 0 : src_port_incr + 1;
      }

      if (flags & RAND_DST_PORT) {
        dst_port = dst_port_min + myrand(worker_id, &rnd_idx, dst_port_delta + 1);
      } else if (flags & STEP_DST_PORT) {
        dst_port = dst_port_min + dst_port_incr;
        dst_port_incr = (dst_port_incr >= dst_port_delta) ? 0 : dst_port_incr + 1;
      }
    }

    /* Rebuild packet */
    *ip_len_ptr = htons(pkt_len - 14);
    *ip_cksum_ptr = 0; /* Reset checksum before calculation */
    *ip_cksum_ptr = checksum(ip_hdr_ptr, 20);
    *udp_len_ptr = htons(pkt_len - 14 - 20);
    *src_port_ptr = htons(src_port);
    *dst_port_ptr = htons(dst_port);

reinject:
    rc = snf_inject_send(hinj, timeout, 0, pkt, pkt_len);
    if (rc == 0) {
      pkts_cnt++;
      bytes_cnt += pkt_len;
    } else if (rc == EAGAIN || rc == EINTR) {
      goto reinject;
    } else {
      fprintf(stderr, "[%d] snf_inject_send error %d (%s)\n", 
          worker_id, rc, strerror(rc));
      break;
    }

    if (rate_limit) {
      mal_cycles_t cyc_elapsed;

      if (rate_limit == RATE_BW)
        cyc_elapsed = (mal_cycles_t)(bytes_cnt * rate_f_delta);
      else /* rate_limit == RATE_PPS */
        cyc_elapsed = (mal_cycles_t)(pkts_cnt * rate_f_delta);
      while ((mal_get_cycles() - t_cyc_start) < cyc_elapsed)
        /* Do nothing */;
    }
  } while ((max_packets == 0) || (pkts_cnt < max_packets));

  rc = snf_inject_close(hinj);

  t_elapsed = (mal_get_cycles() - t_start) * 1000000 / cpc;

  if (rc) {
    fprintf(stderr, "[%d] Cannot close inject handle %d: %s\n", 
        worker_id, worker_id, strerror(rc));
  } else {
    /* Add 24 bytes: 4 bytes CRC, 7 preamble bytes, 1 start byte, 12
     * inter-packet gap bytes */
    double pkt_len_avg = (double) bytes_cnt / (double) pkts_cnt;
    double peak_mpps = 1000.0 / ((double) (pkt_len_avg + 24) * 0.8);
    double mpps = (double) pkts_cnt / t_elapsed;
     
    printf("[%d] Size     Mbps    Mpps  Efficiency\n"
        "    %4d  %7.2f  %6.3f  %9.2f%%\n", worker_id,
        (int) pkt_len_avg, (double) bytes_cnt * 8 / t_elapsed,
        mpps, 100.0 * mpps / peak_mpps * (speed == 1 ? 10 : 1) );
  }
  
done_thread:
  return 0;
}

int
C_API
main(int argc, char **argv)
{
  char c;
  int rc, i;
  char *endptr;
  int portnum, thread_cnt;
  unsigned cpumask;
  int num_threads = 0;
  int num_configs = 0;
#ifndef _WIN32
  pthread_t thread_ids[MAX_THREADS];
  void *thread_rc;
#else
  HANDLE thread_ids[MAX_THREADS];
#endif

  /* Get command line args */
  while ((c = getopt(argc, argv, "vp:b:n:r:R:WE:I:s:S:D:")) != -1) {
    switch (c) {
    case 'v':
      verbose++;
      break;
    case 'p':
    case 'b':
      if (num_configs >= MAX_CONFIGS) {
        fprintf(stderr, "Too many thread configurations specified (max: %d)\n", MAX_CONFIGS);
        usage();
      }

      portnum = 0;
      thread_cnt = 1;
      cpumask = 0;
      rc = sscanf(optarg, "%u:%u:%x", &portnum, &thread_cnt, &cpumask);
      if (rc < 1 || portnum < 0 || thread_cnt < 1) {
        fprintf(stderr, "Invalid thread configuration\n");
        usage();
      }

      if (num_threads + thread_cnt > MAX_THREADS) {
        fprintf(stderr, "Too many threads specified (max: %d)\n", MAX_THREADS);
        usage();
      }

      for (i = num_threads; i < num_threads + thread_cnt; i++) {
        tinfo[i].portnum = portnum;
        tinfo[i].cpumask = cpumask;
      }

      num_threads += thread_cnt; 
      num_configs++;
      break;
    case 'n':
      max_packets = (uint32_t) strtoull(optarg, NULL, 0);
      break;
    case 'r':
      rate_limit = RATE_BW;
      if ((sscanf(optarg, "%lg", &rate_f) != 1) ||
          !(rate_f > 0.0 && rate_f <= 10.0)) {
        printf("Bandwidth in Gbps must be >0 and <=10.0\n");
        usage();
      }
      break;
    case 'R':
      rate_limit = RATE_PPS;
      if ((sscanf(optarg, "%lg", &rate_f) != 1) ||
          !(rate_f > 0.0 && rate_f <= 14.8)) {
        printf("Packet rate in Mpps must be >0 and <=14.8\n");
        usage();
      }
      break;
    case 'W':
      timeout = -1;
      break;
    case 'E':
      rc = sscanf(optarg, "%x:%x:%x:%x:%x:%x-%x:%x:%x:%x:%x:%x",
          &src_mac[0], &src_mac[1], &src_mac[2],
          &src_mac[3], &src_mac[4], &src_mac[5],
          &dst_mac[0], &dst_mac[1], &dst_mac[2],
          &dst_mac[3], &dst_mac[4], &dst_mac[5]);
      if (rc != 12) {
        fprintf(stderr, "Invalid MAC address\n");
        usage();
      }
      break;
    case 'I':
      rc = sscanf(optarg, "%d.%d.%d.%d-%d.%d.%d.%d",
          &src_ip[0], &src_ip[1], &src_ip[2], &src_ip[3],
          &dst_ip[0], &dst_ip[1], &dst_ip[2], &dst_ip[3]);
      if (rc != 8) {
        fprintf(stderr, "Invalid IP address\n");
        usage();
      }
      break;
    case 's':
      pkt_len_min = strtoul(optarg, &endptr, 0);
      if (*endptr == '-') {
        pkt_len_max = strtoul(endptr+1, &endptr, 0);
        if (*endptr == ':') {
          if (endptr[1] == 'R') {
            flags |= RAND_PKT_LEN;
          } else if (endptr[1] == 'S') {
            flags |= STEP_PKT_LEN;
          } else {
            fprintf(stderr, "Invalid len iterator\n\n");
            usage();
          }
        } else {
          fprintf(stderr, "Invalid len seperator\n");
          usage();
        }
        if (pkt_len_min > pkt_len_max) {
          fprintf(stderr, "Minimum packet size cannot be greater than max\n");
          usage();
        }
      } else if (*endptr == '\0') {
        pkt_len_max = pkt_len_min;
      } else {
        fprintf(stderr, "Invalid packet size\n\n");
        usage();
      }
      if (pkt_len_min < 14) {
        fprintf(stderr, "Minimum packet size must be at least 14 bytes\n");
        usage();
      }
      if (pkt_len_max > MAX_PKT_LEN) {
        fprintf(stderr, "Maximum packet size cannot exceed %d bytes\n", MAX_PKT_LEN);
        usage();
      }
      break;
    case 'S':
      src_port_min = strtoul(optarg, &endptr, 0);
      if (*endptr == '-') {
        src_port_max = strtoul(endptr+1, &endptr, 0);
        if (*endptr == ':') {
          if (endptr[1] == 'R') {
            flags |= RAND_SRC_PORT;
          } else if (endptr[1] == 'S') {
            flags |= STEP_SRC_PORT;
          } else {
            fprintf(stderr, "Invalid source port iterator\n\n");
            usage();
          }
        } else {
          fprintf(stderr, "Invalid len seperator\n");
          usage();
        }
        if (src_port_min > src_port_max) {
          fprintf(stderr, "Minimum src port cannot be greater than max\n");
          usage();
        }
      } else if (*endptr == '\0') {
        src_port_max = src_port_min;
      } else {
        fprintf(stderr, "Invalid source port size\n");
        usage();
      }
      break;
    case 'D':
      dst_port_min = strtoul(optarg, &endptr, 0);
      if (*endptr == '-') {
        dst_port_max = strtoul(endptr+1, &endptr, 0);
        if (*endptr == ':') {
          if (endptr[1] == 'R') {
            flags |= RAND_DST_PORT;
          } else if (endptr[1] == 'S') {
            flags |= STEP_DST_PORT;
          } else {
            fprintf(stderr, "Invalid source port iterator\n\n");
            usage();
          }
        } else {
          fprintf(stderr, "Invalid len seperator\n");
          usage();
        }
        if (dst_port_min > dst_port_max) {
          fprintf(stderr, "Minimum dst port cannot be greater than max\n");
          usage();
        }
      } else if (*endptr == '\0') {
        dst_port_max = dst_port_min;
      } else {
        fprintf(stderr, "Invalid source port size\n");
        usage();
      }
      break;
    default:
      usage();
    }
  }

  /* Create at least one thread */
  if (num_threads == 0)
    num_threads = 1;

  /* Use a table of random values so rand() is not called during the send loop. */
  if (flags & (RAND_PKT_LEN | RAND_SRC_PORT | RAND_DST_PORT)) {
    unsigned int now = (unsigned int) time(NULL);

    for (i = 0; i < (int)num_threads; i++) {
      unsigned int seed;
      int j;

      seed = now + i;
      srand(seed);
      if (verbose)
        printf("[%d] seed %u\n", i, seed);

      for (j = 0; j < RND_TBL_SZ; j++) {
        rand_tbl[i][j] = (uint32_t)rand();
      }
    }
  }

  snf_init(SNF_VERSION_API);

  for (i = 0; i < num_threads; i++) {
#ifndef _WIN32
    rc = pthread_create(&thread_ids[i], NULL, work_thread, (void *)(uintptr_t) i);
    if (rc != 0) {
      perror("pthread_create");
      exit(1);
    }
#else
    thread_ids[i] = CreateThread(NULL, 0, work_thread, (void*)(uintptr_t) i, 0, NULL);
    if (thread_ids[i] == NULL) {
      fprintf(stderr, "CreateThread: %d\n", GetLastError());
      exit(1);
    }
#endif
  }

  for (i = 0; i < num_threads; i++) {
#ifndef _WIN32
    pthread_join(thread_ids[i], &thread_rc);
#else
    WaitForSingleObject(thread_ids[i], INFINITE);
#endif
  }

  return 0;
}
