/*************************************************************************
 * The contents of this file are subject to the MYRICOM SNIFFER10G
 * LICENSE (the "License"); User may not use this file except in
 * compliance with the License.  The full text of the License can found
 * in LICENSE.TXT
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Copyright 2010 by Myricom, Inc.  All rights reserved.
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifndef _WIN32
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#define __USE_GNU
#include <pthread.h>
#include <signal.h>
#else /* _WIN32 */
#include <winsock2.h>
#include <windows.h>
#include "getopt.h"
#include "snf_win.h"
#endif

#ifndef _WIN32
#ifdef __FreeBSD__
#include <sys/param.h>
#if __FreeBSD_version < 702100
#define PTHREAD_SETAFFINITY_NP_MISSING
#else
#include <pthread_np.h>
#define cpu_set_t cpuset_t
#endif /* __FreeBSD_version */
#elif defined(__GLIBC__)
#if !__GLIBC_PREREQ(2,4)
#define PTHREAD_SETAFFINITY_NP_MISSING
#endif
#endif
#endif

#include "pcap.h"
#include "snf.h"
#if SNF_TREE
#include "queue.h"
#include "mal_timing.h"
#else
#include <sys/queue.h>
#include "mal_simplified.h"
#endif

#define SNF_Safe(x) do {      \
  int ret = (x);              \
  if (ret != 0) {                                               \
    fprintf(stderr, "SNF Failure at line %d: %s (errno=%d)\n",  \
        __LINE__, strerror(ret), ret);                          \
    exit(1); \
  } } while (0)

#define MAX_THREADS 16

static int verbose = 0;
static int portnum = 0;
static uint64_t max_packets = -1;
static int num_threads = 1;
static int iters = 1;
static uint32_t vlan_tag_be32 = 0;
static enum { NONE = 0, RATE_BW = 1, RATE_PPS = 2 } rate_limit = NONE;
static double rate_f = .0;
static int opt_affinity = 1;
static int replay_timestamps = 0;
static int bin_size_us = 5;
static int pcap_in_memory = 0;
static const char *cap_file = NULL;
#ifdef _WIN32
static CRITICAL_SECTION gPrintCS;
#endif

struct inj_thread_info {
  uint32_t   id;
  uint64_t   pkts_cnt;
  uint64_t   vlan_tag_inserts;
  uint64_t   bytes_cnt;
  uint64_t   t_elapsed_us;
  uint64_t   nerror;
  uint64_t   t_cyc_start;
  uint64_t   cpc;
  double     rate_f_delta;
#ifndef _WIN32
  uint8_t    pad[128-sizeof(uint32_t)-8*sizeof(uint64_t)-sizeof(double)];
#else
  HANDLE     hEvent;
  HANDLE     hThread; 
  uint8_t    pad[128-sizeof(uint32_t)-8*sizeof(uint64_t)-sizeof(double)-2*sizeof(HANDLE)];
#endif
} inj_info[MAX_THREADS];

struct pcap_im_elem {
  uint8_t *data;
  struct pcap_pkthdr hdr;
  STAILQ_ENTRY(pcap_im_elem) next;
};

STAILQ_HEAD( , pcap_im_elem) pcap_im_head;

struct pcap_im_info {
  struct pcap_im_elem *cur_elem;
};

/* Print usage statement and exit */
void
usage(void)
{
  printf("Usage: snf_replay [options] <file.pcap>\n");
  printf("              -v: Print verbose output\n");
  printf("       -p <port>: Myri10G port number to use\n");
  printf("      -n <count>: Maximum number of packets to send from pcap file\n");
  printf("   -t <nthreads>: Number of threads to concurrently send (max: %d)\n", MAX_THREADS);
  printf("      -i <iters>: Number of times to replay the file on each thread\n");
  printf("    -T <vlan_id>: Insert a vlan tag for packets without VLAN tags\n");
  printf("        -r <x.y>: Replay at x.y Gbits/s (per thread, not aggregate)\n");
  printf("        -R <x.y>: Replay at x.y Mpps (per thread, not aggregate)\n");
  printf("              -z: Replay according to packet timestamp\n");
  printf("              -F: Disable setting thread affinity\n");
  printf("              -m: Read pcap file into memory first\n");
  printf("                  Note - Make sure your pcap file fits in available memory\n");
  exit(1);
}

/* Read pcap file and load packets into memory to avoid extra delay between sends */
void read_pcap_into_mem(void) {
  char errbuf[PCAP_ERRBUF_SIZE];
  pcap_t *p;
  struct pcap_pkthdr pkthdr;
  const u_char *pdata;
  struct pcap_im_elem *pie;
  int count = 0;

  if (!(p = pcap_open_offline(cap_file, errbuf))) {
    fprintf(stderr, "Error opening pcap file: %s\n", errbuf);
    exit(1);
  }

  /* Allocate a dummy first entry so that when next packet is called
     we get the true first entry */
  pie = malloc(sizeof(struct pcap_im_elem));
  if (pie == NULL) {
    fprintf(stderr, "Error allocating memory for pcap packet list entry.\n");
    exit(1);
  }
  pie->data = NULL;
  pie->hdr.caplen = 0;
  pie->hdr.len = 0;
  STAILQ_INSERT_HEAD(&pcap_im_head, pie, next);
  
  printf("Reading packets into memory\n");
  while((pdata = pcap_next(p, &pkthdr)) != NULL) {
    pie = malloc(sizeof(struct pcap_im_elem));
    if (pie == NULL) {
      fprintf(stderr, "Error allocating memory for pcap packet list entry.\n");
      exit(1);
    }
    pie->data = malloc(pkthdr.caplen);
    if (pie->data == NULL) {
      fprintf(stderr, "Error allocating memory for pcap packet.\n");
      exit(1);
    }
    memcpy(pie->data, pdata, pkthdr.caplen);
    pie->hdr.ts = pkthdr.ts;
    pie->hdr.caplen = pkthdr.caplen;
    pie->hdr.len = pkthdr.len;
    STAILQ_INSERT_TAIL(&pcap_im_head, pie, next);
    count++;
  }
  printf("Read %d packets into memory\n", count);
  pcap_close(p);
}

/* Initialize pcap source */
int start_packets(void **p, char *pbuf) {
  char errbuf[PCAP_ERRBUF_SIZE];

  if (pcap_in_memory) {
    (*(struct pcap_im_info **)p)->cur_elem = STAILQ_FIRST(&pcap_im_head);
  } else {
    *p = (void *)pcap_open_offline(cap_file, errbuf);
    if (*p == NULL) {
      fprintf(stderr, "%s: Error opening pcap file: %s\n", pbuf, errbuf);
      return -1;
    }
  }
  return 0;
}

/* Close pcap file if necessary */
void finish_packets(void *p) {
  if (!pcap_in_memory) {
    pcap_close(p);
  }
}

/* Get the next packet and its header from the queue or the pcap file */
const u_char *next_packet(void *p, struct pcap_pkthdr *h, int copy)
{
  uint8_t *pdata_cp;
  const u_char *pdata;

  if (pcap_in_memory) {
    struct pcap_im_elem *pie;

    pie = STAILQ_NEXT(((struct pcap_im_info *)p)->cur_elem, next);
    if (pie == NULL) {
      return NULL;
    }
    h->ts = pie->hdr.ts;
    h->caplen = pie->hdr.caplen;
    h->len = pie->hdr.len;
    ((struct pcap_im_info *)p)->cur_elem = pie;
    pdata = pie->data;
  } else {
    pdata = pcap_next((pcap_t *)p, h);
    if (copy && (pdata != NULL)) {
      /* make a copy of this packet data */
      pdata_cp = malloc(h->caplen);
      if (pdata_cp == NULL) {
        fprintf(stderr, "Error allocating memory for pcap packet.\n");
        exit(1);
      }
      memcpy(pdata_cp, pdata, h->caplen);
      pdata = pdata_cp;
    }
  }  

  return pdata;
}

/* Print thread-specific statistics */
void
print_tinfo(struct inj_thread_info *tinfo, const char *prefix)
{
  uint64_t t_usec;
#ifndef _WIN32
  static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock(&mtx);
#else
  EnterCriticalSection(&gPrintCS);
#endif
  t_usec = tinfo->t_elapsed_us;
  printf("%s> Packets: %lld\n", prefix, (long long) tinfo->pkts_cnt);
  if (vlan_tag_be32)
    printf("%s> Packets with vlan insertions: %lld\n", prefix,
        (long long) tinfo->vlan_tag_inserts);
  printf("%s> Bytes: %lld\n", prefix, (long long) tinfo->bytes_cnt);
  printf("%s> Rate: %.2f Mpps\n", prefix, (double) (tinfo->pkts_cnt) / t_usec);
  printf("%s> Throughput: %.3f Gbps in %.3f secs\n", prefix,
      (double) (tinfo->bytes_cnt * 8 / 1e9) / (t_usec / 1e6),
      (double) t_usec / 1e6);
  if (tinfo->nerror)
    printf("%s> Inject errors: %lld\n", prefix, (long long) tinfo->nerror);
#ifndef _WIN32
  pthread_mutex_unlock(&mtx);
#else
  LeaveCriticalSection(&gPrintCS);
#endif
}

/* Assemble the packet if necessary and inject. 
 * Record successful injections in thread statistics and return completion status */
static inline
int
inject(snf_inject_t hinj, struct inj_thread_info *ij, const u_char *pdata, uint32_t length)
{
  int rc;
  struct snf_pkt_fragment vec[3];

  /* If we want vlan and there's not already a vlan tag in the packet,
   * insert it inline */
  if (vlan_tag_be32 && length >= 16 &&
      *(uint16_t *)(pdata + 12) != htons(0x8100)) 
  {
    vec[0].ptr = (void *) pdata;
    vec[0].length = 12; /* dest and src mac */
    vec[1].ptr = &vlan_tag_be32;
    vec[1].length = sizeof(vlan_tag_be32);
    vec[2].ptr = (void *)(pdata + 12);
    vec[2].length = length - 12;
    length += sizeof(vlan_tag_be32);
    ij->vlan_tag_inserts++;
    rc = snf_inject_send_v(hinj, 1000, 0, vec, 3, length);
  }
  else
    rc = snf_inject_send(hinj, 1000, 0, (void *) pdata, length);

  if (rc == 0) {
    ij->pkts_cnt++;
    ij->bytes_cnt += length;
  }
  else if (rc == EINVAL) {
    printf("snf_inject_packet skipping invalid packet, error: %s\n", strerror(rc));
    ij->nerror++;
  }
  else if (rc != EAGAIN) {
    printf("snf_inject_packet unhandled error: %s\n", strerror(rc));
    return 1;
  }
  else
    ij->nerror++;

  if (max_packets != -1 && ((int) ij->pkts_cnt == max_packets)) {
    return 1;
  }

  return 0;
}

/* Send all packets in pcap file. If rate limiting is set, spin until time to send */
int 
send_pkts(snf_inject_t hinj, struct inj_thread_info *ij, void *p)
{
  struct pcap_pkthdr pkthdr;
  const u_char *pdata;
  int rc;

  while ((pdata = next_packet(p, &pkthdr, 0)) != NULL) { 
    rc = inject(hinj, ij, pdata, pkthdr.caplen);
    if (rc != 0)
      return 1;
  
    if (rate_limit) {
      mal_cycles_t cyc_elapsed;
  
      if (rate_limit == RATE_BW)
        cyc_elapsed = (mal_cycles_t)(ij->bytes_cnt * ij->rate_f_delta);
      else /* rate_limit == RATE_PPS */
        cyc_elapsed = (mal_cycles_t)(ij->pkts_cnt * ij->rate_f_delta);

      while ((mal_get_cycles() - ij->t_cyc_start) < cyc_elapsed)
        /* Do nothing */;
    }
  }

  return 0;
}

/* Send all packets in pcap file according to their timestamps.
 * Batch packets into time bins of a fixed width and inject all packets
 * belonging to a bin. Process packets for the next bin and spin until
 * bin time is reached */
int 
send_pkts_ts(snf_inject_t hinj, struct inj_thread_info *ij, void *p)
{
  struct pcap_pkthdr pkthdr;
  const u_char *pdata;
  int rc, i;
  mal_cycles_t cyc_bin = 0;
  const u_char **pdata_bin;
  struct pcap_pkthdr *pkthdr_bin;
  struct timeval tv = {0, 0};
  int bin_num = 1;
  int first = 1;
  int pkt_cnt = 0;
  int copy = 1;

  pdata = NULL;
  memset(&pkthdr, 0, sizeof(struct pcap_pkthdr));
  if (pcap_in_memory)
    copy = 0;

  /* Packets per bin is limited by max rate of ~14.8 packets per usec */
  pdata_bin = (const u_char **) malloc(sizeof(const u_char *) * bin_size_us * 15);
  pkthdr_bin =
      (struct pcap_pkthdr *) malloc(sizeof(struct pcap_pkthdr) * bin_size_us * 15);

  /* All packet timestamps are relative to the first */
  pdata_bin[0] = next_packet(p, &pkthdr_bin[0], copy);
  if (pdata_bin[0] == NULL) {
    free((u_char **)pdata_bin);
    free(pkthdr_bin);
    return 1;
  }
  tv.tv_sec = pkthdr_bin[0].ts.tv_sec;
  tv.tv_usec = pkthdr_bin[0].ts.tv_usec;
  pkt_cnt = 1;

  while (1) {

    /* If this bin is non-empty, add packets until one from another bin is found */
    if (pkt_cnt == 1) {
      while ((pdata_bin[pkt_cnt] = next_packet(p, &pkthdr_bin[pkt_cnt], copy)) != NULL) { 
        long diff_us = (pkthdr_bin[pkt_cnt].ts.tv_sec - tv.tv_sec) * 1000000
            + (pkthdr_bin[pkt_cnt].ts.tv_usec - tv.tv_usec);
        if (diff_us < bin_num * bin_size_us) {
          pkt_cnt++;
        } else {
          /* Save this packet for a future bin */
          pdata = pdata_bin[pkt_cnt];
          pkthdr.ts = pkthdr_bin[pkt_cnt].ts;
          pkthdr.len = pkthdr_bin[pkt_cnt].len;
          pkthdr.caplen = pkthdr_bin[pkt_cnt].caplen;
          break;
        }
      }
    }

    /* Spin for next bin (do nothing for first bin) */
    cyc_bin = (bin_num - 1) * bin_size_us * ij->cpc / 1000000;
    while (mal_get_cycles() - ij->t_cyc_start < cyc_bin)
      /* Do nothing */;

    if (first) {
      ij->t_cyc_start = mal_get_cycles();
      first = 0;
    }

    for (i = 0; i < pkt_cnt; i++) {
      rc = inject(hinj, ij, pdata_bin[i], pkthdr_bin[i].caplen);
      if (rc != 0) {
        free((u_char **)pdata_bin);
        free(pkthdr_bin);
        return 1;
      }
      if (copy) {
        free((u_char *) pdata_bin[i]);
      }
    }

    pkt_cnt = 0;
    bin_num++;

    /* Check if previously saved packet belongs in new bin */
    if (pdata != NULL) {
      long diff_us = (pkthdr.ts.tv_sec - tv.tv_sec) * 1000000
          + (pkthdr.ts.tv_usec - tv.tv_usec);
      if (diff_us < bin_num * bin_size_us) {
        pdata_bin[0] = pdata;
        pkthdr_bin[0].ts = pkthdr.ts;
        pkthdr_bin[0].len = pkthdr.len;
        pkthdr_bin[0].caplen = pkthdr.caplen;
        pdata = NULL;
        pkt_cnt = 1;
      }
    } else {
      /* No saved packet means no more packets in file */
      break;
    }
  }

  free((u_char **)pdata_bin);
  free(pkthdr_bin);
  return 0;
}

/* Set thread CPU affinity to distribute work among cores */
static
void
set_cpu_affinity(int worker_id)
{
#ifdef _WIN32
  SYSTEM_INFO info;
  DWORD cpu;

  GetSystemInfo(&info);
  cpu = worker_id % info.dwNumberOfProcessors;
  SetThreadIdealProcessor(GetCurrentThread(), cpu);
#elif !defined(PTHREAD_SETAFFINITY_NP_MISSING)
  cpu_set_t cpuset;
  unsigned cpu = worker_id % 8;

  CPU_ZERO(&cpuset);
  CPU_SET(cpu, &cpuset);
  pthread_setaffinity_np(pthread_self(), sizeof(cpuset), (void *)&cpuset);
#endif

  if (verbose)
    printf("[%d] Pinned to CPU %d\n", worker_id, (int) cpu);
}

/* Work thread for setup, timing and sending */
#ifndef _WIN32
void *
work_thread(void *arg)
#else
LPVOID WINAPI
work_thread(LPVOID arg)
#endif
{
  struct inj_thread_info *ij = (struct inj_thread_info *)arg;
  void *p = NULL;
  char pbuf[64];
  int i, done;
  snf_inject_t hinj;
  mal_cycles_t test_start;

  snprintf(pbuf, sizeof(pbuf) - 1, "Thread %d", ij->id);

  if (opt_affinity)
    set_cpu_affinity(ij->id);

  if (pcap_in_memory)
    p = malloc(sizeof(struct pcap_im_info));

  /* Compute rate limit delta after the thread is pinned so as to get
   * consistent results from rdtsc */
  mal_cycles_counter_init();
  ij->cpc = mal_cycles_per_second();
  if (rate_limit) {
    ij->t_cyc_start = mal_get_cycles();
    if (rate_limit == RATE_BW)
      ij->rate_f_delta = ij->cpc / (rate_f * 1.0e9 / 8.0);
    else /* rate_limit == RATE_PPS */
      ij->rate_f_delta = ij->cpc / (rate_f * 1.0e6);
  }

  SNF_Safe( snf_inject_open(portnum, 0, &hinj) );

  test_start = mal_get_cycles();

  for (i = 0; i < iters; i++) {
    
    if (start_packets(&p, pbuf) != 0) {
      return NULL;
    }

    if (replay_timestamps) {
      done = send_pkts_ts(hinj, ij, p);
    } else {
      done = send_pkts(hinj, ij, p);
    }
      
    finish_packets(p);

    if (done)
      break;
  }

  SNF_Safe( snf_inject_close(hinj) );

  ij->t_elapsed_us = (mal_get_cycles() - test_start) * 1000000 / ij->cpc;

  print_tinfo(ij, pbuf);

#ifdef _WIN32
  SetEvent(ij->hEvent);
#endif

  return ij;
}

int 
#ifdef _WIN32
__cdecl
#endif
main(int argc, char **argv)
{
  int rc, i;
  char c;
  int tag;
#ifndef _WIN32
  pthread_t thread_ids[MAX_THREADS];
  void *thread_rc;
#endif
  struct inj_thread_info tinfo;

  /* Get command line args */
  while ((c = getopt(argc, argv, "vp:b:n:t:i:T:r:R:Fzm")) != -1) {
    switch (c) {
    case 'v':
      verbose++;
      break;
    case 'b':
    case 'p':
      portnum = strtol(optarg, NULL, 0);
      break;
    case 'n':
      max_packets = (uint64_t) strtoull(optarg, NULL, 0);
      break;
    case 't':
      num_threads = (int) strtol(optarg, NULL, 0);
      if (num_threads > MAX_THREADS) {
        fprintf(stderr, "Invalid number of threads (%d, max:%d)\n",
            num_threads, MAX_THREADS);
        exit(1);
      }
      break;
    case 'i':
      iters = (int) strtol(optarg, NULL, 0);
      break;
    case 'T':
      tag = (int) strtol(optarg, NULL, 0);
      if (!(tag > 0 && tag < 4095)) {
        printf("Invalid VLAN tag for this test: %d\n", tag);
        usage();
      }
      vlan_tag_be32 = htonl(0x8100 << 16 | (tag & 0xfff));
      printf("Sending with vlan tag %d\n", tag);
      break;
    case 'r':
      rate_limit = RATE_BW;
      if ((sscanf(optarg, "%lg", &rate_f) != 1) ||
          !(rate_f > 0.0 && rate_f <= 10.0)) {
        printf("Bandwidth in Gbps must be >0 and <=10.0\n");
        usage();
      }
      break;
    case 'R':
      rate_limit = RATE_PPS;
      if ((sscanf(optarg, "%lg", &rate_f) != 1) ||
          !(rate_f > 0.0 && rate_f <= 14.8)) {
        printf("Packet rate in Mpps must be >0 and <=14.8\n");
        usage();
      }
      break;
    case 'F':
      opt_affinity = 0;
      break;
    case 'z':
      replay_timestamps = 1;
      break;
    case 'm':
      pcap_in_memory = 1;
      break;
    default:
      usage();
    }
  }

  if (argc - optind < 1)
    usage();

  if (replay_timestamps) {
    if (num_threads > 1) {
      printf("Only one thread can be run with timestamp replay\n");
      usage();
    }

    if (rate_limit) {
      printf("Timestamp replay conflicts with rate limit\n");
      usage();
    }
  }

  cap_file = argv[optind];

  if (pcap_in_memory) {
    STAILQ_INIT(&pcap_im_head);
    read_pcap_into_mem();
  }

#ifdef _WIN32
  InitializeCriticalSection(&gPrintCS);
#endif

  snf_init(SNF_VERSION_API);

  /* Start work threads */
  memset(inj_info, 0, sizeof(inj_info));
  for (i = 0; i < num_threads; i++) {
    struct inj_thread_info * info = &inj_info[i];
    info->id = i;
#ifndef _WIN32
    rc = pthread_create(&thread_ids[i], NULL, work_thread, info);
    if (rc != 0) {
      perror("pthread_create");
      exit(1);
    }
#else
    info->hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (info->hEvent == NULL) {
      fprintf(stderr, "CreateEvent failed: %d\n", GetLastError());
      exit(1);
    }

    info->hThread =
        CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) &work_thread, info, 0, NULL);
    if (info->hThread == NULL) {
      fprintf(stderr, "CreateThread failed: %d\n", GetLastError());
      exit(1);
    }
#endif
  }

  tinfo.pkts_cnt = tinfo.bytes_cnt = 0;
  tinfo.vlan_tag_inserts = 0;
  tinfo.t_elapsed_us = 0;
  tinfo.nerror = 0;
  /* Wait for work thread completion */
  for (i = 0; i < num_threads; i++) {
#ifndef _WIN32
    rc = pthread_join(thread_ids[i], &thread_rc);
    if (rc == 0 && thread_rc) {
      struct inj_thread_info *ti = (struct inj_thread_info *) thread_rc;
#else
    rc = WaitForSingleObject(inj_info[i].hEvent, INFINITE);
    if (rc == WAIT_OBJECT_0) {
      struct inj_thread_info *ti = (struct inj_thread_info *) &inj_info[i];
#endif
      tinfo.pkts_cnt += ti->pkts_cnt;
      tinfo.bytes_cnt += ti->bytes_cnt;
      if (ti->t_elapsed_us > tinfo.t_elapsed_us)
        tinfo.t_elapsed_us = ti->t_elapsed_us;
    }
  }

  if (tinfo.pkts_cnt && num_threads > 1) {
    printf("\nAggregate packet send information for %d threads:\n", num_threads);
    print_tinfo(&tinfo, "All threads");
  }

  return 0;
}
