#ifndef _myri_license_defs_h_
#define _myri_license_defs_h_

/*
 * current license version is 3.  version 3 is needed to support
 * non-integer serial numbers.  firmware that supports version 3
 * licenses will also support version 2 licenses.
 *
 * A version 3 license key looks like:
 *  xxxx-xxxx-xxxx-xxxx:3:<serial>:<product>[:Ttttttttt][:V<major>[.<minor>]]
 * This gets stuck into memory as:
 *  <struct_license_info><binary_signature><license_text>
 * The signature is computed on the binary struct_license_info
 *
 * A version 2 license key looks like:
 *  xxxx-xxxx-xxxx-xxxx:2:<serial>:<product>[:Ttttttttt][:V<major>[.<minor>]]
 * This gets stuck into memory as:
 *  <struct_license_info_2><binary_signature><license_text>
 * The signature is computed on the binary struct_license_info
 */

#define MYRI_MAX_KEY_LENGTH 128
#define MYRI_MAX_KEY_PRODUCT_LENGTH 16
#define MYRI_MAX_SIGNATURE_LENGTH 23
#define MYRI_MAX_SERIAL_LENGTH 24
#define MYRI_SIGNATURE_BYTES 8
#define KEY_SECTOR_ADDR 0xe0000

/*
 * Binary form of the license text
 */
struct myri_license_info {
  uint32_t license_version;
  uint8_t serial[MYRI_MAX_SERIAL_LENGTH+1];
  uint8_t product[MYRI_MAX_KEY_PRODUCT_LENGTH];
  uint16_t product_version_major;
  uint16_t product_version_minor;
  uint32_t expire_time;
};
struct myri_license_info_2 {
  uint32_t license_version;
  uint32_t serial;
  uint8_t product[MYRI_MAX_KEY_PRODUCT_LENGTH];
  uint16_t product_version_major;
  uint16_t product_version_minor;
  uint32_t expire_time;
};

/*
 * The entire binary license representation
 */
struct myri_license_key {
  struct myri_license_info info;
  uint8_t signature[MYRI_SIGNATURE_BYTES];
  uint32_t length;
};
struct myri_license_key_2 {
  struct myri_license_info_2 info;
  uint8_t signature[MYRI_SIGNATURE_BYTES];
  uint32_t length;
};


/*
 * return codes for license key validation
 * This is an ordered list.  When multiple keys are present, each key
 * is tested in turn and the lowest ordinal return code is used.
 */
enum {
 MYRI_KEYCHECK_SUCCESS=0,		/* key matches */
 MYRI_KEYCHECK_FAIL_TIME_BUILD, /* the build has expired, not a license problem */
 MYRI_KEYCHECK_MALFORMED,	/* malformed key string */
 MYRI_KEYCHECK_FAIL_SIGNATURE,	/* valid format, bad signature */
 MYRI_KEYCHECK_FAIL_SERIAL,	/* valid format, serial number mismatch */
 MYRI_KEYCHECK_FAIL_VERSION,	/* valid format, version mismatch */
 MYRI_KEYCHECK_FAIL_TIME,	/* valid format, expired */
 MYRI_KEYCHECK_FAIL_PRODUCT,	/* valid format, product mismatch */
 MYRI_KEYCHECK_UNSUPPORTED_HW,	/* hardware not supported */
 MYRI_KEYCHECK_NOTLOADED,	/* no license key loaded */
 MYRI_KEYCHECK_MAXVAL		/* a number greater than any valid value */
};

static char *myri_keycheck_str[] = {
  "Valid",
  "Pre-release build has expired",
  "Malformed key string",
  "Invalid key signature",
  "Serial number mismatch",
  "No key for this product version",
  "Key expired",
  "No key for this product",
  "Hardware not supported",
  "License key not loaded"
};

/*
 * parse the fields from a key string
 */
MAL_FUNC(int) parse_key(char *key_string, int *license_versionp,
  uint8_t *signature, char *key_serialp, char *key_productp,
  int *key_version_majorp, int *key_version_minorp, uint32_t *key_secondsp);
MAL_FUNC(int) expand_key(char *key, char *keybuf);
MAL_FUNC(void) key_to_text(char *key, char *outbuf);
MAL_FUNC(void) key_to_string(char *key, char *outbuf, int *skip);
MAL_FUNC(void) convert_key_info_2(struct myri_license_info_2 *key2, struct myri_license_info *key);

#endif
