#define _GNU_SOURCE 1
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#ifndef _WIN32
#include <unistd.h>
#include <ctype.h>
#endif

#include "mal.h"
#include "myri_license_defs.h"

/*
 * parse the fields from a key string
 */
MAL_FUNC(int)
parse_key(
  char *key_string,
  int *license_versionp,
  uint8_t *signature,
  char *key_serialp,
  char *key_productp,
  int *key_version_majorp,
  int *key_version_minorp,
  uint32_t *key_secondsp)
{
  char key_product[MYRI_MAX_KEY_PRODUCT_LENGTH];
  int license_version;
  char key_serial[MYRI_MAX_SERIAL_LENGTH+1];
  int key_version_major;	/* product version from key */
  int key_version_minor;
  uint32_t key_seconds;	/* product timeout from key */
  uint32_t s1,s2,s3,s4;
  char *prodp;
  char *cp;
  uint32_t tmp;
  int i;

  /* set key values to defaults */
  key_version_major = -1;
  key_version_minor = -1;
  key_seconds = 0;

  /*
   * first, make sure key string is null terminated
   * within character limit, saves headaches later
   */
  i = 0;
  while (key_string[i] != '\0' &&
	 i < MYRI_MAX_KEY_LENGTH) {
    i++;
  }
  if (i >= MYRI_MAX_KEY_LENGTH) {
    return MYRI_KEYCHECK_MALFORMED;
  }

  /* parse key string */
  cp = key_string;
  if (signature != NULL) {
    sscanf(cp, "%x-%x-%x-%x:", &s1, &s2, &s3, &s4);
    *(uint16_t *)signature = htons(s1);
    *(uint16_t *)(signature+2) = htons(s2);
    *(uint16_t *)(signature+4) = htons(s3);
    *(uint16_t *)(signature+6) = htons(s4);
  }

  /* find ':' before license version number */
  cp = strchr(cp, ':');
  if (cp == NULL) return MYRI_KEYCHECK_MALFORMED;

  /* skip to license version number */
  cp++;
  license_version = atoi(cp);

  /* find ':' before serial number */
  cp = strchr(cp, ':');
  if (cp == NULL) return MYRI_KEYCHECK_MALFORMED;

  /* skip to and extract serial number */
  cp++;
  i = 0;
  while (*cp != ':' && i < (MYRI_MAX_SERIAL_LENGTH))
    key_serial[i++] = toupper(*cp++);
  key_serial[i] = '\0';

  /* now skip to ':' before product code */
  /* probably already there, so check first */
  if (*cp != ':') cp = strchr(cp, ':');
  if (cp == NULL) return MYRI_KEYCHECK_MALFORMED;

  /* skip to product code, copy into buffer */
  cp++;
  prodp = key_product;
  while (*cp != '\0' && *cp != ':') {
    *prodp++ = *cp++;
  }
  *prodp = '\0';	/* terminate it */

  /*
   * now, we process the rest of the string in a loop looking
   * for optional fields
   */
  while (*cp == ':') {
    cp++;
    switch (*cp) {

    /* product version number */
    case 'V':
      cp++;
      tmp = 0;
      while (isdigit(*cp)) {
	tmp *= 10;
	tmp += *cp & 0x0F;
	cp++;
      }
      key_version_major = tmp;

      /* If next character is a '.', there is a minor number */
      if (*cp == '.') {
	cp++;
	tmp = 0;
	while (isdigit(*cp)) {
	  tmp *= 10;
	  tmp += *cp & 0x0F;
	  cp++;
	}
	key_version_minor = tmp;
      }
      break;

    /* product time lock */
    case 'T':
      cp++;
      tmp = 0;
      while (isdigit(*cp)) {
	tmp *= 10;
	tmp += *cp & 0x0F;
	cp++;
      }
      key_seconds = tmp;
      break;

    /* Skip anything else */
    default:
      while (*cp != ':' && *cp != '\0') {
	cp++;
      }
      break;
    }
  }

  /* copy out values if we get here */
  if (license_versionp != NULL) *license_versionp = license_version;
  if (key_serialp != NULL) strcpy(key_serialp, key_serial);
  if (key_productp != NULL) strcpy(key_productp, key_product);
  if (key_version_majorp != NULL) *key_version_majorp = key_version_major;
  if (key_version_minorp != NULL) *key_version_minorp = key_version_minor;
  if (key_secondsp != NULL) *key_secondsp = key_seconds;

  return MYRI_KEYCHECK_SUCCESS;
}

static int
encode_key_info_2(
  char *key_buffer,
  uint8_t *signature,
  char *serial,
  char *product,
  int version_major,
  int version_minor,
  uint32_t seconds,
  char *text_key)
{
  struct myri_license_info_2 *lip;
  struct myri_license_key_2 *lkp;
  uint32_t sn;
  int key_len;
  char *cp;

  /* clear the whole buffer */
  memset(key_buffer, 0, MYRI_MAX_KEY_LENGTH);

  lkp = (struct myri_license_key_2 *) key_buffer;
  lip = &lkp->info;

  /* encode the license information in binary */
  memcpy(lkp->signature, signature, MYRI_SIGNATURE_BYTES);

  lip->license_version = htonl(2);
  sn = atoi(serial);
  lip->serial = htonl(sn);
  memset(lip->product, 0, MYRI_MAX_KEY_PRODUCT_LENGTH);
  strcpy((char *)lip->product, product);
  lip->product_version_major = htons(version_major);
  lip->product_version_minor = htons(version_minor);
  lip->expire_time = htonl(seconds);

  /* add the text to the end */
  cp = key_buffer + sizeof(*lkp);
  strcpy(cp, text_key);

  /* round up to multiple of 4 and add size of binary part */
  key_len = (int)strlen(text_key)+1;
  key_len = (key_len+3) & ~3;
  key_len += sizeof(*lkp);

  /* save total length in the key */
  lkp->length = htonl(key_len);

  return key_len;
}

static int
encode_key_info_3(
  char *key_buffer,
  uint8_t *signature,
  char *serial,
  char *product,
  int version_major,
  int version_minor,
  uint32_t seconds,
  char *text_key)
{
  struct myri_license_info *lip;
  struct myri_license_key *lkp;
  int key_len;
  char *cp;

  /* clear the whole buffer */
  memset(key_buffer, 0, MYRI_MAX_KEY_LENGTH);

  lkp = (struct myri_license_key *) key_buffer;
  lip = &lkp->info;

  /* encode the license information in binary */
  memcpy(lkp->signature, signature, MYRI_SIGNATURE_BYTES);

  lip->license_version = htonl(3);
  strncpy((char *)lip->serial, serial, MYRI_MAX_SERIAL_LENGTH);
  lip->serial[MYRI_MAX_SERIAL_LENGTH] = '\0';
  memset(lip->product, 0, MYRI_MAX_KEY_PRODUCT_LENGTH);
  strcpy((char *)lip->product, product);
  lip->product_version_major = htons(version_major);
  lip->product_version_minor = htons(version_minor);
  lip->expire_time = htonl(seconds);

  /* add the text to the end */
  cp = key_buffer + sizeof(*lkp);
  strcpy(cp, text_key);

  /* round up to multiple of 4 and add size of binary part */
  key_len = (int) strlen(text_key)+1;
  key_len = (key_len+3) & ~3;
  key_len += sizeof(*lkp);

  /* save total length in the key */
  lkp->length = htonl(key_len);

  return key_len;
}

MAL_FUNC(int)
encode_key_info(
  char *key_buffer,
  int license_version,
  uint8_t *signature,
  char *serial,
  char *product,
  int version_major,
  int version_minor,
  uint32_t seconds,
  char *text_key)
{
  if (license_version == 2) {
    return encode_key_info_2(key_buffer, signature, serial, product,
	version_major, version_minor, seconds, text_key);
  } else if (license_version == 3) {
    return encode_key_info_3(key_buffer, signature, serial, product,
	version_major, version_minor, seconds, text_key);
  } else {
    errno = EINVAL;
    return -1;
  }
}

MAL_FUNC(int)
expand_key(
  char *key,
  char *keybuf)
{
  char key_product[MYRI_MAX_KEY_PRODUCT_LENGTH];
  uint8_t signature[MYRI_SIGNATURE_BYTES];
  int license_version = 0;
  char key_serial[MYRI_MAX_SERIAL_LENGTH+1];
  int key_version_major = 0;
  int key_version_minor = 0;
  uint32_t key_seconds = 0;
  int key_len;

  parse_key(key, &license_version, signature, key_serial, key_product,
      &key_version_major, &key_version_minor, &key_seconds);

  memset(keybuf, 0, MYRI_MAX_KEY_LENGTH);
  key_len = encode_key_info(keybuf, license_version, signature, key_serial,
      key_product, key_version_major, key_version_minor, key_seconds, key);

  return key_len;
}

static void
key_to_partial_string(
  char *buf,
  struct myri_license_info *lip)
{
  char *bp;
  time_t expire;
  struct tm *when;
  int16_t vmaj;
  int16_t vmin;

  bp = buf;

  bp += sprintf(bp, "%s", lip->product);

  vmaj = ntohs(lip->product_version_major);
  vmin = ntohs(lip->product_version_minor);
  if (vmaj != -1) {
    bp += sprintf(bp, ", V%d", vmaj);
    if (vmin != -1) {
      bp += sprintf(bp, ".%d", vmin);
    }
  } else {
    bp += sprintf(bp, " (all versions)");
  }

  expire = ntohl(lip->expire_time);
  if (expire != 0) {
    when = localtime(&expire);
    bp += sprintf(bp, " (expires %d/%02d/%d)",
        when->tm_mon+1, when->tm_mday, when->tm_year+1900);
  }
}

MAL_FUNC(void)
key_to_text(
  char *key,
  char *outbuf)
{
  int version;
  int skip;

  version = *(uint32_t *)key;
  version = ntohl(version);

  skip = (version == 2) ? sizeof(struct myri_license_key_2) : sizeof(struct myri_license_key);

  sprintf(outbuf, "%s", key + skip);
}

MAL_FUNC(void)
key_to_string(
  char *key,
  char *outbuf,
  int *length)
{
  struct myri_license_key *kp;
  struct myri_license_key_2 *kp2;
  struct myri_license_info keyinfo;
  struct myri_license_info *ki;

  int version;
  char buf[128];
  char buf2[128];

  version = *(uint32_t *)key;
  version = ntohl(version);

  if (version < 3) {
    kp2 = (struct myri_license_key_2 *)key;
    convert_key_info_2(&kp2->info, &keyinfo);
    ki = &keyinfo;
    *length = ntohl(kp2->length);
  } else if (version == 3)  {
    kp = (struct myri_license_key *)key;
    *length = ntohl(kp->length);
    ki = &kp->info;
  } else {
    sprintf(outbuf, "Unrecognized key data\n");
    return;
  }

  key_to_text(key, buf);
  key_to_partial_string(buf2, ki);
  sprintf(outbuf, "%s # %s", buf, buf2);
}

/*
 * This function is really only useful for extracting data for printing, 
 * since signature of v2 struct will not match v3 for validation
 */
MAL_FUNC(void)
convert_key_info_2(
  struct myri_license_info_2 *key2, 
  struct myri_license_info *key)
{
  key->license_version = key2->license_version;
  sprintf((char *)key->serial, "%d", ntohl(key2->serial));
  strcpy((char *)key->product, (char *)key2->product);
  key->product_version_major = key2->product_version_major;
  key->product_version_minor = key2->product_version_minor;
  key->expire_time = key2->expire_time;
}
