/*************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET          *
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the       *
 * "License"); User may not use this file except in compliance with the  *
 * License.  The full text of the License can found in LICENSE.TXT       *
 *                                                                       *
 * Software distributed under the License is distributed on an "AS IS"   *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See  *
 * the License for the specific language governing rights and            *
 * limitations under the License.                                        *
 *                                                                       *
 * Copyright 2003 - 2010 by Myricom, Inc.  All rights reserved.          *
 *************************************************************************/

/* modifications for MX kernel lib made by
 * Brice.Goglin@ens-lyon.org (LIP/INRIA/ENS-Lyon) */

#include "mx_arch.h"
#include "mx_misc.h"
#include "mx_instance.h"
#include "mx_malloc.h"
#include "mx_pio.h"
#include "mal_stbar.h"
#include "myri_version.h"
#include "mal_cpu.h"
#include "mcp_events.h"

#include "kraw.h"
#include "myri_raw.h"


#include "mx_ether_common.h"


#include "bsd/quad.h"

#ifndef MX_DISABLE_COMMON_COPYBLOCK
#define MX_DISABLE_COMMON_COPYBLOCK 0
#endif

#ifndef MX_DMA_VPAGE_SIZE 
#define MX_DMA_VPAGE_SIZE MX_VPAGE_SIZE
#define MX_ALLOC_DMA_PAGE(is, addr, pin, len) mx_alloc_dma_page(is, addr, pin)
#define MX_FREE_DMA_PAGE(is, addr, pin, len) mx_free_dma_page(is, addr, pin)
#endif

int mx_initialized = 0;
uint32_t myri_mx_small_message_threshold = 1024;
uint32_t myri_mx_medium_message_threshold = 4*1024;
uint32_t myri_security_disabled = 0;
mx_spinlock_t mx_lanai_print_spinlock;

mx_sync_t mx_global_mutex;

uint32_t myri_debug_mask;

uint32_t myri_msi_level = MAL_OS_LINUX && MAL_CPU_powerpc64;

mx_atomic_t myri_max_user_pinned_pages;

uint32_t myri_clksync_period = 1000; /* in millisecs */
uint32_t myri_clksync_error = 30000;
uint32_t myri_clksync_error_count = 0;
uint32_t myri_clksync_verbose = 0;

enum myri_timeesource_loss_mode {
  MYRI_TIMESOURCE_LOSS_ZERO = 0,
  MYRI_TIMESOURCE_LOSS_FREERUN = 1,
  MYRI_TIMESOURCE_LOSS_HOSTTIME = 2 /* might not make sense, not used now */
};

uint32_t myri_timesource_loss_mode = MYRI_TIMESOURCE_LOSS_FREERUN;

int
mx_init_driver(void)
{
#if MAL_CPU_x86 && MX_ENABLE_SSE2 && !defined _MSC_VER
  if (!mal__cpu_has_sse2()) {
    MX_WARN(("Processor without sse2: recompile with --disable-sse2.\n"));
    return EIO;
  }
#endif
  MX_INFO(("%s Version %s\n", MYRI_DRIVER_STR, MYRI_VERSION_STR));
  MX_INFO(("Build %s\n", MYRI_BUILD_STR));
  MX_INFO(("Debug %s\n", MAL_DEBUG ? "ON" : "OFF"));
  mx_set_default_hostname();

  mx_lx_init_board_ops();
  mx_lz_init_board_ops();

  myri_kraw_init_ops();
  myri_snf_init_ops();

  MAL_DEBUG_PRINT 
    (MAL_DEBUG_BOARD_INIT,
     (MYRI_DRIVER_STR " configured for %d instances\n", myri_max_instance));
  mx_instances = mx_kmalloc 
    (sizeof(mx_instances[0]) * myri_max_instance, MX_MZERO);
  if (mx_instances == 0) {
    return ENOMEM;
  }


  mx_sync_init(&mx_global_mutex, NULL, -1, "mx_global_mutex");
  mx_spin_lock_init(&mx_lanai_print_spinlock, NULL, -1, "lanai print spinlock");
  mx_initialized = 1;

  if (myri_security_disabled) {
    MX_WARN(("Security is disabled\n"));
  }

  return 0;
}

int
mx_finalize_driver(void)
{
  if (!mx_initialized) {
    return -1;
  }
  mx_kfree(mx_instances);
  mx_instances = 0;
  mx_sync_destroy(&mx_global_mutex);
  mx_spin_lock_destroy(&mx_lanai_print_spinlock);
  return 0;
}

int
mx_alloc_zeroed_dma_page(mx_instance_state_t *is, char **addr, 
			 mx_page_pin_t *pin)
{
  int status;

  status = mx_alloc_dma_page(is, addr, pin);
  if (!status)
    bzero(*addr, PAGE_SIZE);
  return status;
}

int
mx_alloc_zeroed_dma_page_relax_order(mx_instance_state_t *is, char **addr, 
				     mx_page_pin_t *pin)
{
  int status;

  status = mx_alloc_dma_page_relax_order(is, addr, pin);
  if (!status)
    bzero(*addr, PAGE_SIZE);
  return status;
}

mx_instance_state_t *
mx_get_instance(uint32_t unit)
{
  mx_instance_state_t *is;

  /* make sure the requested unit is in range  */
  if (unit >= myri_max_instance) {
    return NULL;
  }

  /* grab the global mutex to make sure the global array of instances
     is consistant */
  mx_mutex_enter(&mx_global_mutex);
  is = mx_instances[unit];

  /* make sure its a valid instance */
  if (is == 0) {
    mx_mutex_exit(&mx_global_mutex);      
    return NULL;
  }
  
  mx_atomic_add(1, &is->ref_count);
  mx_mutex_exit(&mx_global_mutex);
  return is;
}

void
mx_release_instance(mx_instance_state_t *is)
{
  mx_atomic_subtract(1, &is->ref_count);
}

static int
mx_get_logging(mx_instance_state_t *is, uint32_t size, uaddr_t out,
	       uint32_t is_kernel)
{
  char *sram_ptr, *bounce;
  int status = 0;
  uint32_t logging_offset, dont_care, copy_size;

  /* sanity check */
  if (mx_is_dead(is)) {
    return (ENXIO);
  }
  
  /* sanity check */
  if (size == 0) {
    return (0);
  }
  
  /* get the SRAM offset of the logging data */
  logging_offset = MCP_GETVAL(is, logging_offset);
  
  /* sanity check */
  if (logging_offset == 0) {
    MX_WARN (("Logging is not enabled in the MCP build\n"));
    return (ENXIO);
  }
  
  /* the MCP is now logging */
  mx_mutex_enter(&is->sync);
  if (is->flags & MX_IS_LOGGING) {
    mx_mutex_exit(&is->sync);
    return EBUSY;
  }
  is->flags |= MX_IS_LOGGING;
  mx_mutex_exit(&is->sync);
  
  sram_ptr = (char *)is->lanai.sram + (size_t) logging_offset;
  bounce = mx_kmalloc(MX_VPAGE_SIZE, MX_WAITOK|MX_MZERO);
  if (bounce == 0) {
    status = ENOMEM;
    goto abort_with_sync;
  }

  /* tell the MCP to start logging */
  status = mx_mcp_command(is, MCP_CMD_START_LOGGING, size, 0, 0, &dont_care);
  if (status) {
    MX_WARN (("Couldn't start logging\n"));
    goto abort_with_buffer;
  }

  /* wait for the MCP to log the requested amount of information */
  status = mx_sleep(&is->logging.sync, MX_LOGGING_WAIT, MX_SLEEP_INTR);
  if (status) {
    if (status == EAGAIN)
      MX_WARN(("time out waiting for logging buffer\n"));
    mx_sync_reset(&is->logging.sync);
    goto abort_with_buffer;
  }
  
  /* retrieve the logging data */
  size = (size > is->logging.size) ? is->logging.size : size;
  while (size > 0) {
    copy_size = (size > MX_VPAGE_SIZE) ? MX_VPAGE_SIZE : size;
    mx_pio_bcopy_read(sram_ptr, bounce, copy_size);
    status = mx_copyout(bounce, out, copy_size, is_kernel);
    if (status) {
      MX_WARN(("failed to copy logging buffer out to 0x%lx\n", 
	       (unsigned long) out));
      break;
    }

    size -= copy_size;
    sram_ptr += copy_size;
    out += copy_size;
  }

 abort_with_buffer:
  mx_kfree(bounce);
   
 abort_with_sync:
  mx_mutex_enter(&is->sync);
  is->flags &= ~MX_IS_LOGGING;
  mx_mutex_exit(&is->sync);
  return status;
}

static int
mx_run_dmabench(mx_instance_state_t *is, mx_dmabench_t *x)
{
  mx_page_pin_t pin;
  char *addr;
  uint32_t dont_care;
  uint16_t count_size;
  int status, cmd;

  if ((x->count > 100) ||
      (1 << x->log_size) > MX_VPAGE_SIZE)
    return EINVAL;

  mx_mutex_enter(&is->sync);
  if (is->dmabench.busy) {
    mx_mutex_exit(&is->sync);
    return EBUSY;
  }
  is->dmabench.busy = 1;
  mx_mutex_exit(&is->sync);
  status = mx_alloc_dma_page_relax_order(is, &addr, &pin);
  if (status)
    goto abort_busy;

  if (x->dma_read)
    cmd = MCP_CMD_START_DMABENCH_READ;
  else 
    cmd = MCP_CMD_START_DMABENCH_WRITE;
  count_size = (x->count << 8) | x->log_size;
  status = mx_mcp_command(is, cmd, count_size, pin.dma.high,
			  pin.dma.low, &dont_care);

  if (status) {
    MX_WARN(("Could not start DMA benchmark, status = %d\n", status));
    MX_WARN(("0x%x 0x%x\n", pin.dma.high, pin.dma.low));
    goto abort_with_dmapage;
  }

  status = mx_sleep(&is->dmabench.wait_sync, MCP_COMMAND_TIMEOUT * 2,
		    MX_SLEEP_NOINTR);

  x->count = is->dmabench.count;
  x->cycles = is->dmabench.cycles;

 abort_with_dmapage:
  mx_free_dma_page(is, &addr, &pin);

 abort_busy:
  mx_mutex_enter(&is->sync);
  is->dmabench.busy = 0;
  mx_mutex_exit(&is->sync);
  return status;
}

#if !MX_DISABLE_COMMON_COPYBLOCK

#if MAL_OS_WINDOWS
void mx_win_alloc_copyblock(mx_copyblock_t *cb);
void mx_win_free_copyblock(mx_copyblock_t *cb);
int mx_win_alloc_huge_copyblock(mx_instance_state_t *is, 
				mx_huge_copyblock_t *hcb,
				unsigned long lib_offset,
				int node_id);
void mx_win_free_huge_copyblock(mx_instance_state_t *is,
				mx_huge_copyblock_t *hcb);
#endif

void
mx_common_free_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb)
{
  unsigned i;

  if (cb->pins != NULL) {
    for (i = 0; i < (cb->size / PAGE_SIZE); i++) {
      if (cb->pins[i].va != 0) {
	mx_unpin_page(is, &cb->pins[i], MX_PIN_KERNEL | MX_PIN_CONSISTENT);
	mx_unreserve_page((void *)(uintptr_t)cb->pins[i].va);
      }
    }
    mx_kfree(cb->pins);
  }
  if (cb->alloc_addr != NULL) {
#if MAL_OS_WINDOWS
    mx_win_free_copyblock(cb);
#else
    mx_kfree(cb->alloc_addr);
#endif
  }
  if (cb->dmas != NULL)
    mx_kfree(cb->dmas);

  cb->size = 0;
  bzero(cb, sizeof(*cb));
}

int
mx_common_alloc_copyblock(mx_instance_state_t *is, mx_copyblock_t *cb)
{
  int i, status;
  uint64_t va;
  uint32_t len;

  len = cb->size;

  /* make sure that the length is page aligned */
  mal_always_assert((len & (PAGE_SIZE - 1)) == 0);

#if MAL_OS_WINDOWS
  mx_win_alloc_copyblock(cb);
#else
  cb->alloc_addr = mx_kmalloc(len + PAGE_SIZE, MX_MZERO|MX_WAITOK);
#endif
  cb->pins = mx_kmalloc(sizeof(cb->pins[0]) * (len / PAGE_SIZE), MX_MZERO|MX_WAITOK);
  cb->dmas = mx_kmalloc(sizeof(mcp_dma_addr_t) * (len / MX_VPAGE_SIZE), MX_MZERO|MX_WAITOK);
  if ((cb->alloc_addr == NULL) || (cb->pins == NULL) || (cb->dmas == NULL)) {
    MX_WARN(("copyblock pin info allocation failed due to lack of memory\n"));
    status = ENOMEM;
    goto abort_with_cb;
  }
  cb->addr =  (void *)(uintptr_t)MX_PAGE_ALIGN((uintptr_t)cb->alloc_addr);
  
  for (i = 0, va = (uint64_t)(size_t)cb->addr; 
       va < (uint64_t)(size_t)((char *)cb->addr + len); 
       va += PAGE_SIZE, i++) {
    cb->pins[i].va = va;
    status = mx_pin_page(is, &cb->pins[i], MX_PIN_KERNEL|MX_PIN_CONSISTENT, 0);
    if (status) {
      MX_WARN(("Failed to pin copyblock, status = %d\n", status));	
      cb->pins[i].va = 0;
      goto abort_with_cb;
    }
    mx_reserve_page((void *)(uintptr_t)va);
  }
  
  /* fill up the dma ring */
  for (i = 0; i < ((len + (MX_VPAGE_SIZE - 1)) / MX_VPAGE_SIZE); i++) {
    cb->dmas[i].high = htonl(cb->pins[i/(PAGE_SIZE/MX_VPAGE_SIZE)].dma.high);
    cb->dmas[i].low = htonl(cb->pins[i/(PAGE_SIZE/MX_VPAGE_SIZE)].dma.low);
  }
       
  return(0);

 abort_with_cb:
  mx_common_free_copyblock(is, cb);
  return (status);
}
#endif

void
mx_free_huge_copyblock(mx_instance_state_t *is, mx_huge_copyblock_t *hcb)
{
  char *c;
  unsigned long i;
  unsigned long npages = (unsigned long)((hcb->size + PAGE_SIZE - 1) / PAGE_SIZE);

#if MAL_OS_WINDOWS
  mx_win_free_huge_copyblock(is, hcb);
  return;
#endif

  if (hcb->kva != NULL)
    mx_unmap_huge_copyblock(hcb);

  if (hcb->pins != NULL) {
    for (i = 0; i < npages; i++) {
      if (hcb->pins[i].va != 0) {
	c = (char *)(uintptr_t)hcb->pins[i].va;
	mx_unreserve_page(c);
	mx_free_dma_page(is, &c, &hcb->pins[i]);
      }
    }
    mx_kfree(hcb->pins);
  }
  hcb->pins = NULL;
  if (hcb->cb.pins != NULL)
    mx_free_copyblock(is, &hcb->cb);
  hcb->cb.pins = NULL;
  hcb->size = 0;
  mx_atomic_add(hcb->pinned_pages, &myri_max_user_pinned_pages);
  hcb->pinned_pages = 0;
}


int
mx_alloc_huge_copyblock(mx_instance_state_t *is, 
			mx_huge_copyblock_t *hcb,
			unsigned long lib_offset,
                        int node_id)
{
  mcp_dma_addr_t *dma;
  char *c;
  unsigned long i, addr, host_index, host_offset, pin_offset, pin_index;
  unsigned long npages = (unsigned long)((hcb->size + PAGE_SIZE - 1) / PAGE_SIZE);
  unsigned long nvpages =  (unsigned long)((hcb->size - lib_offset + MX_VPAGE_SIZE - 1) / MX_VPAGE_SIZE);
  int status;

#if MAL_OS_WINDOWS
  return mx_win_alloc_huge_copyblock(is, hcb, lib_offset, node_id);
#endif

  hcb->pinned_pages = 0;
  if (mx_atomic_read(&myri_max_user_pinned_pages) < npages) {
    hcb->size = 0;
    status = ENOMEM;
    goto abort;
  }
  mx_atomic_subtract(npages, &myri_max_user_pinned_pages);
  hcb->pinned_pages = npages;

  /* do the hard part first, and allocate big gobs of RAM */
  hcb->pins = mx_kmalloc_node(sizeof(hcb->pins[0]) * npages, MX_MZERO|MX_WAITOK, node_id);
  if (hcb->pins == NULL) {
    MX_WARN(("huge copyblock allocation failed due to lack of memory\n"));
    status = ENOMEM;
    goto abort;
  }

  for (i = 0; i < npages; i++) {
    status = mx_alloc_zeroed_dma_page(is, &c, &hcb->pins[i]);
    if (status) {
      MX_WARN(("Failed to alloc copyblock page %ld/%ld, status = %d\n",
	       i, npages, status));
      hcb->pins[i].va = 0;
      goto abort;
    }
    mx_reserve_page(c);
  }

  /* Allocate some pinnable memory to hand to the MCP, just use
     existing copyblock routines for ease
   */
  hcb->cb.size = nvpages * sizeof(mcp_dma_addr_t);
  status = mx_alloc_copyblock(is, &hcb->cb);
  if (status) {
    MX_WARN(("mx_alloc_huge_copyblock:copyblock alloc failure\n"));
    goto abort;
  }

  /* fill the copyblock with the the vpage dma addresses, we start
     lib_offset pages into the copyblock, so as to allow the lib
     to map some blank pages in front of the data copyblock
  */
  pin_index = 0;
  pin_offset = 0;
  for (addr = lib_offset; addr < hcb->size; addr += MX_VPAGE_SIZE) {
    dma = (mcp_dma_addr_t *)(uintptr_t)(hcb->cb.pins[pin_index].va + pin_offset);
    host_index = (unsigned long)MX_ATOP(addr);
    host_offset = (uint32_t)(addr - MX_PAGE_TRUNC(addr));
#if 0
    printk("Filling: addr: %ld host_index, %ld host_offset: %ld,  pin_index: %ld,  pin_offset: %ld\n",
	   addr, host_index, host_offset, pin_index, pin_offset);
#else
    dma->high = htonl(hcb->pins[host_index].dma.high);
    dma->low = htonl(hcb->pins[host_index].dma.low + host_offset);
#endif
    pin_offset += sizeof(mcp_dma_addr_t);
    if (pin_offset >= MX_PAGE_SIZE) {
      pin_offset = 0;
      pin_index++;
    }
  }

  return 0;

 abort:
  mx_free_huge_copyblock(is, hcb);
  return status;
}

int
mx_dma_map_copyblock(mx_endpt_state_t *es, mx_copyblock_t *cb, uint32_t offset)
{
  mx_instance_state_t *is;
  mcp_dma_addr_t *dma;
  uint32_t host_offset;
  uaddr_t mcp_index, host_index;
  unsigned long addr;

  is = es->is;
  dma = (mcp_dma_addr_t *)((char *)(is->lanai.sram + (unsigned long)offset));

  /* Walk the copyblock, storing the DMA address for each virtual page
     down on the nic.  Note that on hosts with large page sizes, two
     or more virtual pages may map to offsets within the same host
     page. */
     
  for (addr = 0; addr < cb->size; addr += MX_VPAGE_SIZE) {
    mcp_index = MX_ATOVP(addr);
    host_index = MX_ATOP(addr);

    /* compute the offset into the (possibly larger) host page */
    host_offset = (uint32_t)(addr - MX_PAGE_TRUNC(addr));
    
#if 1   /* assume 64 bit-writes are not causing padding fifo overflow */
    MX_PIO_WRITE((uint64_t*)&dma[mcp_index], 
		 mx_hton_u64(((uint64_t)cb->pins[host_index].dma.high << 32) +
			     cb->pins[host_index].dma.low + host_offset));
#else
    MX_PIO_WRITE(&(dma[mcp_index].high), htonl(cb->pins[host_index].dma.high));
    MX_PIO_WRITE(&(dma[mcp_index].low), htonl(cb->pins[host_index].dma.low + host_offset));
#endif
    MAL_STBAR();
    
  }
  return (0);
}

static int
mx_dma_map_copyblocks(mx_endpt_state_t *es)
{
  uint32_t offset;
  int status = 0;
  
  /* send queue */
  status = mx_mcp_command(es->is, MCP_CMD_GET_ENDPT_SENDQ_OFFSET, 
			  es->endpt, 0, 0, &offset);
  if (status) {
    MX_WARN(("could not find get endpoint sendq map offset\n"));
    return (status);
  }

  status |= mx_dma_map_copyblock(es, &es->sendq, offset);

  /* recv queue */
  status = mx_mcp_command(es->is, MCP_CMD_GET_ENDPT_RECVQ_MPL_OFFSET, 
			  es->endpt, 0, 0, &offset);
  if (status) {
    MX_WARN(("could not find get endpoint recvq MPL offset\n"));
    return (status);
  }
  
  status |= mx_dma_map_copyblock(es, &es->recvq.cb, offset);

  /* eventq queue */
  status = mx_mcp_command(es->is, MCP_CMD_GET_ENDPT_EVENTQ_OFFSET, 
			  es->endpt, 0, 0, &offset);
  if (status) {
    MX_WARN(("could not find get endpoint eventq map offset\n"));
    return (status);
  }

  status |= mx_dma_map_copyblock(es, &es->eventq, offset);

  return (status);
}



/* basic opening of an endpoint from scratch, still needs to be enabled afterwards */
static int
mx_open_mcp_endpoint(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;
  uint32_t dummy, offset, license_valid;
  int status;

  /* We don't try to be very verbose here because licensed products should make
   * their own license check ioctls before trying to open an endpoint */
  status = myri_check_license(is, &license_valid);
  if (status != 0 || license_valid != 0) {
    MX_WARN (("License invalid or not present\n"));
    return status ? status : EINVAL;
  }

  es->user_mmapped_sram.size = MX_PAGE_ALIGN(MCP_UMMAP_SIZE);
  status = mx_mcp_command(is, MCP_CMD_GET_USER_MMAP_OFFSET,
			  es->endpt, 0, 0, &offset);
  if (status) {
    MX_WARN (("Can't determine user mmap offset \n"));
    return status;
  }

  mal_always_assert((MX_PAGE_SIZE <= MCP_UMMAP_SIZE && 
		     offset % MX_PAGE_SIZE == 0) ||
		    (MX_PAGE_SIZE > MCP_UMMAP_SIZE &&
		     offset % MX_PAGE_SIZE ==
		     (es->endpt * MCP_UMMAP_SIZE) % MX_PAGE_SIZE));
  /* with big page size (> UMMAP_SIZE), we have to map the whole page
     (i.e. we also map some siblings endpoints */
  offset &= ~(MX_PAGE_SIZE - 1);

  es->user_mmapped_sram.addr = 
    ((char *)(is->lanai.sram + (unsigned long)offset));

  if (offset + es->user_mmapped_sram.size > is->sram_size) {
    MX_WARN (("User mapped sram has bad size (0x%x) or location (0x%x)\n",
	      es->user_mmapped_sram.size, offset));
  }
  es->user_mmapped_zreq.addr = (char*)is->lanai.sram + (1<<23) + (es->endpt << 16);
  if (MAL_IS_ZE_BOARD(is->board_type)) {
    /* we have only 4MB of virtual space */
    if (es->endpt >= (0x400000 >> 16)) {
      MX_WARN (("Endpoint number %d too big\n", es->endpt));
      return EINVAL;
    }
    es->user_mmapped_zreq.size = MX_PAGE_ALIGN(MCP_UMMAP_SIZE);
  }

  /* open the endpoint on the mcp */
  status = mx_mcp_command(is, MCP_CMD_OPEN_ENDPOINT, es->endpt, 0, 0, &dummy);
  if (status != 0) {
    MX_WARN(("%s: endpt %d, mcp open failed\n", is->is_name, es->endpt));
    return (status);
  }

  return 0;
}

static
int
myri_init_common_mcp_copyblocks(mx_endpt_state_t *es)
{
  int status = 0;

  /* First static copyblocks, can't be changed */
  es->sendq.flags = MYRI_COPYBLOCK_F_STATIC | MYRI_COPYBLOCK_F_KALLOC;
  es->sendq.size  = (uint32_t) MX_VPTOA(MCP_SENDQ_VPAGE_CNT);
  status = mx_endpt_alloc_copyblock(es, &es->sendq);
  if (status)
    goto abort;

  es->eventq.flags = MYRI_COPYBLOCK_F_STATIC | MYRI_COPYBLOCK_F_KALLOC;
  es->eventq.size = (uint32_t) MX_VPTOA(myri_eventq_vpage_cnt);
  status = mx_endpt_alloc_copyblock(es, &es->eventq);
  if (status)
    goto abort_with_sendq;

  /* Receive queue currently uses a huge copyblock and its size can be
   * controlled */
  es->recvq.cb.flags = MYRI_COPYBLOCK_F_KALLOC | MYRI_COPYBLOCK_F_HUGE;
  return 0;

abort_with_sendq:
  mx_free_copyblock(es->is, &es->sendq);
  es->sendq.size = 0;
abort:
  return status;
}

static
int
myri_common_set_copyblock(mx_endpt_state_t *es, myri_set_copyblock_t *req)
{
  int status = 0;
  mx_copyblock_t *cb;
  mx_huge_copyblock_t *hcb;

  /*
   * Really only the Huge copyblocks can be "set" (i.e. dynamically allocated),
   * but we currently leave the infrastructure in place to support more
   */
  switch (req->cb_type) {
    case MYRI_CB_SENDQ:
      cb = &es->sendq;
      break;

    case MYRI_CB_EVENTQ:
      cb = &es->eventq;
      break;

    case MYRI_CB_RECVQ:
      cb = &es->recvq.cb;
      break;

    default:
      return EINVAL;
      break;
  }

  /* Can't modify static copyblocks and can't modify any copyblocks once the
   * MCP session is enabled */
  if ((cb->flags & MYRI_COPYBLOCK_F_STATIC) ||
      (es->flags & (MX_ES_MCP_ENABLED|MX_ES_MCP_SESSION)) != MX_ES_MCP_SESSION)
    return EINVAL;

  /* TODO (SNF-33): Support user passing pages to pin -- use cb_base/cb_size to
   * pin */
  if (req->cb_mode != MYRI_CB_KALLOC)
    return EINVAL;

  /* Only support dynamic sizing of huge copyblocks since right now, regular
   * copyblocks must be static */
  if (!(cb->flags & MYRI_COPYBLOCK_F_HUGE))
    return EINVAL;

  /* Free any existing copyblocks before allocating this one */
  hcb = mx_cb_to_huge_to_copyblock(cb);
  if (hcb->size)
    mx_free_huge_copyblock(es->is, hcb);

  /* mcp expects full pages of 8-byte dma addresses, hence the alignup
   * requirement */
  hcb->size =
    (uint64_t) MAL_ROUND_UP(req->cb_size,
                            MX_VPTOA(MX_VPAGE_SIZE / sizeof(mcp_dma_addr_t)));
  status = mx_endpt_alloc_huge_copyblock(es, hcb);
  if (status)
    hcb->size = 0;

  return status;
}

int
myri_endpt_enable_common_copyblocks(mx_endpt_state_t *es)
{
  myri_set_copyblock_t req;
  int status = 0;

  /* Can only enable once and can only enable mcp endpoints */
  if ((es->flags & MX_ES_MCP_ENABLED) || !(es->flags & MX_ES_MCP_SESSION))
    return EINVAL;

  /* Receive queue may or may not be already allocated */
  if (es->recvq.size == 0) {
    bzero(&req, sizeof(req));
    req.cb_type = MYRI_CB_RECVQ;
    req.cb_mode = MYRI_CB_KALLOC;
    req.cb_size = (uint64_t) MX_VPTOA(myri_recvq_vpage_cnt);
    status = myri_common_set_copyblock(es, &req);
  }

  /* tell the MCP about the copyblock's DMA addresses */
  status = mx_dma_map_copyblocks(es);
  if (status)
    MX_WARN(("%s: endpt %d, mapping copyblock failed\n", es->is->is_name, es->endpt));
  return status;
}

/* either finish the opening or an endpoint, or re-enable it after recovery */
int
myri_endpt_enable_mcp(mx_endpt_state_t *es)
{
  mx_instance_state_t *is = es->is;
  uint32_t dummy;
  int status;

  /* Can only enable driver endpoints that use mcp sessions */
  if ((es->flags & MX_ES_MCP_ENABLED) || !(es->flags & MX_ES_MCP_SESSION))
    return EINVAL;

  /* set the endpoint session id */
  status = mx_mcp_command(is, MCP_CMD_SET_ENDPOINT_SESSION, es->endpt,
			  es->session_id, 0, &dummy);
  if (status != 0) {
    MX_WARN(("%s: endpt %d, set endpt session_id failed\n", is->is_name, es->endpt));
    return (status);
  }

  /* enable the endpoint on the mcp */
  status = mx_mcp_command(is, MCP_CMD_ENABLE_ENDPOINT, es->endpt, 0, 0, &dummy);
  if (status != 0) {
    MX_WARN(("%s: endpt %d, endpt enable failed\n", is->is_name, es->endpt));
    return (status);
  }

  es->flags |= MX_ES_MCP_ENABLED;

  return 0;
}

extern myri_endpt_ops_t myri_ops_kraw;

static
myri_endpt_ops_t *
myri_endpt_getops(enum myri_endpt_type es_type)
{
  if (es_type == MYRI_ES_RAW)
    return &myri_ops_kraw;
  else if (es_type >= 300 && es_type < 400)
    return &myri_ops_snf;
  else
    return NULL;
}

int
myri_common_open_args(int cmd, uaddr_t arg, myri_set_endpt_t *set_endpt)
{
  int status;
  switch (cmd) {
    case MYRI_SET_ENDPOINT:
    case MX_SET_ENDPOINT:
      status = mx_arch_copyin(arg, set_endpt, sizeof(*set_endpt));
      if (status != 0)
        return status;
      if (cmd == MX_SET_ENDPOINT)
        set_endpt->endpt_type = MYRI_ES_MX;
      break;
    case MX_SET_RAW:
      set_endpt->endpoint = -1;
      set_endpt->endpt_type = MYRI_ES_RAW;
      break;
    default:
      return EINVAL;
  }
  return 0;
}

int
myri_common_open(int32_t unit, mx_endpt_state_t *es, 
                 myri_set_endpt_t *set_endpt, uaddr_t arg)
{
  int status = 0;
  mx_instance_state_t *is;
  myri_endpt_ops_t *eops;
  enum myri_endpt_type es_type;
  int endpoint = set_endpt->endpoint;

  es_type = set_endpt->endpt_type;

  is = mx_get_instance(unit);
  if (is == 0)
    return (ENODEV);

  mx_mutex_enter(&is->sync);

  if (mx_is_dead(is)) {
    mx_mutex_exit(&is->sync);
    mx_release_instance(is);
    return ENXIO;
  }

  es->es_type = es_type;

  eops = myri_endpt_getops(es_type);

  if (eops) {
    status = eops->alloc(es, is, set_endpt);
    if (status) {
      mx_mutex_exit(&is->sync);
      mx_release_instance(is);
      return status;
    }
  } 
  else {
    mx_mutex_exit(&is->sync);
    mx_release_instance(is);
    return EINVAL;
  }

  /* Common fields */
  es->ref_count = 0;
  es->eops = eops;
  myri_update_numa_config(es);
  mx_sync_init(&es->sync, is, endpoint, "es->sync");
  mx_sync_init(&es->close_sync, is, endpoint, "es->close_sync");
  mx_sync_init(&es->wait_sync, is, endpoint, "es->wait_sync");
  mx_sync_init(&es->vq_sync, is, endpoint, "es->vq_sync");
  mx_mutex_exit(&is->sync);

  /* Soft endpoints don't consume an actual mcp endpoint. */
  if (!(es->flags & MX_ES_SOFT)) {
    /* allocate the flow window */
    status = mx_alloc_zeroed_dma_page(is, (char **)&es->flow_window,
				      &es->flow_window_pin);
    if (status) {
      MX_WARN(("%s: endpt %d, flow window allocation failed\n", is->is_name, es->endpt));
      mx_common_close(es);
      return (status);
    }
    mx_reserve_page(es->flow_window);

    status = myri_init_common_mcp_copyblocks(es);
    if (status) {
      MX_WARN(("%s: endpt %d, can't initialize common copyblocks\n", is->is_name, es->endpt));
      mx_common_close(es);
      return (status);
    }

    status = mx_open_mcp_endpoint(es);
    if (status != 0) {
      MX_WARN(("%s: endpt %d, mx_open_mcp_endpoint failed\n", is->is_name, es->endpt));
      mx_common_close(es);
      return (status);
    }

    es->flags |= MX_ES_MCP_SESSION;
  }

  /* TODO: XXX Some endpoints don't yet support split open and enable.
   * As support for split open/enable is added to every endpoint type, this
   * flag will disappear */
  if (es->flags & MX_ES_EARLY_ENABLE) {
    status = es->eops->enable(es, arg);
    if (status != 0) {
      MX_WARN(("%s: endpt %d, early enable failed\n", is->is_name, es->endpt));
      mx_common_close(es);
      return (status);
    }
  }

  return 0;
}

mx_endpt_state_t *
mx_get_endpoint(mx_instance_state_t *is, int endpt)
{
  mx_endpt_state_t *es;

  if ((uint32_t) endpt >= is->es_count)
    return NULL;

  mx_mutex_enter(&is->sync);
  es = is->es[endpt];
  if (es == NULL)
    goto abort_with_is;
  mx_mutex_enter(&es->sync);
  if (es->flags & MX_ES_CLOSING) {
    mx_mutex_exit(&es->sync);
    es = NULL;
    goto abort_with_is;
  }
  es->ref_count++;
  mx_mutex_exit(&es->sync);

 abort_with_is:
  mx_mutex_exit(&is->sync);
  return es;
}

void
mx_put_endpoint(mx_endpt_state_t *es)
{
  mx_mutex_enter(&es->sync);
  if ((es->flags & MX_ES_CLOSING) && (es->ref_count == 1))
    mx_wake(&es->sync);
  es->ref_count--;
  mx_mutex_exit(&es->sync);
}

void
mx_common_close(mx_endpt_state_t *es)
{

  mx_instance_state_t *is;
  uint32_t dummy;
  int status = 0;
  int close_loop_cnt = 0;
  
  is = es->is;

  /* wait for all references to go away */
 again:
  mx_mutex_enter(&es->sync);
  if (es->flags & MX_ES_RECOVERY) {
    es->flags &= ~MX_ES_RECOVERY;
    mx_wake(&es->is->ready_for_recovery_sync);
  }
  es->flags |= MX_ES_CLOSING;
  if (es->ref_count != 0) {
    if (close_loop_cnt == 0)
      MX_INFO(("%s: closing endpoint %d with %d references, sleeping\n", 
              es->is->is_name, es->endpt, es->ref_count));
    mx_mutex_exit(&es->sync);
    mx_sleep(&es->sync, MX_CLOSE_WAIT, MX_SLEEP_NOINTR);
    close_loop_cnt++;
    goto again;
  }
  mx_mutex_exit(&es->sync);

  /* If this endpoint uses up an MCP session, close it */
  if (es->flags & MX_ES_MCP_SESSION) {
    status = mx_mcp_command(is, MCP_CMD_CLOSE_ENDPOINT, es->endpt, 0, 0, &dummy);

    if (status != 0)
      MX_WARN(("%s: Failed to close endpoint %d on mcp\n",
	       is->is_name, es->endpt));

    /* sleep for up to 10 seconds waiting for the mcp to close the
       endpoint before we begin to free resources which it may
       depend on
     */
    else {
      status = mx_sleep(&es->close_sync, MX_CLOSE_WAIT, MX_SLEEP_NOINTR);
      if (status) {
        MX_WARN(("%s: Timed out waiting for MCP to close endpoint %d\n",
	        is->is_name, es->endpt));
        mx_mark_board_dead(is, MX_DEAD_ENDPOINT_CLOSE_TIMEOUT, es->endpt);
      }
    }
  }

  /* Now go to close */
  es->eops->close(es);

  if (es->flow_window) {
    mx_unreserve_page(es->flow_window);
    mx_free_dma_page(is, (char **)&es->flow_window, &es->flow_window_pin);
    es->flow_window = NULL;
  }

  mx_sync_destroy(&es->close_sync);
  mx_sync_destroy(&es->wait_sync);
  mx_sync_destroy(&es->sync);
  mx_release_instance(is);
}

void
mx_dump_interrupt_ring(mx_instance_state_t *is)
{
  int slot;
  mcp_intr_t *desc;
  
  MX_PRINT(("--------- Dumping interrupt ring state ----- \n"));
  MX_PRINT(("currently expecting interrupts on slot=%d, seq=0x%x\n\n", 
	    is->intr.slot, is->intr.seqnum));
  /*  MX_PRINT(("REQ_ACK_0 = %x\n", mx_read_lanai_special_reg_u32(is, lx.ISR) & MX_LX_REQ_ACK_0));*/
  for (slot = 0; slot < MX_VPAGE_SIZE / sizeof(mcp_intr_t); slot++) {
    desc = &is->intr.ring[slot];
    MX_PRINT(("[%d]: seq=0x%x flags=%d valid=%d\n",
	      slot, ntohs(desc->seqnum), ntohl(desc->flags), desc->valid));
  }
}

static 
int
mx_common_ep_irq(mx_endpt_state_t *es, struct ep_irq *ei)
{
  mcp_intr_t *desc;
  uint32_t flags, ringsize = MX_VPAGE_SIZE / sizeof(mcp_intr_t);
  int status;

  desc = &ei->intr.ring[ei->intr.slot];
  flags = ntohl(desc->flags);

  if (!desc->valid) {
    MX_WARN(("Invalid endpoint interrupt descriptor !\n"));
    return 0;
  }

  if (!(es->flags & MX_ES_CLOSING)) {
    /* Unless the endpoint is waiting for a close on the host, process endpoint
     * specific interrupts other than closing and error interrupts */
    if (es->eops && es->eops->interrupt_ep)
      status = es->eops->interrupt_ep(es, flags);
    else
      status = EINVAL;

    if (status != 0)
      MX_WARN(("%s: error handling interrupt on endpt %d with flags=%#x\n",
              es->is->is_name, es->endpt, flags));
  }

  if (flags & MCP_INTR_ENDPT_ERROR) {
    MX_WARN(("%s:mcp detected endpoint error on endpoint 0x%x\n",
	     es->is->is_name, es->endpt));
    es->endpt_error = MX_WAIT_ENDPT_ERROR;
  }
  
  if (flags & MCP_INTR_ENDPT_CLOSED) {
    mx_wake(&es->close_sync);
  }
  
  /* bookkeeping */
  desc->valid = 0;
  ei->intr.slot = (ei->intr.slot + 1) % ringsize;

  return 1;
}

#ifdef MYRI_MSI
int
mx_common_msix_ep_irq(struct ep_irq *ei)
{
	mx_instance_state_t *is = ei->is;
	mx_endpt_state_t *es = *ei->ep;
	int handled;

	if (!es)
		return 0;

	/* process interrupt */
	handled = mx_common_ep_irq(es, ei);

	/* claim interrupt (as late as possible) */
	is->board_ops.claim_interrupt(is, 1, es->endpt);

	return handled;
}
#endif

int
mx_common_interrupt(mx_instance_state_t *is)
{
  mcp_intr_t *desc;
  uint32_t slot, flags, tmp;
  
  slot = is->intr.slot;
#if MAL_DEBUG
  is->intr.cnt++;
#endif  
  
  /* don't try to handle unless interrupt queues have been setup */
  if (is->intr.ring == 0)
    return 0;
  
  desc = &is->intr.ring[slot];
  
  /* PIO read to wait for desc, if enabled */
  if (myri_irq_sync && !desc->valid) {
    tmp = is->board_ops.read_irq_level(is);
  }
  
  /* check if we have a valid intr descriptor */
  if (!desc->valid) {
    is->intr.spurious++;
    return 0;
  }
  
  /* check interrupt seqnum */
  if (ntohs(desc->seqnum) != is->intr.seqnum) {
    uint16_t cmd;
    
    MX_WARN(("bad interrupt sequence number (slot %d, got %d, expected %d), "
	     "disabling interrupts\n", slot, ntohs(desc->seqnum), 
	     is->intr.seqnum));
    mx_dump_interrupt_ring(is);
    is->board_ops.disable_interrupt(is);
    /* in case we use MSI, disable DMA */
    mx_read_pci_config_16(is, MX_PCI_COMMAND, &cmd);
    mx_write_pci_config_16(is, MX_PCI_COMMAND, cmd & ~MX_PCI_COMMAND_MASTER);
    return 0;
  }
  
  /* claim the interrupt immediately if legacy */
  if (!is->msi_enabled && !is->msix_enabled)
    is->board_ops.claim_interrupt(is, 0, 0);
  
  /* process interrupt desc */
  flags = ntohl(desc->flags);
  
  /* if no interrupt flags, get out of here */
  if (!flags) {
    MX_WARN(("%s: slot %d: empty intr desc\n", is->is_name, slot));
    mx_dump_interrupt_ring(is);
    is->board_ops.disable_interrupt(is);
  }
  
  if (flags & MCP_INTR_ETHER_TX)
    mx_ether_tx_done(is, ntohl(desc->ether_tx_cnt));

  
  if (flags & MCP_INTR_ETHER_RX) {
    mx_ether_rx_done(is);
#ifdef HAVE_LRO
    mx_ether_lro_flush(is);
#endif
  }
  
  if (flags & MCP_INTR_ETHER_DOWN)
    mx_wake(&is->ether->down_sync);
  
  if (flags & MCP_INTR_LINK_STATE)
    {
      uint32_t old_state = is->link_state;
      uint32_t nic_state = ntohl(desc->link_state);
      int port = 0;
      const char *state_str;
      
      is->link_state &= ~((MCP_LINK_STATE_UP + MCP_LINK_STATE_MISMATCH) << port);
      is->link_state |= (nic_state << port);
      mx_ether_link_change_notify(is);
      if (nic_state == MCP_LINK_STATE_UP)
	state_str = "UP";
      else if (!(nic_state & MCP_LINK_STATE_UP))
	state_str = "DOWN";
      else
	state_str = "on Wrong network";
      if (old_state != is->link_state)
	MX_INFO(("%s: Link%d is %s\n",is->is_name, port, state_str));
    }
  
  if (flags & MCP_INTR_DMABENCH) {
    is->dmabench.cycles = ntohs(desc->dmabench_cpuc);
    is->dmabench.count = ntohs(desc->dmabench_cnt);
    mx_wake(&is->dmabench.wait_sync);
  }
  
  if (flags & MCP_INTR_LOGGING) {
    is->logging.size = ntohs(desc->logging_size);
    mx_wake(&is->logging.sync);
  }
  
#if 0
  /* FIXME: use INTR_PRINT for 10g */
  if (flags & MCP_INTR_PRINT)
    mx_lanai_print(is, ntohs(desc->print_size));
#endif
  
  if (flags & MCP_INTR_ENDPT) {
    mx_endpt_state_t *es;
    uint32_t i;
      
    for (i = 0; i < MCP_ENDPOINTS_CNT; i++) {
      if (desc->endpt_bitmap[i/8] == 0) {
	i += 7;
	continue;
      }
      if (desc->endpt_bitmap[i/8] & (1 << (i & 7))) {
	if (i >= myri_max_endpoints) {
	  MX_WARN(("%s: intr on invalid endpoint 0x%x\n", is->is_name, i));
	  break;
	}
	if ((es = is->es[i]) == 0) {
	  MX_WARN(("%s: intr on closed endpoint 0x%x\n", is->is_name, i));
	  break;
	}
	
	/* process endpoint interrupt desc */
	mx_common_ep_irq(es, &is->ep_irq[i]);
      }
    }
  }
  
  if (flags & MCP_INTR_RAW)
    myri_ops_kraw.interrupt_is(is, NULL);
  
  if (flags & MCP_INTR_SNF_WAKE)
    myri_ops_snf.interrupt_is(is, NULL);

  if (flags & MCP_INTR_TIMESYNC) {
    myri_clock_intr(is);
  }
  
  desc->valid = 0;
  is->intr.seqnum++;
  is->intr.slot = (slot + 1) & ((MX_VPAGE_SIZE / sizeof(mcp_intr_t)) - 1);
  
  /* claim interrupt as late as possible if MSI/MSIX */
  if (is->msi_enabled || is->msix_enabled)
    is->board_ops.claim_interrupt(is, 0, 0);
  
  return 1;
}

#if MAL_OS_WINDOWS
#define WINDOWS_BOUNDS_CHECK_COPY(len, limit) do       \
  if (len > limit)                                     \
    {                                                  \
      status = EINVAL;                                 \
      goto abort;                                      \
    }                                                  \
  while (0);
#else
#define WINDOWS_BOUNDS_CHECK_COPY(len, limit)
#endif


int
myri_sample_hwclock(mx_instance_state_t *is,
                    uint64_t *nic_ticks, uint64_t *host_nsecs,
                    uint64_t *host_nsecs_delay);

int
mx_endptless_ioctl(uint32_t cmd, const uaddr_t in, uint32_t privileged,
		   uint32_t is_kernel)
{
  int status = 0;
  mx_instance_state_t *is;
  uaddr_t out = in;

  privileged |= myri_security_disabled;

  switch (cmd) {

  case MYRI_GET_PORT_COUNT:
    {
      uint32_t board;

      status = mx_copyin(in, &board, sizeof(board), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      status = mx_copyout(&is->num_ports, out, sizeof(is->num_ports), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_LOGGING:
    {
      myri_get_logging_t x;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      status = mx_get_logging(is, x.size, (uaddr_t) x.buffer, is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_NIC_TICKS:
    {
      myri_get_nic_ticks_t x;

      if ((status = mx_copyin(in, &x, sizeof(x), is_kernel)))
	  goto abort_with_nothing;

      is = mx_get_instance((uint32_t)x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      status =  myri_sample_hwclock(is, &x.nic_ticks, &x.host_nsecs, &x.host_nsecs_delay);
      if (!status)
	status = mx_copyout(&x, out, sizeof(x), is_kernel);

      mx_release_instance(is);
    }
    break;

  case MYRI_GET_COUNTERS_CNT:
  case MYRI_GET_COUNTERS_STR:
  case MYRI_GET_COUNTERS_VAL:
  case MYRI_GET_COUNTERS_VAL64:
  case MYRI_GET_COUNTERS_CLASS:
    {
      uint32_t board, cnt;
      const char **counters;
      const uint32_t *cclass;
      
      status = mx_copyin(in, &board, sizeof(board), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      status = myri_counters(is->board_type, &counters, &cnt, &cclass);
      if (status) {
	MX_WARN (("Don't know which counters we have\n"));
	status = ENXIO;
	goto abort_with_is;
      }
      
      if (cmd == MYRI_GET_COUNTERS_CNT) {
	status = mx_copyout(&cnt, out, sizeof(cnt), is_kernel);
      } else if (cmd == MYRI_GET_COUNTERS_STR) {
	char tmp[MYRI_MAX_STR_LEN];
	int i;
	for (i = 0; i < cnt; ++i, out += MYRI_MAX_STR_LEN) {
	  strncpy(tmp, counters[i], MYRI_MAX_STR_LEN-1); 
	  tmp[MYRI_MAX_STR_LEN-1] = '\0';
	  status = mx_copyout(tmp, out, MYRI_MAX_STR_LEN, is_kernel);
	  if (status)
	    goto abort_with_is;
	}
      } else if (cmd == MYRI_GET_COUNTERS_CLASS) {
        void *cclass_ptr = (void *) cclass;
        status = mx_copyout(cclass_ptr, out, cnt * sizeof(*cclass), is_kernel);
      } else if (cmd == MYRI_GET_COUNTERS_VAL64) {
	uint64_t *counters = mx_kmalloc(is->counters.cnt * sizeof(uint64_t), 
					MX_MZERO|MX_WAITOK);
	if (counters == NULL) {
	  status = ENOMEM;
	} else {
	  status = myri_counters64_get(is, counters, is->counters.cnt, 
				       MYRI_COUNTERS_GET_ALL);
	  if (status == 0)
	    status = mx_copyout(counters, out, cnt * 8, is_kernel);
	  mx_kfree(counters);
	}
      } else if (cmd == MYRI_GET_COUNTERS_VAL) {
	status = mx_copyout(is->counters.addr, out, cnt * 4, is_kernel);
      } else
	status = ENOTTY;
      
      mx_release_instance(is);
    }
    break;
    
  case MYRI_GET_COUNTERS_IRQ:
    {
      uint32_t board;
      myri_get_counters_irq_t irq;
      
      status = mx_copyin(in, &board, sizeof(board), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      
      irq.count = is->intr.cnt;
      irq.events = is->intr.seqnum;
      irq.spurious = is->intr.spurious;
      status = mx_copyout(&irq, out, sizeof(irq), is_kernel);
      mx_release_instance(is);
    }
    break;
    
  case MYRI_GET_COUNTERS_KB:
    {
      myri_get_counters_kb_t x;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      
      x.num_ports = 1;
      x.send_kb_0 = ((uint32_t *)is->counters.addr)[2];
      x.recv_kb_0 = ((uint32_t *)is->counters.addr)[3];

      if ((MAL_NUM_PORTS(is->board_type)) > 1) {
	x.num_ports = 2;
	x.recv_kb_0 = ((uint32_t *)is->counters.addr)[4];
	x.send_kb_1 = ((uint32_t *)is->counters.addr)[3];
	x.recv_kb_1 = ((uint32_t *)is->counters.addr)[5];
      }
      
      status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_CLEAR_COUNTERS:
    {
      uint32_t board_num;
      
      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      
      if (!privileged) {
	status = EPERM;
	goto abort_with_is;
      }
      
      /* We clear the accumulation counters on the host prior to the mcp
       * counters. */
      myri_counters64_clear(is);
      status = mx_mcp_command(is, MCP_CMD_RESET_COUNTERS, 0, 0, 0, &board_num);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_BOARD_COUNT: 
    status = mx_copyout(&mx_num_instances, out, sizeof(mx_num_instances), is_kernel);
    break;

  case MYRI_GET_BOARD_MAX:
    status = mx_copyout(&myri_max_instance, out, sizeof(myri_max_instance), is_kernel);
    break;

  case MYRI_GET_ENDPT_MAX:
    {
      uint32_t num_endpoints = myri_max_endpoints;
      /* Return total(hard+soft) endpoints available so myri_endpoint_info
       * can display all endpoints as set in mx_instance.c.
       * only TX endpoints are allocated as hard(MCP) endpoints so
       * myri_snf_ops.c must divide by 6 to get hard endpoints after
       * ioctl MYRI_GET_ENDPT_MAX
       */
      num_endpoints = 6*myri_max_endpoints;
      status = mx_copyout(&num_endpoints, out, sizeof(num_endpoints), is_kernel);
    }
    break;
    
  case MYRI_GET_NIC_ID:
    {
      myri_get_nic_id_t x;
      int i;
      
      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      x.nic_id = 0;
      for (i = 0; i < 6; i++)
	x.nic_id = (x.nic_id << 8) + is->mac_addr[i];

      status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
    }
    break;

  /* Still shared with dbl */
  case MX_GET_MAX_SEND_HANDLES:
    status = mx_copyout(&myri_mx_max_send_handles, out, 
			sizeof(myri_mx_max_send_handles), is_kernel);
    break;

  case MX_NIC_ID_TO_BOARD_NUM:
    {
      mx_nic_id_to_board_num_t x;
      int i;

      if ((status = mx_copyin(in, &x, sizeof(x), is_kernel))) {
	goto abort_with_nothing;
      }
      mx_mutex_enter(&mx_global_mutex);
      for (i = 0; i < myri_max_instance; ++i) {
	is = mx_instances[i];
	if (!is)
	  continue;
	if ((((uint64_t)is->mac_addr[0] << 40) |
	     ((uint64_t)is->mac_addr[1] << 32) |
	     ((uint64_t)is->mac_addr[2] << 24) |
	     ((uint64_t)is->mac_addr[3] << 16) |
	     ((uint64_t)is->mac_addr[4] <<  8) |
	     ((uint64_t)is->mac_addr[5])) == x.nic_id) {
	  break;
	}
      }
      if (i >= myri_max_instance) {
	status = ENODEV;
	mx_mutex_exit(&mx_global_mutex);
	goto abort_with_nothing;
      }
      mx_mutex_exit(&mx_global_mutex);
      x.board_number = i;
      status = mx_copyout(&x, out, sizeof(x), is_kernel);
    }
    break;

  case MYRI_RAW_GET_PARAMS:
    {
      myri_raw_params_t x;
      
      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;
      is = mx_get_instance(x.board_number);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      x.raw_mtu = MCP_RAW_MTU;
      x.raw_max_route = MYRI_RAW_MAXROUTE;
      x.raw_num_tx_bufs = MCP_RAW_TX_CNT;
      status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MX_GET_SERIAL_NUMBER:
    {
      uint32_t board_number;
      uint32_t serial;

      if ((status = mx_copyin(in, &board_number,
                              sizeof(board_number), is_kernel))) {
        goto abort_with_nothing;
      }
      is = mx_get_instance(board_number);
      if (!is) {
        status = ENODEV;
        goto abort_with_nothing;
      }
      serial = myri_atoi(is->lanai.serial);
      status = mx_copyout(&serial, out, sizeof(serial), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_ENDPT_OPENER:
    {
      myri_get_endpt_opener_t x;
      mx_endpt_state_t *es = 0;
      int raw = 0;
      int ether = 0;

      if ((status = mx_copyin(in, &x, sizeof(x), is_kernel)))
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }

      switch ((int) x.endpt) {
      case -2: 
	ether = 1;
	break;
      case -1:
	raw = 1;
	break;
      default:
	if (x.endpt >= is->es_count) {
	  status = EINVAL;
	  goto abort_with_is;
	}
	break;
      }
      mx_mutex_enter(&is->sync);
      if (raw) {
#if MYRI_ENABLE_KRAW
	es = is->raw.es;
#else
	es = NULL;
#endif
      } else if (ether)
	es = NULL;
      else
	es = is->es[x.endpt];

      x.closed = 0;
      if (es) {
	bcopy(&es->opener, &x.opener, sizeof(x.opener));
      } else if (!ether || !is->ether_is_open) {
	  x.closed = 1;
      }
      mx_mutex_exit(&is->sync);
      status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MX_GET_CACHELINE_SIZE:
    {
      status = mx_copyout(&mx_cacheline_size, out, sizeof(mx_cacheline_size), is_kernel);
    }
    break;

  case MX_RUN_DMABENCH:
    {
      mx_dmabench_t x;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board_number);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      
      status = mx_run_dmabench(is, &x);
      if (!status)
	status = mx_copyout(&x, out, sizeof(x), is_kernel);

      mx_release_instance(is);
    }
    break;
    
    
  case MX_GET_BOARD_STATUS:
    {
      uint32_t board_num, board_status;

      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }

      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      board_status = is->saved_state.reason;
      status = mx_copyout(&board_status, out, sizeof(board_status), is_kernel);
      mx_release_instance(is);
    }
    break;
  case MX_GET_CPU_FREQ:
    {
      uint32_t board_num;

      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }

      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      status = mx_copyout(&is->cpu_freq, out, sizeof(is->cpu_freq), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MX_GET_PCI_FREQ:
    {
      uint32_t board_num;

      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }

      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      status = mx_copyout(&is->pci_freq, out, sizeof(is->pci_freq), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_INTR_COAL:
  case MYRI_SET_INTR_COAL:
    {
      myri_intr_coal_t x;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      
      if (cmd == MYRI_SET_INTR_COAL) {
	if (!privileged) {
	  status = EPERM;
	  goto abort_with_is;
	}
	if (x.delay > 1000) {
	  status = ERANGE;
	  goto abort_with_is;
	}
	MCP_SETVAL(is, intr_coal_delay, x.delay);
      } else
	x.delay = MCP_GETVAL(is, intr_coal_delay);

      status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
    }
    break;
    
  case MYRI_GET_TIMESOURCE_STATE:
    {
      uint32_t board_num;
      uint32_t ts_state;

      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }
      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      ts_state = (uint32_t) is->clk_ts_state;
      status = mx_copyout(&ts_state, out, sizeof(ts_state), is_kernel);
      mx_release_instance(is);
    }
    break;

  case MX_GET_LINK_STATE:
    {
      uint32_t board_num;

      status = mx_copyin(in, &board_num, sizeof(board_num), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }

      is = mx_get_instance(board_num);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      status = mx_copyout(&is->link_state, out, sizeof(is->link_state), 
			  is_kernel);
      mx_release_instance(is);
    }
    break;
    
  case MX_GET_SERIAL_STRING:
  case MX_GET_PRODUCT_CODE:
  case MX_GET_PART_NUMBER:
    {
      mx_get_eeprom_string_t x;
      char *unknown = "unknown";
      char *ptr;
      int len;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status) {
	goto abort_with_nothing;
      }

      is = mx_get_instance(x.board_number);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      if (cmd == MX_GET_SERIAL_STRING)
	ptr = is->lanai.serial;
      else if (cmd == MX_GET_PRODUCT_CODE)
	ptr = is->lanai.product_code;
      else
	ptr = is->lanai.part_number;

      if (ptr == NULL) {
	ptr = unknown;
      }

      len = (int)(strlen(ptr) + 1);
      if (len > MYRI_MAX_STR_LEN)
	len = MYRI_MAX_STR_LEN;

      status = mx_copyout(ptr, (uaddr_t)x.buffer, len, is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_GET_VERSION:
    {
      myri_get_version_t x;
      x.driver_api_magic = MYRI_DRIVER_API_MAGIC;
      strncpy(x.version_str,MYRI_VERSION_STR, sizeof(x.version_str));
      strncpy(x.build_str,MYRI_BUILD_STR, sizeof(x.build_str));
      x.version_str[sizeof(x.version_str) - 1] = 0;
      x.build_str[sizeof(x.build_str) - 1] = 0;
      status = mx_copyout(&x, out, sizeof x, is_kernel);
    }
    break;

  case MYRI_GET_SRAM_SIZE:
  case MYRI_GET_BOARD_TYPE:
  case MYRI_GET_BOARD_NUMA_NODE:
    {
      uint32_t x;
      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      x = ((cmd == MYRI_GET_SRAM_SIZE)  ? is->sram_size :
	   (cmd == MYRI_GET_BOARD_TYPE) ? is->board_type :
	   (cmd == MYRI_GET_BOARD_NUMA_NODE) ? mx_get_board_numa_node(is) :
	   0);
      status = mx_copyout(&x, out, sizeof x, is_kernel);
      mx_release_instance(is);
    }
    break;

  case MYRI_PCI_CFG_READ:
  case MYRI_PCI_CFG_WRITE:
    {
      uint8_t val8;
      uint16_t val16;
      uint32_t val32;
      myri_pci_cfg_t x;

      if (!privileged) {
	status = EPERM;
	goto abort_with_nothing;
      }
      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
	goto abort_with_nothing;

      is = mx_get_instance(x.board);
      if (!is) {
	status = ENODEV;
	goto abort_with_nothing;
      }
      if (cmd == MYRI_PCI_CFG_READ) {
	switch (x.width) {
	case 1:
	  status = mx_read_pci_config_8(is, x.offset, &val8);
	  x.val = val8;
	  break;
	case 2:
	  status = mx_read_pci_config_16(is, x.offset, &val16);
	  x.val = val16;
	  break;
	case 4:
	  status = mx_read_pci_config_32(is, x.offset, &val32);
	  x.val = val32;
	  break;
	default:
	  status = EINVAL;
	}
      } else {
	switch (x.width) {
	case 1:
	  status = mx_write_pci_config_8(is, x.offset, x.val);
	  break;
	case 2:
	  status = mx_write_pci_config_16(is, x.offset, x.val);
	  break;
	case 4:
	  status = mx_write_pci_config_32(is, x.offset, x.val);
	  break;
	default:
	  status = EINVAL;
	}
      }
      if (!status)
	status = mx_copyout(&x, out, sizeof x, is_kernel);
      mx_release_instance(is);
    }
    break;
#if MYRI_ENABLE_PTP
  case MYRI_PTP_RX_AVAIL:
  case MYRI_PTP_GET_RX_MSG:
  case MYRI_PTP_STORE_RX_MSG:
  case MYRI_PTP_START_STOP:
  case MYRI_PTP_PUT_TX_TIMESTAMP:
  case MYRI_PTP_GET_TX_TIMESTAMP:
    {
      status = myri_ptp_ioctl(cmd, in, is_kernel);
    }
    break;
#endif

  case MYRI_CHECK_LICENSE:
    {
      uint32_t x;

      status = mx_copyin(in, &x, sizeof(x), is_kernel);
      if (status)
        goto abort_with_nothing;

      is = mx_get_instance(x);
      if (!is) {
        status = ENODEV;
        goto abort_with_nothing;
      }

      status = myri_check_license(is, &x);
      if (status == 0)
        status = mx_copyout(&x, out, sizeof(x), is_kernel);
      mx_release_instance(is);
      break;
    }

  case MYRI_GET_DRIVER_NAME:
    {
      int len;
      len = (int)(strlen(MYRI_DRIVER_STR) + 1);
      if (len >= MYRI_MAX_STR_LEN) {
	len = MYRI_MAX_STR_LEN;
      }
      mx_copyout(MYRI_DRIVER_STR, out, len, is_kernel); 
      break;
    }

  default:
    status = ENOTTY;
  }
  
  return status;

 abort_with_is:
  mx_release_instance(is);
 abort_with_nothing:
  return status;
}

/*
 * Common copyblocks are the set of copyblocks that are common to all
 * MCP-enabled endpoints.  The offsets here must be kept in sync with
 * associated mmap handling in mx_mmap_off_to_kva().
 */
int
myri_endpt_get_common_copyblocks(const mx_endpt_state_t *es, mx_get_copyblock_t *desc,
                                 uint32_t *offp)
{
  uint32_t off = 0;

  /* First the A mappings (exist at open time) */
  desc->user_mmapped_sram_offset = off;
  desc->user_mmapped_sram_len = es->user_mmapped_sram.size;
  off += desc->user_mmapped_sram_len;

  desc->user_mmapped_zreq_offset = off;
  desc->user_mmapped_zreq_len = es->user_mmapped_zreq.size;
  off += desc->user_mmapped_zreq_len;

  desc->kernel_window_offset = off;
  desc->kernel_window_len = (es->is->kernel_window ? MX_PAGE_SIZE : 0);
  off += desc->kernel_window_len;

  desc->flow_window_offset = off;
  desc->flow_window_len = (es->flow_window ? MX_PAGE_SIZE : 0);
  off += desc->flow_window_len;

  desc->send_compcnt_offset = off;
  desc->send_compcnt_len = (es->is->tx_done.addr ? MX_PAGE_SIZE : 0);
  off += desc->send_compcnt_len;

  /* Then the B mappings (exist at enable time) */
  desc->sendq_offset = off;
  desc->sendq_len = es->sendq.size;
  off += desc->sendq_len;

  desc->recvq_offset = off;
  desc->recvq_len = (uint32_t)es->recvq.size;
  off += desc->recvq_len;

  desc->eventq_offset = off;
  desc->eventq_len = es->eventq.size;
  off += desc->eventq_len;

  if (offp)
    *offp = off;

  return 0;
}

int
mx_common_ioctl(mx_endpt_state_t *es, uint32_t cmd, const uaddr_t in)
{
  uaddr_t out = in;
  int status = 0;

  if (mx_is_dead(es->is)) {
    /* if the board is dead, return an error to the application...
     * except in case of parity since the recovery is notified to the
     * application through the wait_status or get_board_status
     * (other ioctls can still be processed safely since they do not
     * touch the board for anything that won't be redone a afterwards)
     */
    if (!(es->is->flags & MX_PARITY_RECOVERY)) {
      MX_WARN(("firmware dead on board %d, ignoring ioctl\n", es->is->id));
      return EIO;
    }
  }

  if (es->eops) {
    status = es->eops->ioctl(es, cmd, in);
    if (status != ENOTTY)
      return status;
  }

  switch (cmd) {

  case MX_GET_COPYBLOCKS:
    {
      mx_get_copyblock_t desc;
      /* Endpoint must be mcp-enabled before copyblocks can be retrieved.
       * mx_mmap_off_to_kva() refers to A and B mappings which are respectively
       * available after endpoint open and after endpoint enable.  Since this
       * common_get_copyblock() is inherited from the mx days where A and B
       * were allocated simultaneously, it can only be called once the endpoint
       * is enabled. */
      if (!(es->flags & MX_ES_MCP_ENABLED)) {
        status = EINVAL;
        break;
      }

      status = myri_endpt_get_common_copyblocks(es, &desc, NULL);
      if (status)
        break;
      status = mx_copyout(&desc, out, sizeof(desc), es->is_kernel);
    }
    break;

#if MAL_OS_UDRV
  case MYRI_UDRV_DOORBELL:
    {
      myri_udrv_doorbell_t x;
      mx_copyblock_t *cb;

      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
	goto abort;

      cb = (MAL_IS_ZE_BOARD(es->is->board_type) ?
	    &es->user_mmapped_zreq : &es->user_mmapped_sram);
      mal_always_assert(x.offset >= 0 && x.len <= 64 &&
			x.offset + x.len <= cb->size);
      mx_pio_memcpy((char *)cb->addr + x.offset, x.data, x.len, 0);
    }
    break;
#endif

  case MYRI_ENABLE_ENDPOINT:
    if (es->eops == NULL)
      status = EINVAL;
    else
      status = es->eops->enable(es, 0);
    break;

  case MYRI_SET_COPYBLOCK:
  {
    myri_set_copyblock_t x;
    status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
    if (status)
      goto abort;
    mx_mutex_enter(&es->sync);
    status = myri_common_set_copyblock(es, &x);
    mx_mutex_exit(&es->sync);
    if (!status)
      mx_copyout(&x, out, sizeof(x), es->is_kernel);
  }
  break;

#if MAL_OS_LINUX || MAL_OS_FREEBSD || MAL_OS_WINDOWS
  case MYRI_SOFT_RX:
    {
      myri_soft_rx_t x;
      status = mx_copyin(in, &x, sizeof(x), es->is_kernel);
      if (status)
	goto abort;
      status = myri_ether_soft_rx(es->is, &x);
    }
    break;
#endif /* MAL_OS_LINUX || MAL_OS_FREEBSD */

  default:
    status = ENOTTY;
  }

 abort:
  return status;
}

int
mx_instance_status_string(int unit, char **str, int *len)
{
  char *c;
  char mcp_status_string[64];
  mx_instance_state_t *is;
  uint32_t mcp_status;
  int ret = 0;

  is = mx_get_instance(unit);
  if (!is) {
    ret = ENODEV;
    goto abort_with_nothing;
  }

  c = mx_kmalloc(512, MX_MZERO|MX_WAITOK);
  if (c == NULL) {
    ret = ENOMEM;
    goto abort_with_is;
  }

  if (mx_is_dead(is)) {
    mcp_status = MCP_GETVAL(is, mcp_status);
    sprintf(mcp_status_string, "not responding (%d,%d,0x%x)", 
	    is->saved_state.reason, is->saved_state.arg, mcp_status);
  } else {
    sprintf(mcp_status_string, "running)");
  }

  sprintf(c, "%s: Driver %s (0x%x), status:%s, uptime %ld, %d ports\n",
	  is->is_name, MYRI_DRIVER_STR, MYRI_DRIVER_API_MAGIC, mcp_status_string, 
	  (long)ntohl(*(uint32_t *)is->counters.addr), is->num_ports);

  *len = (int)strlen(c);
  *str = c;
  mx_release_instance(is);
  return 0;


 abort_with_is:
  mx_release_instance(is);
 abort_with_nothing:
  return ret;
}

int
mx_direct_get_common (mx_shm_seg_t *dst_segs, uint32_t dst_nsegs,
		      void * src_space, mx_shm_seg_t *src_segs, uint32_t src_nsegs,
		      uint32_t length)
{
  mx_shm_seg_t *current_src_seg, *current_dst_seg;
  uint32_t current_src_index, current_dst_index;
  uint32_t current_src_offset, current_dst_offset;
  int status;

  /* start with first src segment */
  current_src_index = 0;
  current_src_seg = &src_segs[0];
  current_src_offset = 0;
  /* start with first dst segment */
  current_dst_index = 0;
  current_dst_seg = &dst_segs[0];
  current_dst_offset = 0;

  while (current_src_index < src_nsegs && current_dst_index < dst_nsegs
	 && length > 0) {
    uint32_t chunk = length;
    /* copy until one segment ends */
    chunk = chunk < current_src_seg->len - current_src_offset
      ? chunk : current_src_seg->len - current_src_offset;
    chunk = chunk < current_dst_seg->len - current_dst_offset
      ? chunk : current_dst_seg->len - current_dst_offset;

    if (chunk) {
      status = mx_arch_copy_user_to_user((uaddr_t)(current_dst_seg->vaddr + current_dst_offset),
					 (uaddr_t)(current_src_seg->vaddr + current_src_offset),
					 src_space,
					 chunk);
      if (status)
	return status;
    }

    /* update src status */
    current_src_offset += chunk;
    if (current_src_offset == current_src_seg->len) {
      current_src_seg++;
      current_src_index++;
      current_src_offset = 0;
    }
    /* update dst status */
    current_dst_offset += chunk;
    if (current_dst_offset == current_dst_seg->len) {
      current_dst_seg++;
      current_dst_index++;
      current_dst_offset = 0;
    }

    length -= chunk;
  }

  return 0;
}

static
int
mx_isxdigit(int d)
{
  if (d >= '0' && d <= '9')
    return 1;
  if (d >= 'a' && d <= 'f')
    return 1;
  if (d >= 'A' && d <= 'F')
    return 1;
  return 0;
}

static
int
mx_isdelim(int d)
{
  /* all valid delimiters except ':' which is needed for macaddr */
  switch (d) {
    case ',':
    case ';':
    case '!':
    case '#':
    case '/':
    case '%':
    case ' ':
    case '\t':
    case '.':
      return 1;
    default:
      return 0;
  }
}

int
mx_maccmp(const char *mac_a, const char *list_of_mac_b_substr)
{
  int ca, cb;
  int lena, lenaa;
  int matched;
  const char *a = mac_a;
  const char *b = list_of_mac_b_substr;
  const char *be = b;
  const char *bb;

  lenaa = (int)strlen(mac_a);
  if (!lenaa) /* Need a valid input mac_a */
    return 0;

  while (*b != '\0') {
    matched = 0;
    while (mx_isdelim(*b)) 
      b++;
    if (*b == '\0')
      return 0; /* nothing */
    bb = be = b;
    while (*be != '\0' && !mx_isdelim(*be)) {
      if (mx_isxdigit(*be) || *be == ':')
        be++;
      else {
        return -1; /* invalid character */
      }
    }
    b = be; /* save for next round */
    lena = lenaa;
    a = mac_a + lena - 1;
    while (be != bb && lena) {
      be--;
      ca = *a;
      cb = *be;
      if (ca >= 'a' && ca <= 'z')
        ca -= 'a' - 'A';
      if (cb >= 'a' && cb <= 'z')
        cb -= 'a' - 'A';
      if (ca != cb) {
        matched = 0;
        break;
      }
      else
        matched = 1;
      a--;
      lena--;
    }
    if (matched)
      return 1;
  }
  return 0;
}

/*
 * Clock synchronization between NIC and host
 */

static uint64_t myri_read_hwclock(mx_instance_state_t *is);

/* values returned by myri_get_timesync in the tsstate variable */
enum myri_timesync_status{
  MYRI_TIMESYNC_UNSUPPORTED, /* not a pps nic? or coldfire not responding */
  MYRI_TIMESYNC_BAD,         /* coldfire doesn't have irig signal */
  MYRI_TIMESYNC_GOOD,        /* everything is good */
};

/* get a nic_ticks to timesource nanos translation from the lanai */
/* returns 0 if we successfully obtained timesync information , -1 otherwise.
   tsstate is defined in the enum above. */

static int
myri_get_timesync(mx_instance_state_t *is, uint64_t *tsnanos, uint64_t *nticks,
		  enum myri_timesync_status *tsstate)
{
  int iters = 0;
  uint8_t ts_status;
  uint64_t ts_nanos;
  uint64_t nic_ticks;
  int fail = 0;
  unsigned long flags;
  uint32_t seqno;

  flags = 0;

  mx_spin_lock_irqsave(&is->clk_spinlock, flags);

  seqno = ++is->clk_hwclock_seqno;
  /* the timesync doorbell the next uint32 after the clk update db. */
  MX_PIO_WRITE((uint32_t *)(is->lanai.sram + is->clk_db_offset + 4), seqno);
  MAL_STBAR();
  while (is->clk_update->nticks_seqnum != seqno) {
    if (iters++ > 1000000) {
      MX_PRINT_ONCE(("%d: Missed timesync request response!\n", is->id));
      fail = 1;
      break;
    }
    MAL_READBAR();
  }
  
  nic_ticks = ((uint64_t) ntohl(is->clk_update->nticks_high) << 32) +
    ((uint64_t) ntohl(is->clk_update->nticks_low));
  ts_nanos = ((uint64_t) ntohl(is->clk_update->ts_secs) * 1000000000) +
    ((uint64_t) ntohl(is->clk_update->ts_nanos));
  ts_status = is->clk_update->ts_status;

  mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);

  if (fail) {
    return -1;
  }

  *tsnanos = ts_nanos;
  *nticks = nic_ticks;
  if (ts_status == 0x07){
    *tsstate = MYRI_TIMESYNC_GOOD;
  } else {
    if ( ts_status == 0xff ) {
      MX_INFO(("%d: Timecode features not working\n", is->id));
      *tsstate = MYRI_TIMESYNC_UNSUPPORTED;
    } else {
      /*MX_INFO(("%d: Timecode status is not good: %2x\n", is->id, ts_status));*/
      *tsstate = MYRI_TIMESYNC_BAD;
    }
  }


  return 0;
}

int
myri_sample_hwclock(mx_instance_state_t *is,
                    uint64_t *nic_ticks, uint64_t *host_nsecs,
                    uint64_t *host_nsecs_delay)
{
  uint64_t tsB_nsecs, tsE_nsecs;
  struct timespec tsB, tsE;
  unsigned long flags;
  int attempts = 10;

  flags = 0; 

  if (mx_is_dead(is))
    return -EIO;

try_again:
  mx_spin_lock_irqsave(&is->clk_spinlock, flags);
  mx_getnstimeofday(&tsB);
  *nic_ticks = myri_read_hwclock(is);
  mx_getnstimeofday(&tsE);
  mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);

  tsB_nsecs = mx_timespec_to_ns(&tsB);
  tsE_nsecs = mx_timespec_to_ns(&tsE);

  if (tsB_nsecs > tsE_nsecs) {
    static int once = 0;
    if (attempts-- > 0)
      goto try_again;
    if (!once) {
      MX_WARN(("System clock going backwards? Can't sync host/NIC clocks\n"));
      once = 1;
    }
    return -EIO;
  }

  *host_nsecs = (tsB_nsecs + tsE_nsecs) / 2;
  *host_nsecs_delay = tsE_nsecs - tsB_nsecs;
  return 0;
}

int
myri_clock_init(mx_instance_state_t *is)
{
  uint32_t offset;
  uint64_t host_nsecs_delay;
  int status = 0;
  myri_clksync_nticks_t clk_nticks;

  is->clk_ts_state = MYRI_CLK_NOT_INITIALIZED;

  status = mx_mcp_command(is, MCP_CMD_GET_HWCLOCK_OFFSET,
			  0, 0, 0, &offset);
  if (status) /* no HW clock on this firmware */
    return status;

  is->clk_db_offset = offset;
  mx_spin_lock_init(&is->clk_spinlock, is, -1, "clock spinlock");

  /* We use the last bytes of the counters page to read clock updates.
   * There is a check in the counters code to ensure this never gets clobbered
   */
  mal_assert(is->counters.addr);
  is->clk_update = (mcp_hwclock_update_t *)
    (is->counters.addr + MX_VPAGE_SIZE - sizeof(mcp_hwclock_update_t));

  /* check if we have a timesource capable nic */
  if (is->timesource_enable == 1) {
    /* the check for PPS can probably be removed once all 'beta' timesync
       nics have been returned to myricom as we think they are the only
       ones that got into the hands of customers with the PPS product code*/
    if ((strstr(is->lanai.product_code, "PPS") != NULL) ||
	(strstr(is->lanai.product_code, "SYNC") != NULL)) {
      is->timesource_enable = 1;
    } else {
      is->timesource_enable = 0;
    }
  }

  /* lanai freq is (156.25Mhz * 14 / 6) or about 364.58333... .
     So the conversion from cpuc to nanos is 96/35 */

  if (is->timesource_enable == 1) {
    /* the actual work of synching with the timecode is done in the
       myri_clock_ts_process */
    MX_INFO(("Enabling timesource timestamping.\n"));
    switch (myri_timesource_loss_mode) {
    case MYRI_TIMESOURCE_LOSS_ZERO:
      MX_INFO(("Timestamps are zeroed on timesource loss.\n"));
      break;
    case MYRI_TIMESOURCE_LOSS_FREERUN:
      MX_INFO(("Timestamps freerun on timesource loss.\n"));
      break;
    default:
      MX_INFO(("Unsupported timesource loss mode, using freerun.\n"));
      myri_timesource_loss_mode = MYRI_TIMESOURCE_LOSS_FREERUN;
      break;
    }
    clk_nticks.ticks = 0;
    clk_nticks.nsecs = 0;
    clk_nticks.ticks_mask = 0;
    clk_nticks.nt_numerator = 0; 
    clk_nticks.nt_denominator = 1;
    is->clkp_latest.error_nsecs = 0;
    is->clkp_latest.hostnic_drift_nsecs = 0;
    is->clkp_latest.multiplier = 0;
    is->clkp_latest.nsecs_base = clk_nticks.nsecs;
    is->kernel_window->clksync.seqnum = 0;
    is->kernel_window->clksync.nticks = clk_nticks;
    is->kernel_window->clksync.clkp = is->clkp_latest;
    is->clk_ts_state = MYRI_CLK_TS_INIT_PHASE1;
  } else {
    /* we aren't running on a timecode capable nic */
    MX_INFO(("Enabling host timestamping.\n"));
    clk_nticks.ticks_mask = -1;
    clk_nticks.nt_numerator = 96;  /* 96/35 is the factor of cpuc to nanos */
    clk_nticks.nt_denominator = 35;
    
    if ((status = myri_sample_hwclock(is, &clk_nticks.ticks, &clk_nticks.nsecs,
				      &host_nsecs_delay)))
      return status;
    
    /* Start off by assuming no drift, and default to syncing every second. */
    is->clk_sync_init_cnt = 1;
    is->clk_sync_init_nsecs = clk_nticks.nsecs;
    is->clkp_latest.error_nsecs = 0;
    is->clkp_latest.hostnic_drift_nsecs = 0;
    is->clkp_latest.multiplier = 0;
    is->clkp_latest.nsecs_base = clk_nticks.nsecs;
    
    is->kernel_window->clksync.seqnum = 0;
    is->kernel_window->clksync.nticks = clk_nticks;
    is->kernel_window->clksync.clkp = is->clkp_latest;
    is->clk_ts_state = MYRI_CLK_HOST_INITIALIZED;
  }
  return 0;
}

void
myri_clock_intr(mx_instance_state_t *is)
{
  myri_clksync_nticks_t clk_nticks;
  unsigned long flags;
  uint64_t nic_nsecs, cycles;
  uint64_t secs_now;
  int rc;
  enum myri_timesync_status ts_status;
  enum myri_timesource_state state;

  flags = 0; 

  if (is->timesource_enable != 1 || is->clk_ts_state == MYRI_CLK_NOT_INITIALIZED ||
      is->clk_ts_state == MYRI_CLK_TS_ERROR) {
    /* not using timesource, or the clock has not yet been initialized, or
       we encountered a fatal error with the timestamp so ignore the interrupt */
    return;
  }

  /* need to check the reason for the interrupt */
  rc = myri_get_timesync(is, &nic_nsecs, &cycles, &ts_status);

  if (rc  != 0 || ts_status == MYRI_TIMESYNC_UNSUPPORTED) {
    state = MYRI_CLK_TS_ERROR;
  } else if (ts_status == MYRI_TIMESYNC_BAD) {
    secs_now = myri_timesync_ticks2nsecs(&is->kernel_window->clksync, cycles);
    MX_INFO(("%d: Lost communication with timesource at %u seconds.\n", 
	     is->id, (uint32_t)(secs_now / 1000000000)));
    state = MYRI_CLK_TS_NO_TIMECODE;
  } else { /*good*/
    MX_INFO(("%d: Regained communication with timesource.\n", is->id));
    state = MYRI_CLK_TS_INIT_PHASE1;
  }

  if (myri_timesource_loss_mode == MYRI_TIMESOURCE_LOSS_ZERO) {
    clk_nticks.ticks = 0;
    clk_nticks.nsecs = 0;
    clk_nticks.ticks_mask = 0;
    clk_nticks.nt_numerator = 0;
    clk_nticks.nt_denominator = 1;
  }

  /* write the values out */
  mx_spin_lock_irqsave(&is->clk_spinlock, flags);
  is->kernel_window->clksync.seqnum++;
  MAL_WRITEBAR();
  is->clk_ts_state = state;
  is->timesource_data.attempts = 0;
  if (myri_timesource_loss_mode == MYRI_TIMESOURCE_LOSS_ZERO) {
    is->kernel_window->clksync.nticks = clk_nticks;
  }
  MAL_WRITEBAR();
  is->kernel_window->clksync.seqnum++;
  mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);

}

void
myri_clock_destroy(mx_instance_state_t *is)
{
  if (is->clk_ts_state == MYRI_CLK_NOT_INITIALIZED)
    return;
  mx_spin_lock_destroy(&is->clk_spinlock);
  is->clk_ts_state = MYRI_CLK_NOT_INITIALIZED;
}

/*
 * Must hold the clk_spinlock as there can only be one HWCLOCK request at a
 * time
 */
static uint64_t
myri_read_hwclock(mx_instance_state_t *is)
{
  uint64_t nic_ticks;
  uint32_t seqno;
  int iters = 0;

  seqno = ++is->clk_hwclock_seqno;
  MX_PIO_WRITE((uint32_t *)(is->lanai.sram + is->clk_db_offset), seqno);
  MAL_STBAR();
  while (is->clk_update->nticks_seqnum != seqno) {
    if (iters++ > 1000000) {
      MX_PRINT_ONCE(("Missed receive hwclock update!\n"));
      break;
    }
    MAL_READBAR();
  }
  nic_ticks = ((uint64_t) ntohl(is->clk_update->nticks_high) << 32) +
              ((uint64_t) ntohl(is->clk_update->nticks_low));

  return nic_ticks;
}

/*
 * Every so often, we sync the host and NIC clocks.
 */
static int
myri_sync_hwclock(mx_instance_state_t *is,
                  myri_clksync_nticks_t *clk_nticks,
                  uint64_t *host_nsecs_sync,
                  uint64_t *nic_nsecs_sync,
                  int64_t *hnic_drift_nsecs_o)
{
  int i, rc, cnt = 0;
  uint64_t nic_ticks, nic_nsecs, host_nsecs;
  uint64_t tsN_begin = 0, tsH_begin = 0;
  int64_t drift_sum = 0;

#define CLOCK_NTRIALS 8
#define CLOCK_KEEP 5 /* leave out worst 3 outliers */
  struct {
    uint64_t host_delay_nsecs;
    int64_t  hnic_drift_nsecs;
  } trials[CLOCK_NTRIALS], trial;

  for (cnt = 0; cnt < CLOCK_NTRIALS; cnt++) {
    if ((rc = myri_sample_hwclock(is, &nic_ticks, &host_nsecs,
                                  &trial.host_delay_nsecs)))
      return rc;

    nic_nsecs = myri_clksync_nticks_update(clk_nticks, nic_ticks);
    if (cnt == 0) {
      tsN_begin = nic_nsecs;
      tsH_begin = host_nsecs;
    }
    trial.hnic_drift_nsecs = (int64_t) host_nsecs - nic_nsecs;
    i = cnt - 1;
    while (i >= 0) {
      if (trials[i].host_delay_nsecs <= trial.host_delay_nsecs)
        break;
      trials[i+1] = trials[i];
      i--;
    }
    trials[i+1] = trial;
  }

  /* The middle trial is where host and nic time should line up on average */
  *nic_nsecs_sync = (tsN_begin + nic_nsecs) / 2;
  *host_nsecs_sync = (tsH_begin + host_nsecs) / 2;

  cnt = CLOCK_KEEP;
  for (i = 0; i < cnt; i++)
    drift_sum += trials[i].hnic_drift_nsecs;
  *hnic_drift_nsecs_o = myri_s64_div(drift_sum, cnt);
  return 0;
}

static
void
myri_clock_handle_error(mx_instance_state_t *is, uint64_t host_nsecs,
                        int64_t error)
{
  if (is->clk_sync_init_cnt == 0) {
    myri_clksync_error_count++;
    if (myri_clksync_verbose)
      MX_INFO(("%s: clksync offset=%+6"PRId64" nsecs (elapsed=%d secs)\n",
          is->is_name, error,
          (int)(myri_u64_div((host_nsecs - is->clk_sync_init_nsecs), 1000000000ull))));
  }
  else {
    /* During initialization */
    const int64_t max_error = 5000;
    const int noerror_cnt = 3;
    const uint64_t max_delay_msecs = 10000ull; /* 10 secs */

    if (MAL_ABS64(error) < max_error) {
      if (++is->clk_sync_init_cnt == noerror_cnt+1) {
        if (myri_clksync_verbose)
          MX_INFO(("%s: clksync init to %lld\n", is->is_name,
                  (long long) error));
        is->clk_sync_init_cnt = 0; /* we're done */
      }
    }
    else {
      uint64_t delay = myri_u64_div((host_nsecs - is->clk_sync_init_nsecs), 1000000ull);
      is->clk_sync_init_cnt = 1; /* start over, need init_error_cnt in a row */
      if (delay >= max_delay_msecs && myri_clksync_verbose) {
        MX_INFO(("%s: clksync init offset=%+6"PRId64" nsecs after %5d msecs\n",
          is->is_name, error, (int) delay));
      }
    }
  }
}

/* 
   Two things slightly complicate synching the time.  The first is that
   the timesource doesn't have a year, so we depend on the
   time of the host for the year.  The second, is that the time we get
   from the timesource isn't known until after the ontime edge, so things
   get a little weird around the end of the year.

   To overcome this we do the following:
 
   We get the year first, then check the timesource time twice at least 
   one second apart. We then check the year a second time.  If the year is 
   still the same, and the timesource has not rolled over, then have a good
   sync, else repeat.

   The way things are done below, the system clock has to be less than
   1 second off of the timesource time.  

   (TODO: we probably set a lower bound so we don't have to 
   do the second checks unless we are on the 365th day of the year
   or later).

*/
/* process state changes for the timesource */

static void 
myri_clock_ts_process(mx_instance_state_t *is, enum myri_timesource_state ts_state) {
  uint64_t nic_nsecs, cycles;
  myri_clksync_nticks_t clk_nticks;
  int rc, year;
  enum myri_timesync_status ts_status;
  unsigned long flags;
  struct timespec ts;

  flags = 0;
  
  switch (ts_state) {
  case MYRI_CLK_TS_INITIALIZED:
  case MYRI_CLK_TS_NO_TIMECODE:
    /* nothing to do */
    return;
    break;

  case MYRI_CLK_TS_INIT_PHASE1:
    /* take the first sample */
    /*MX_INFO(("%d: TS_INIT_PHASE1\n", is->id));*/

    mx_getnstimeofday(&ts);
    year = myri_secs_to_year(ts.tv_sec);

    rc = myri_get_timesync(is, &nic_nsecs, &cycles, &ts_status);
    if (rc != 0 || ts_status == MYRI_TIMESYNC_UNSUPPORTED) {
      MX_WARN(("%d: Fatal error with timesource. rc=%d, ts_status=%d\n",
		 is->id, rc, ts_status));
      is->clk_ts_state = MYRI_CLK_TS_ERROR;
      return;
    } else if (ts_status == MYRI_TIMESYNC_BAD) {
      /* coldfire doesn't have a good irig signal so go back to NO_TIMECODE */
      is->clk_ts_state = MYRI_CLK_TS_NO_TIMECODE;
      is->timesource_data.attempts = 0;
      return;
    }

    is->timesource_data.cycles = cycles;
    is->timesource_data.nsecs = nic_nsecs;
    is->timesource_data.year = year;

    mx_spin_lock_irqsave(&is->clk_spinlock, flags);
    if (ts_state == is->clk_ts_state) {
      is->clk_ts_state = MYRI_CLK_TS_INIT_PHASE2;
    }
    mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);

    is->timesource_data.attempts++;
    if (is->timesource_data.attempts > 5) {
      /* 5 rollovers?, not likely, something else is going on */
      MX_WARN(("%d Failed to sync with timesource after 5 attempts.\n",
		 is->id));
      is->clk_ts_state = MYRI_CLK_TS_ERROR;
      return;
    }

    return;
    break;
  case MYRI_CLK_TS_INIT_PHASE2:
    /* take the second sample */
    /* MX_INFO(("%d: TS_INIT_PHASE2\n", is->id));*/

    rc = myri_get_timesync(is, &nic_nsecs, &cycles, &ts_status);
    if (rc != 0 || ts_status == MYRI_TIMESYNC_UNSUPPORTED) {
      MX_WARN(("%d: Fatal error with timesource. rc=%d, ts_status=%d, p1\n",
		 is->id, rc, ts_status));
      is->clk_ts_state = MYRI_CLK_TS_ERROR;
      return;
    } else if (ts_status == MYRI_TIMESYNC_BAD) {
      /* coldfire doesn't have a good irig signal so go back to NO_TIMECODE */
      is->clk_ts_state = MYRI_CLK_TS_NO_TIMECODE;
      is->timesource_data.attempts = 0;
      return;
    }

    mx_getnstimeofday(&ts);
    year = myri_secs_to_year(ts.tv_sec);

    if ((is->timesource_data.year == year) && 
	(is->timesource_data.nsecs < nic_nsecs)) {
      /* no year changes on host, or rollovers on timesource, good! */
      clk_nticks.ticks = is->timesource_data.cycles;
      clk_nticks.nsecs = is->timesource_data.nsecs + 
	((uint64_t)(myri_year_to_secs(year)) * 1000000000);
      clk_nticks.nt_numerator = 96;
      clk_nticks.nt_denominator = 35;
      clk_nticks.ticks_mask = 0;

      /* write the values out */
      mx_spin_lock_irqsave(&is->clk_spinlock, flags);
      if (ts_state == is->clk_ts_state) {
	is->clk_ts_state = MYRI_CLK_TS_INITIALIZED;

	is->kernel_window->clksync.seqnum++;
	MAL_WRITEBAR();
	is->kernel_window->clksync.clkp = is->clkp_latest;
	is->kernel_window->clksync.nticks = clk_nticks;
	MAL_WRITEBAR();
	is->kernel_window->clksync.seqnum++;
      }
      mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);

      MX_INFO(("%d: Synched with timesource at %u seconds.\n", is->id,
	       (uint32_t)(clk_nticks.nsecs / 1000000000)));
      return;
    } else {
      /* try again */
      mx_spin_lock_irqsave(&is->clk_spinlock, flags);
      if (ts_state == is->clk_ts_state) {
	is->clk_ts_state = MYRI_CLK_TS_NO_TIMECODE;
      }
      mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);
      return;
    }

    break;
  case MYRI_CLK_TS_ERROR:
    /* something is not right, just print error and go back to
       NO_TIMECODE */
    MX_INFO(("%d: Something bad happened during timesource read.  This might be fatal.\n", is->id));
    mx_spin_lock_irqsave(&is->clk_spinlock, flags);
    if (ts_state == is->clk_ts_state) {
      is->clk_ts_state = MYRI_CLK_TS_NO_TIMECODE;
    }
    mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);
    
    return;
    break;
  default:
    MX_WARN(("%d: Bad timesource state: %d\n", is->id, ts_state));
    return;
    break;
  }


}

int
myri_clock_sync_now(mx_instance_state_t *is)
{
  int64_t hnic_drift_nsecs, sync_interval;
  int64_t multiplier, off_delta_nsecs, error;
  uint64_t nic_nsecs, host_nsecs;
  unsigned long flags;
  myri_clksync_nticks_t clk_nticks;

  enum myri_timesource_state ts_state;

  flags = 0;
  ts_state = is->clk_ts_state;

  if (ts_state == MYRI_CLK_NOT_INITIALIZED) {
    return 0;
  } else if (ts_state != MYRI_CLK_HOST_INITIALIZED){
    /*timesource related clock states*/
    myri_clock_ts_process(is, ts_state);
    return 0;
  }
  /* Use the nticks-to-nsecs state as read-copy-update so we don't
   * expose an intermediate state in the kernel window as we're syncing
   * clocks */
  clk_nticks = is->kernel_window->clksync.nticks;

  if (myri_sync_hwclock(is, &clk_nticks, &host_nsecs, &nic_nsecs, &hnic_drift_nsecs))
    return 0;

  off_delta_nsecs = hnic_drift_nsecs - is->clkp_latest.hostnic_drift_nsecs;
  sync_interval = (int64_t) nic_nsecs - is->clkp_latest.nsecs_base;

  /* XXX Need to revisit 64-bit division in 32-bit mode for large divisors */
  if (4 == sizeof(void *))
    mal_assert(sync_interval < 0x100000000LL);

  if (sync_interval >= 15000) {
    /* Ensure our sync interval is large enough */
    multiplier = myri_s64_div((off_delta_nsecs << MYRI_CLKSYNC_MULTIPLIER_SHIFT), sync_interval);
    error = is->clkp_latest.error_nsecs = (int64_t)
      host_nsecs - myri_clksync_convert_nsecs(&is->kernel_window->clksync, nic_nsecs);

    if (is->clk_sync_init_cnt || MAL_ABS64(error) >= (int64_t) myri_clksync_error)
      myri_clock_handle_error(is, host_nsecs, error);

    /* Update the "latest clock" using 75% the new multiplier and 25% the
     * previous multiplier */
    is->clkp_latest.multiplier =
      (3 * multiplier + is->clkp_latest.multiplier) / 4;
    is->clkp_latest.nsecs_base = nic_nsecs;
    is->clkp_latest.hostnic_drift_nsecs = hnic_drift_nsecs;

    mx_spin_lock_irqsave(&is->clk_spinlock, flags);
    is->kernel_window->clksync.seqnum++;
    MAL_WRITEBAR();
    is->kernel_window->clksync.clkp = is->clkp_latest;
    is->kernel_window->clksync.nticks = clk_nticks;
    MAL_WRITEBAR();
    is->kernel_window->clksync.seqnum++;
    mx_spin_unlock_irqrestore(&is->clk_spinlock, flags);
  }
  return 0;
}

/*
 * atoi() for the kernel
 */
int
myri_atoi(
  char *s)
{
  int val;
  int sign;

  sign = 1;
  val = 0;

  if (*s == '-') {
    sign = -1;
    s++;
  }
  while (*s >= '0' && *s <= '9') {
    val = (val * 10) + (*s & 0x0F);
    s++;
  }
  if (sign < 0) {
    val = 0 - val;
  }
  return val;
}

#if MX_KERNEL_LIB
#include "mx_arch_klib.h"
mx_klib_symbol(mx_common_ioctl);
mx_klib_symbol(mx_endptless_ioctl);
mx_klib_symbol(mx_mmap_off_to_kva);
mx_klib_symbol(myri_common_open);
mx_klib_symbol(mx_common_close);
#endif
