/************************************************************************
 * The contents of this file are subject to the MYRICOM MYRINET		*
 * EXPRESS (MX) NETWORKING SOFTWARE AND DOCUMENTATION LICENSE (the	*
 * "License"); User may not use this file except in compliance with the	*
 * License.  The full text of the License can found in LICENSE.TXT	*
 *									*
 * Software distributed under the License is distributed on an "AS IS"	*
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See	*
 * the License for the specific language governing rights and		*
 * limitations under the License.					*
 *									*
 * Copyright 2011 by Myricom, Inc.  All rights reserved.		*
 ************************************************************************/

#include <linux/delay.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/if_arp.h>
#include <linux/kthread.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/pci.h>
#include <linux/rtnetlink.h>

#include "myri_eosnf_kcompat.h"
#include "snf.h"

#define MYRI_EOSNF_NAME	"Myricom Ethernet over SNF driver"
#define MYRI_EOSNF_VER	"1.0"

MODULE_AUTHOR("Maintainer: help@myri.com");
MODULE_DESCRIPTION(MYRI_EOSNF_NAME);
MODULE_LICENSE("Dual BSD/GPL");
MODULE_VERSION(MYRI_EOSNF_VER);

unsigned int cpu_mask = 0;
module_param(cpu_mask, uint, 0);
MODULE_PARM_DESC(cpu_mask,
		 "Mask to specify on which CPUs the receive threads run");

unsigned int num_rings = 0;
module_param(num_rings, uint, 0);
MODULE_PARM_DESC(num_rings,
		 "Number of transmit and receive threads per device");

unsigned int rx_timeout_ms = 1000;
module_param(rx_timeout_ms, uint, 0);
MODULE_PARM_DESC(rx_timeout_ms,
		 "Length of time to wait for an incoming packet (in msecs)");

#define STATS_TIMER_LEN		1000
#define ETHER_MAX_LEN_JUMBO	9018

struct myri_eosnf_rx {
	int cpu_id;
	struct net_device *netdev;
	snf_ring_t ring;
	struct task_struct *kthread;
	/* Ring stats */
	struct snf_ring_stats ring_stats;
	u64 rx_packets;
	u64 rx_bytes;
	u64 rx_dropped;
	u64 rx_length_errors;
};

struct myri_eosnf_tx {
	snf_inject_t inj;
	/* tx stats */
	struct snf_inject_stats inj_stats;
	u64 tx_packets;
	u64 tx_bytes;
	u64 tx_dropped;
};

struct myri_eosnf_dev {
	struct list_head list;
	struct net_device *netdev;
	u32 msg_enable;

	cpumask_t cpu_mask;
	uint32_t portnum;
	snf_handle_t handle;

	/* rx vars */
	int num_rx_threads;
	struct myri_eosnf_rx *rx;

	/* tx vars */
	int num_queues;
	struct myri_eosnf_tx *tx;

	/* stats vars */
	struct delayed_work stats_work;
	struct net_device_stats stats;
};

struct list_head dev_list;

#define for_each_myri_dev(d) while ((d = pci_get_device(PCI_VENDOR_ID_MYRICOM, PCI_ANY_ID, d)) != NULL)

static void
myri_eosnf_get_drvinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
	strcpy(info->driver, KBUILD_MODNAME);
	strcpy(info->version, MYRI_EOSNF_VER);
}

static const char myri_eosnf_nic_stats[][ETH_GSTRING_LEN] = {
	"nic_pkt_recv", "nic_bytes_recv", "nic_pkt_overflow", "nic_pkt_bad",
	"snf_pkt_overflow",
	"nic_pkt_send", "nic_bytes_send",
};

static const char myri_eosnf_ring_stats[][ETH_GSTRING_LEN] = {
	"ring", "ring_pkt_recv",
	"rx_packets", "rx_bytes", "rx_dropped", "rx_length_errors",
};

static const char myri_eosnf_inj_stats[][ETH_GSTRING_LEN] = {
	"queue", "inj_pkt_send",
	"tx_packets", "tx_bytes", "tx_dropped",
};

static int myri_eosnf_get_stats_count(struct net_device *dev)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	int count;

	/* common stats */
	count = ARRAY_SIZE(myri_eosnf_nic_stats);

	/* rx ring stats */
	count += snf->num_rx_threads * ARRAY_SIZE(myri_eosnf_ring_stats);

	/* tx stats */
	count += snf->num_queues * ARRAY_SIZE(myri_eosnf_inj_stats);

	return count;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,32)
static int myri_eosnf_get_sset_count(struct net_device *dev, int sset)
{
	switch (sset) {
	case ETH_SS_STATS:
		return myri_eosnf_get_stats_count(dev);
	default:
		return -EOPNOTSUPP;
	}
}
#endif

static void myri_eosnf_get_strings(struct net_device *dev, u32 sset, u8 *data)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	int i;

	switch (sset) {
	case ETH_SS_STATS:
		memcpy(data, *myri_eosnf_nic_stats,
		       sizeof(myri_eosnf_nic_stats));
		data += sizeof(myri_eosnf_nic_stats);
		for (i = 0; i < snf->num_rx_threads; i++) {
			memcpy(data, *myri_eosnf_ring_stats,
			       sizeof(myri_eosnf_ring_stats));
			data += sizeof(myri_eosnf_ring_stats);
		}
		for (i = 0; i < snf->num_queues; i++) {
			memcpy(data, *myri_eosnf_inj_stats,
			       sizeof(myri_eosnf_inj_stats));
			data += sizeof(myri_eosnf_inj_stats);
		}
		break;
	}
}

static void myri_eosnf_get_ethtool_stats(struct net_device *dev,
					 struct ethtool_stats *stats,
					 u64 *data)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	int i = 0, ring;

	data[i++] = snf->rx[0].ring_stats.nic_pkt_recv;
	data[i++] = snf->rx[0].ring_stats.nic_bytes_recv;
	data[i++] = snf->rx[0].ring_stats.nic_pkt_overflow;
	data[i++] = snf->rx[0].ring_stats.nic_pkt_bad;
	data[i++] = snf->rx[0].ring_stats.snf_pkt_overflow;

	data[i++] = snf->tx[0].inj_stats.nic_pkt_send;
	data[i++] = snf->tx[0].inj_stats.nic_bytes_send;

	for (ring = 0; ring < snf->num_rx_threads; ring++) {
		data[i++] = ring;
		data[i++] = snf->rx[ring].ring_stats.ring_pkt_recv;
		data[i++] = snf->rx[ring].rx_packets;
		data[i++] = snf->rx[ring].rx_bytes;
		data[i++] = snf->rx[ring].rx_dropped;
		data[i++] = snf->rx[ring].rx_length_errors;
	}

	for (ring = 0; ring < snf->num_queues; ring++) {
		data[i++] = ring;
		data[i++] = snf->tx[ring].inj_stats.inj_pkt_send;
		data[i++] = snf->tx[ring].tx_packets;
		data[i++] = snf->tx[ring].tx_bytes;
		data[i++] = snf->tx[ring].tx_dropped;
	}
}

static struct ethtool_ops myri_eosnf_ethtool_ops = {
	.get_drvinfo = myri_eosnf_get_drvinfo,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,33)
	.get_stats_count = myri_eosnf_get_stats_count,
#else
	.get_sset_count = myri_eosnf_get_sset_count,
#endif
	.get_strings = myri_eosnf_get_strings,
	.get_ethtool_stats = myri_eosnf_get_ethtool_stats,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,3,0)
	.get_sg = ethtool_op_get_sg,
	.set_sg = ethtool_op_set_sg,
	.get_tx_csum = ethtool_op_get_tx_csum,
	.set_tx_csum = ethtool_op_set_tx_hw_csum,
#endif
};

static int myri_eosnf_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	struct snf_pkt_fragment spf[MAX_SKB_FRAGS + 1];
	int i, rc, queue;
	u32 len;

	queue = skb_get_queue_mapping(skb);

	/* This should only be necessary if SG is enabled */
	if (skb->ip_summed == CHECKSUM_PARTIAL) {
		rc = skb_checksum_help(skb);
		if (rc)
			goto err;
	}

	spf[0].ptr = skb->data;
	if (skb_shinfo(skb)->nr_frags)
		spf[0].length = len = skb_headlen(skb);
	else {
		/* FIXME - the docs state that it should not need to be padded,
		 * but it currently does need to be.
		 */
		if (skb->len < ETH_ZLEN) {
			rc = skb_padto(skb, ETH_ZLEN);
			if (rc)
				goto err;

			skb->len = ETH_ZLEN;
		}

		spf[0].length = len = skb->len;
	}

	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
		skb_frag_t *frag = &skb_shinfo(skb)->frags[i];

		spf[i + 1].ptr = skb_frag_address(frag);
		spf[i + 1].length = skb_frag_size(frag);

		len += skb_frag_size(frag);
	}

	rc = snf_inject_send_v(snf->tx[queue].inj, 0, 0, spf, i + 1, len);
	if (rc) {
		if (rc == EAGAIN)
			return NETDEV_TX_BUSY;
		else
			goto err;
	}

	dev_kfree_skb_any(skb);

	snf->tx[queue].tx_packets++;
	snf->tx[queue].tx_bytes += skb->len;

#ifndef RHEL_RELEASE_CODE
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,7,0)
        dev->trans_start = jiffies;
#else
        netif_trans_update(dev);
#endif
#else /* else handle RHEL case */
#if RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(7,4)
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,7,0)
        dev->trans_start = jiffies;
#else
        netif_trans_update(dev);
#endif
#endif
#endif
	return NETDEV_TX_OK;

err:
	snf->tx[queue].tx_dropped++;
	return NETDEV_TX_OK;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void snf_eth_stats_thread(void *data)
{
	struct myri_eosnf_dev *snf = data;
#else
static void snf_eth_stats_thread(struct work_struct *work)
{
	struct delayed_work *dw = container_of(work, struct delayed_work, work);
	struct myri_eosnf_dev *snf = container_of(dw, struct myri_eosnf_dev,
						  stats_work);
#endif
	enum snf_link_state state;
	int i, rc;
	u64 tx_packets = 0, tx_bytes = 0, tx_dropped = 0;
	u64 rx_packets = 0, rx_bytes = 0, rx_dropped = 0, rx_length_errors = 0;

	rc = snf_get_link_state(snf->handle, &state);
	if (!rc && netif_carrier_ok(snf->netdev) != state) {
		if (state) {
			netif_carrier_on(snf->netdev);
			netif_info(snf, link, snf->netdev, "Link Up\n");
		} else {
			netif_carrier_off(snf->netdev);
			netif_info(snf, link, snf->netdev, "Link Down\n");
		}
	}

	for (i = 0; i < snf->num_rx_threads; i++) {
		snf_ring_getstats(snf->rx[i].ring, &snf->rx[i].ring_stats);
		rx_packets += snf->rx[i].rx_packets;
		rx_bytes += snf->rx[i].rx_bytes;
		rx_dropped += snf->rx[i].rx_dropped;
		rx_length_errors += snf->rx[i].rx_length_errors;
	}
	snf->stats.rx_packets = rx_packets;
	snf->stats.rx_bytes = rx_bytes;
	snf->stats.rx_dropped = rx_dropped;
	snf->stats.rx_errors = rx_dropped;
	snf->stats.rx_length_errors = rx_length_errors;

	for (i = 0; i < snf->num_queues; i++) {
		snf_inject_getstats(snf->tx[i].inj, &snf->tx[i].inj_stats);
		tx_packets += snf->tx[i].tx_packets;
		tx_bytes += snf->tx[i].tx_bytes;
		tx_dropped += snf->tx[i].tx_dropped;
	}
	snf->stats.tx_packets = tx_packets;
	snf->stats.tx_bytes = tx_bytes;
	snf->stats.tx_dropped = tx_dropped;
	snf->stats.tx_errors = tx_dropped;

	schedule_delayed_work(&snf->stats_work,
			      msecs_to_jiffies(STATS_TIMER_LEN));
}

static int snf_eth_worker(void *data)
{
	struct myri_eosnf_rx *rx = data;
	struct net_device *dev = rx->netdev;
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	struct snf_recv_req recv_req;
	struct sk_buff *skb;
	u32 cpu_id, len;
	int rc;

	cpu_id = rx->cpu_id;

	while (!kthread_should_stop()) {
		rc = snf_ring_recv(rx->ring, rx_timeout_ms, &recv_req);
		if (rc) {
			if (rc == EINTR || rc == EAGAIN)
				continue;
			else {
				snf->stats.rx_errors++;
				netdev_err(dev, "snf_ring_recv failed on cpu "
					   "%d, err = %d\n", cpu_id, rc);
				return -rc;
			}
		}

		len = recv_req.length;

		if (len > ETHER_MAX_LEN_JUMBO) {
			rx->rx_length_errors++;
			rx->rx_dropped++;
			continue;
		}

		skb = alloc_skb(len, GFP_KERNEL);
		if (skb == NULL) {
			rx->rx_dropped++;
			continue;
		}

		/* copy the packet into the skb */
		skb_copy_to_linear_data(skb, recv_req.pkt_addr, len);
		skb_put(skb, len);
		skb->ip_summed = CHECKSUM_NONE;
		skb->protocol = eth_type_trans(skb, dev);
		skb->dev = dev;
		skb_record_rx_queue(skb, cpu_id);

		netif_rx_ni(skb);

		rx->rx_packets++;
		rx->rx_bytes += len;
	}

	return 0;
}

static int myri_eosnf_close(struct net_device *dev)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	int i, rc;

	rc = snf_stop(snf->handle);
	if (rc)
		netdev_err(dev, "snf_stop failed\n");

	cancel_delayed_work_sync(&snf->stats_work);

	for (i = 0; i < snf->num_rx_threads; i++) {
		kthread_stop(snf->rx[i].kthread);

		rc = snf_ring_close(snf->rx[i].ring);
		if (rc)
			netdev_err(dev, "snf_ring_close failed\n");
	}

	for (i = 0; i < snf->num_queues; i++) {
		rc = snf_inject_close(snf->tx[i].inj);
		if (rc)
			netdev_err(dev, "snf_inject_close failed\n");
	}

	snf_close(snf->handle);

	return 0;
}

static int myri_eosnf_open(struct net_device *dev)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);
	int rc, cpu, q, i = 0;

	/* Open snf with 1 receive worker on each requested CPU */
	rc = snf_init(SNF_VERSION_API);
	if (rc) {
		netdev_err(dev, "snf_init failed with %d\n", rc);
		return -rc;
	}

	rc = snf_open(snf->portnum, snf->num_rx_threads, NULL, 0, -1,
		      &snf->handle);
	if (rc) {
		netdev_err(dev, "snf_open failed with %d\n", rc);
		return -rc;
	}

	/* Initialize transmit */
	for (q = 0; q < snf->num_queues; q++) {
		rc = snf_inject_open(snf->portnum, 0, &snf->tx[q].inj);
		if (rc) {
			netdev_err(dev, "snf_inject_open failed with %d\n", rc);
			rc = -rc;
			goto err;
		}
	}

	/* Create Rx handler threads on each cpu */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
	for_each_cpu(cpu, &(snf->cpu_mask)) {
#else
	for_each_cpu_mask(cpu, snf->cpu_mask) {
#endif
		snf->rx[i].cpu_id = cpu;
		snf->rx[i].netdev = dev;

		rc = snf_ring_open_id(snf->handle, i, &snf->rx[i].ring);
		if (rc) {
			netdev_err(dev, "snf_ring_open_id for cpu %d fail\n",
				   i);
			rc = -rc;
			goto err1;
		}

		snf->rx[i].kthread = kthread_create(snf_eth_worker, &snf->rx[i],
						    "eosnf_rx%d", i);
		rc = IS_ERR(snf->rx[i].kthread);
		if (rc) {
			netdev_err(dev, "kthread_create fail\n");
			goto err1;
		}
		kthread_bind(snf->rx[i].kthread, cpu);
		wake_up_process(snf->rx[i].kthread);

		netdev_info(dev, "Starting rx_work on cpu %d\n", cpu);
		i++;
	}

	/* Create delayed thread to query and update stats */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	INIT_WORK(&snf->stats_work, snf_eth_stats_thread, snf);
#else
	INIT_DELAYED_WORK(&snf->stats_work, snf_eth_stats_thread);
#endif
	rc = schedule_delayed_work(&snf->stats_work,
				   msecs_to_jiffies(STATS_TIMER_LEN));
	if (!rc)
		goto err1;

	rc = snf_start(snf->handle);
	if (rc) {
		netdev_err(dev, "snf_start failed\n");
		rc = -rc;
		goto err2;
	}

	return 0;

err2:
	cancel_delayed_work_sync(&snf->stats_work);
err1:
	while (--i >= 0) {
		kthread_stop(snf->rx[i].kthread);
		snf_ring_close(snf->rx[i].ring);
	}
err:
	while (--q >= 0)
		snf_inject_close(snf->tx[q].inj);
	snf_close(snf->handle);
	return rc;
}

static struct net_device_stats *myri_eosnf_get_stats(struct net_device *dev)
{
	struct myri_eosnf_dev *snf = netdev_priv(dev);

	return &snf->stats;
}

static int myri_eosnf_change_mtu(struct net_device *dev, int new_mtu)
{
	if (new_mtu < ETH_ZLEN || new_mtu > ETHER_MAX_LEN_JUMBO)
		return -EINVAL;

	dev->mtu = new_mtu;

	return 0;
}

#ifdef HAVE_NET_DEVICE_OPS
static const struct net_device_ops myri_eosnf_netdev_ops = {
	.ndo_open = myri_eosnf_open,
	.ndo_stop = myri_eosnf_close,
	.ndo_start_xmit = myri_eosnf_xmit,
	.ndo_get_stats = myri_eosnf_get_stats,
#ifndef RHEL_RELEASE_CODE
	.ndo_change_mtu = myri_eosnf_change_mtu,
#else
#if RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 4)
	.ndo_change_mtu = myri_eosnf_change_mtu,
#else
#if RHEL_RELEASE_CODE <= RHEL_RELEASE_VERSION(7, 9)
	.extended.ndo_change_mtu = myri_eosnf_change_mtu,
#else
	.ndo_change_mtu = myri_eosnf_change_mtu,
#endif
#endif
#endif
};
#endif

static void myri_eosnf_init_netdev_ops(struct net_device *dev)
{
#ifdef HAVE_NET_DEVICE_OPS
	dev->netdev_ops = &myri_eosnf_netdev_ops;
#else
	dev->open = myri_eosnf_open;
	dev->stop = myri_eosnf_close;
	dev->get_stats = myri_eosnf_get_stats;
	dev->hard_start_xmit = myri_eosnf_xmit;
	dev->change_mtu = myri_eosnf_change_mtu;
#endif
	/* NOTE: It is not possible to have a transmit timeout */
}

static void myri_eosnf_detach(struct myri_eosnf_dev *snf)
{
	struct net_device *dev = snf->netdev;

	netdev_info(dev, "EoSNF device removed\n");

	unregister_netdev(dev);
	kfree(snf->tx);
	kfree(snf->rx);
	free_netdev(dev);
}

static void myri_eosnf_purge(void)
{
	struct myri_eosnf_dev *snf;

	list_for_each_entry(snf, &dev_list, list)
	    myri_eosnf_detach(snf);
}

cpumask_t mask; // size too large for stack so moved out for now.
static int __init
myri_eosnf_attach(struct pci_dev *pdev, int portnum, char *mac_addr)
{
	struct myri_eosnf_dev *snf;
	struct net_device *dev;
	u32 num_cpus;
	int rc;
	const cpumask_t *numa_mask;

	numa_mask = cpumask_of_pcibus(pdev->bus);

	if (cpu_mask) {
		u32 cpu, m = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
		for_each_cpu(cpu, numa_mask) {
#else
		for_each_cpu_mask(cpu, *numa_mask) {
#endif
			m = (1 << cpu);
			if (m & cpu_mask) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
				cpumask_set_cpu(cpu, &mask);
#else
				cpu_set(cpu, mask);
#endif
				m &= ~m;
			}
		}

		if (m)
			pr_err("CPUs %x are not on this node, ignoring\n", m);
	} else if (num_rings) {
		int cpu, rings;

		if (num_rings > cpumask_weight(numa_mask)) {
			pr_err("Specified more %d rings than CPUs in node\n",
				num_rings - cpumask_weight(numa_mask));
			num_rings = cpumask_weight(numa_mask);
		}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
		cpumask_clear(&mask);
		rings = num_rings;
		for_each_cpu(cpu, numa_mask) {
			cpumask_set_cpu(cpu, &mask);
#else
		cpus_clear(mask);
		rings = num_rings;
		for_each_cpu_mask(cpu, *numa_mask) {
			cpu_set(cpu, mask);
#endif
			if (--rings == 0)
				break;
		}
	} else
		cpumask_copy(&mask, numa_mask);

	num_cpus = cpumask_weight(&mask);
        if (num_cpus == 0) {
                /* Not all kernels capable of producing useful pcibus cpumasks */
                if (cpumask_weight(numa_mask) == 0) {
                        pr_err("NUMA configuration cannot list cpus of PCI bus "
                               "(using 1 ring)\n");
                }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0)
                cpumask_set_cpu(1, &mask);
#else
                cpu_set(1, mask);
#endif
                num_cpus = 1;
        }
	dev = alloc_etherdev_mq((sizeof(struct myri_eosnf_dev)), num_cpus);
	if (!dev)
		return -ENOMEM;

	SET_NETDEV_DEV(dev, &pdev->dev);

	memcpy(dev->dev_addr, mac_addr, dev->addr_len);
	/* Set the "locally administered" bit to make it different than the
	 * MAC address on the underlying hardware.
	 */
	dev->dev_addr[0] |= 0x2;
	memcpy(dev->perm_addr, dev->dev_addr, dev->addr_len);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
	SET_ETHTOOL_OPS(dev, &myri_eosnf_ethtool_ops);
#else
        netdev_set_default_ethtool_ops(dev, &myri_eosnf_ethtool_ops);
#endif
	myri_eosnf_init_netdev_ops(dev);
	dev->features = NETIF_F_HIGHDMA;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,3,0)
	dev->hw_features = NETIF_F_HIGHDMA | NETIF_F_SG | NETIF_F_HW_CSUM;
#endif
	snf = netdev_priv(dev);
	cpumask_copy(&snf->cpu_mask, &mask);
	snf->netdev = dev;
	snf->msg_enable = ~0;
	snf->portnum = portnum;
	snf->num_rx_threads = num_cpus;
	snf->num_queues = NUM_TX_QUEUES(num_cpus);
	netif_set_real_num_tx_queues(dev, snf->num_queues);
	netif_set_real_num_rx_queues(dev, snf->num_rx_threads);

	/* Since stats can be queried when offline, alloc them now */
	snf->rx = kcalloc(snf->num_rx_threads, sizeof(struct myri_eosnf_rx),
			  GFP_KERNEL);
	if (!snf->rx) {
		rc = -ENOMEM;
		goto err;
	}

	snf->tx = kcalloc(snf->num_queues, sizeof(struct myri_eosnf_tx),
			  GFP_KERNEL);
	if (!snf->tx) {
		rc = -ENOMEM;
		goto err1;
	}

	netif_carrier_off(dev);
	rc = register_netdev(dev);
	if (rc)
		goto err2;

	netdev_info(dev, "EoSNF device ready\n");

	list_add_tail(&snf->list, &dev_list);

	return 0;

err2:
	kfree(snf->tx);
err1:
	kfree(snf->rx);
err:
	free_netdev(dev);
	return rc;
}

static struct pci_dev *__init myri_find_pdev(char *mac_addr)
{
	struct net_device *dev;
	struct pci_dev *pdev = NULL;

	rtnl_lock();
	dev = dev_getbyhwaddr_rcu(&init_net, ARPHRD_ETHER, mac_addr);
	rtnl_unlock();
	if (!dev)
		goto err;

	for_each_myri_dev(pdev)
		if (dev->base_addr == pci_resource_start(pdev, 0))
			break;

	if (!pdev)
		pr_err("No PCI device found for %s\n", dev->name);

err:
	return pdev;
}

static int __init myri_eosnf_mod_init(void)
{
	struct snf_ifaddrs *ifa, *ifan;
	int rc;

	pr_info("%s, version %s\n", MYRI_EOSNF_NAME, MYRI_EOSNF_VER);

	INIT_LIST_HEAD(&dev_list);

	rc = snf_getifaddrs(&ifa);
	if (rc)
		return -rc;

	for (ifan = ifa; ifan; ifan = ifan->snf_ifa_next) {
		char *macaddr = ifan->snf_ifa_macaddr;
		struct pci_dev *pdev;

		pdev = myri_find_pdev(macaddr);
		if (!pdev) {
			rc = -ENODEV;
			goto err;
		}

		rc = myri_eosnf_attach(pdev, ifan->snf_ifa_portnum, macaddr);
		if (rc)
			goto err;
	}

	snf_freeifaddrs(ifa);
	return 0;

err:
	myri_eosnf_purge();
	return rc;
}

static void __exit myri_eosnf_mod_exit(void)
{
	myri_eosnf_purge();
}

module_init(myri_eosnf_mod_init);
module_exit(myri_eosnf_mod_exit);
